! $Id: AllocatableStmts.f90 2810 2019-03-23 23:08:54Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the AllocatableStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /allocatable-stmts/.

MODULE AllocatableStmts
  
  USE CompilerKinds
  USE Identifiers
  USE ParseNodes
  USE SourceLocations
  USE SubscriptLists
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseAllocatableStmt
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing an /allocatable-decl/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: AllocatableDecl
    !> The name of the object being declared.
    TYPE(Identifier) :: name
    !> An /array-spec/, if present, or not allocated if not.
    TYPE(DimSpec), ALLOCATABLE :: array_spec(:)
    !> A /coarray-spec/, if present, or not allocated if not.
    TYPE(DimSpec), ALLOCATABLE :: coarray_spec(:)
  CONTAINS
    PROCEDURE :: Write => ad_Write
    PROCEDURE :: Query => ad_Query
    PROCEDURE :: Modify => ad_Modify
    PROCEDURE, NOPASS :: What => ad_What
    PROCEDURE :: Advertise => ad_Advertise
    PROCEDURE :: GetLocation => ad_GetLocation
  END TYPE AllocatableDecl
  
  !> The results of parsing an /allocatable-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: AllocatableStmt
    !> The items in the /allocatable-decl-list/
    TYPE(AllocatableDecl), ALLOCATABLE :: decls(:)
  CONTAINS
    PROCEDURE :: Dump => alloc_Dump
    PROCEDURE :: Query => alloc_Query
    PROCEDURE :: Modify => alloc_Modify
    PROCEDURE, NOPASS :: What => alloc_What
  END TYPE AllocatableStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /allocatable-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    alloc_stmt        The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseAllocatableStmt(tlist, alloc_stmt, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(AllocatableStmt), INTENT(OUT) :: alloc_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /allocatable-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R526'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: list_count     ! Number of items in the /allocatable-decl-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /allocatable-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'ALLOCATABLE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! option :: before name list.
    
    CALL Expect(  &
        tlist, 2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  &
        '"::" or /object-name/',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    
    !---------------------------------------------------------------------------
    ! /allocatable-decl-list/
    
    list_count = ListItemCount(tlist(it:))
    ALLOCATE(alloc_stmt%decls(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, SIZE(tlist), item_start, item_finish)
      CALL parse_allocatable_decl(  &
          tlist,  &
          item_start,  &
          item_finish,  &
          alloc_stmt%decls(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! That's it!
    
  END SUBROUTINE ParseAllocatableStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /allocatable-decl/ (R527).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the start of the 
  !! /allocatable-decl/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the end of the 
  !! /allocatable-decl/ in @a tlist.
  !!
  !! @param[out]    decl              Results of parsing the 
  !! /allocatable-decl/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_allocatable_decl(tlist, start, finish, decl, err_list)
    
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(AllocatableDecl), INTENT(OUT) :: decl
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /allocatable-decl/.
    CHARACTER(*), PARAMETER :: comp = 'R527'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err          ! Error object for Expect calls.
    INTEGER :: opt              ! Expect option
    INTEGER :: closing          ! Index of closing parenthesis.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: it               ! Current token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ExpectName(tlist, start, '/object-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    decl%name = IdentifierIfortBug(tlist(start))
    
    ! That can be it.
    IF (start == finish) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional ( /array-spec/ ).  While the syntax rules do this indirectly 
    ! (unlike for a POINTER statement - why??) this spec must be a 
    ! /deferred-shape-spec-list/ due to constraint C532.  We enforce that here.
    
    it = start + 1
    CALL Expect(  &
        tlist, it,  &
        [ittDelimiter, ittDelimiter],  &
        [scck_'(', scck_'['],  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN    ! we have a (
      ! Find location of closing ')'.
      CALL ExpectClosing(tlist, start + 1, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ! Parse the array-spec.  We need to check that it is a deferred 
      ! shape spec elsewhere.
      CALL ParseArraySpec(  &
          tlist(start+2:closing-1),  &
          .FALSE.,  &
          decl%array_spec,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      ! An alternative to this might be to call ParseDeferredShapeSpecList, 
      ! like we do when parsing pointer statements.  This might give a 
      ! clearer message though.
      IF (.NOT. IsDeferredShapeSpecList(decl%array_spec)) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C532',  &
            LOCATION=QueryLocation(tlist(start+2)),  &
            MSG='An array with the ALLOCATABLE attribute shall have a &
              &/array-spec/ that is a /deferred-shape-spec-list/.' )
      END IF
      it = closing + 1
      
      ! That can be it.
      IF (it > finish) RETURN
      
      CALL Expect(tlist, it, ittDelimiter, scck_'[', err) 
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    END IF
    
    ! it here index of opening lbracket.
    
    !---------------------------------------------------------------------------
    ! lbracket /coarray-spec/ rbracket.
    
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ParseCoarraySpec(  &
        tlist(it+1:closing-1),  &
        decl%coarray_spec,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! That's all folks.
    CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
    
  END SUBROUTINE parse_allocatable_decl
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableStmt%Dump - write a debugging representation 
  !! of the results of parsing an allocatable statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE alloc_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocatableStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Declaration list index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/allocatable-decl-list/')") pfx
    DO i = 1, SIZE(pnode%decls)
      ASSOCIATE(decl => pnode%decls(i))
        WRITE (unit, "(A,A)", ADVANCE='NO') pfx // '  ', .NameOf. decl%name
        IF (ALLOCATED(decl%array_spec)) THEN
          WRITE (unit, "('(')", ADVANCE='NO')
          CALL WriteDimSpec(unit, decl%array_spec)
          WRITE (unit, "(')')", ADVANCE='NO')
        END IF
        IF (ALLOCATED(decl%coarray_spec)) THEN
          WRITE (unit, "('[')", ADVANCE='NO')
          CALL WriteDimSpec(unit, decl%coarray_spec)
          WRITE (unit, "(']')", ADVANCE='NO')
        END IF
        WRITE (unit, "()")
      END ASSOCIATE
    END DO
    
  END SUBROUTINE alloc_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE alloc_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocatableStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE alloc_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object to execution the operation.
  
  SUBROUTINE alloc_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocatableStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE alloc_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'allocatable-stmt'.
  
  FUNCTION alloc_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'allocatable-stmt'
    
  END FUNCTION alloc_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for AllocatableDecl.
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableDecl%Write - write a text representation of 
  !! the parse node to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE ad_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocatableDecl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "(A)", ADVANCE='NO') .NameOf. pnode%name
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      WRITE (unit, "(' ( ')", ADVANCE='NO')
      CALL WriteDimSpec(unit, pnode%array_spec)
      WRITE (unit, "(' ) ')", ADVANCE='NO')
    END IF
    
    IF (ALLOCATED(pnode%coarray_spec)) THEN
      WRITE (unit, "(' [ ')", ADVANCE='NO')
      CALL WriteDimSpec(unit, pnode%coarray_spec)
      WRITE (unit, "(' ] ')", ADVANCE='NO')
    END IF
    
  END SUBROUTINE ad_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableDecl%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE ad_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(AllocatableDecl), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Subscript index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      DO i = 1, SIZE(pnode%array_spec)
        CALL pnode%array_spec(i)%Query(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%coarray_spec)) THEN
      DO i = 1, SIZE(pnode%coarray_spec)
        CALL pnode%coarray_spec(i)%Query(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ad_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableDecl%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the modify 
  !! action.
  
  SUBROUTINE ad_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(AllocatableDecl), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Subscript index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      DO i = 1, SIZE(pnode%array_spec)
        CALL pnode%array_spec(i)%Modify(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%coarray_spec)) THEN
      DO i = 1, SIZE(pnode%coarray_spec)
        CALL pnode%coarray_spec(i)%Modify(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ad_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableDecl%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION ad_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'allocatable-decl'
    
  END FUNCTION ad_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocatableDecl%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION ad_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(AllocatableDecl), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION ad_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Advertise identifiers in an allocatable declarator.
  !!
  !! @param[in]     pnode             The declarator to advertise.
  !!
  !! @param[in,out] ident_list        The list of identifiers.
  
  SUBROUTINE ad_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocatableDecl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureDefAllocatable],  &
          LOCATIONS=[.LocationOf. pnode%name] )
    
  END SUBROUTINE ad_Advertise
  
END MODULE AllocatableStmts
