! $Id: AsynchronousStmts.f90 2205 2016-06-09 23:36:40Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the AsynchronousStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /asynchronous-stmt/'s.

MODULE AsynchronousStmts
  
  USE Identifiers
  USE ParseNodes
  USE Strings
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseAsynchronousStmt
  
  !> Parse node for an asynchronous statement.
  !!
  !! @todo Consider whether this could be an extension of NameListPNode
  TYPE, EXTENDS(SCParseNode), PUBLIC :: AsynchronousStmt
    !> The list of objects in the statement.
    TYPE(Identifier), ALLOCATABLE :: object_name_list(:)
  CONTAINS
    PROCEDURE :: Dump => async_Dump
    PROCEDURE :: Query => async_Query
    PROCEDURE :: Modify => async_Modify
    PROCEDURE, NOPASS :: What => async_What
    PROCEDURE :: Advertise => async_Advertise
  END TYPE AsynchronousStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /asynchronous-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up 
  !! the /asynchronous-stmt/.
  !!
  !! @param[out]    async_stmt        The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseAsynchronousStmt(tlist, async_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(AsynchronousStmt), INTENT(OUT) :: async_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    CHARACTER(*), PARAMETER :: comp = 'R528'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err      ! Error object for Expect calls.
    INTEGER :: it           ! Current token index.
    INTEGER :: opt          ! Expect option return.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /asynchronous-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'ASYNCHRONOUS', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! :: or /object-name-list/
    
    CALL Expect(  &
        tlist, 2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  & 
        '"::" or /object-name/',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    
    !---------------------------------------------------------------------------
    ! The /object-name-list/
    
    CALL ParseNameList(  &
        tlist,  &
        it, SIZE(tlist),  &
        '/object-name/',  &
        async_stmt%object_name_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! That's it!
    
  END SUBROUTINE ParseAsynchronousStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of AsynchronousStmt%Dump - write a debugging representation 
  !! of the results of parsing an allocatable statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE async_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AsynchronousStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Declaration list index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/object-name-list/')") pfx
    DO i = 1, SIZE(pnode%object_name_list)
      WRITE (unit, "(A,A)") pfx // '  ', .NameOf. pnode%object_name_list(i)
    END DO
    
  END SUBROUTINE async_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of AsynchronousStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE async_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AsynchronousStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE async_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AsynchronousStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE async_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AsynchronousStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE async_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of AsynchronousStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'asynchronous-stmt'.
  
  FUNCTION async_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'asynchronous-stmt'
    
  END FUNCTION async_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of AsynchronousStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE async_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    USE IntentSpecs
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AsynchronousStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Identifier index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(SIZE(pnode%object_name_list)))
    
    DO i = 1, SIZE(pnode%object_name_list)
      ident_list(i) = IdentifierNature(  &
          NAME=.NameOf. pnode%object_name_list(i),  &
          NATURES=[NatureDefAsynchronous], &
          LOCATIONS=[.LocationOf. pnode%object_name_list(i)] )
    END DO
    
  END SUBROUTINE async_Advertise
  
END MODULE AsynchronousStmts
