! $Id: BackspaceStmts.f90 2892 2019-04-07 00:34:44Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the BackspaceStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /backspace-stmt/'s (R924).

MODULE BackspaceStmts
  
  USE ParseNodes
  USE PositionSpecs
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseBackspaceStmt
  
  !> Describes the results of parsing a backspace statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: BackspaceStmt
    !> The position specification in the statement.
    TYPE(PositionSpecList) :: position_specs
  CONTAINS
    PROCEDURE :: Dump => backspace_Dump
    PROCEDURE :: Query => backspace_Query
    PROCEDURE :: Modify => backspace_Modify
    PROCEDURE, NOPASS :: What => backspace_What
  END TYPE BackspaceStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /backspace-stmt/ (R924).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    backspace_stmt    The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseBackspaceStmt(tlist, backspace_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(BackspaceStmt), INTENT(OUT) :: backspace_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /backspace-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R924'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing parenthesis.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /backspace-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! BACKSPACE...
    
    CALL ExpectKeyword(tlist, 1, scck_'BACKSPACE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (SIZE(tlist) == 1) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(SIZE(tlist))),  &
          MSG='Expected /file-unit-number/ or "(" /position-spec-list/ ")".' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    IF (TestPositionStmtFirstForm(tlist(2:))) THEN
      CALL ParseExpr(  &
          tlist,  &
          2,  SIZE(tlist),  &
          backspace_stmt%position_specs%unit,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ELSE
      ! Failure not tested by StmtParsingTests because a missing 
      ! parenthesis causes the first form to be selected.
      CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ExpectClosing(tlist, 2, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ParsePositionSpecList(  &
          tlist,  &
          3, closing - 1,  &
          backspace_stmt%position_specs,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      
      CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    END IF
    
  END SUBROUTINE ParseBackspaceStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of BackspaceStmt%Dump - write a debugging representation 
  !! of the results of parsing an backspace statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE backspace_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BackspaceStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A)", ADVANCE='NO') pfx
    CALL pnode%position_specs%Write(unit)
    WRITE (unit, "()")
    
  END SUBROUTINE backspace_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of BackspaceStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE backspace_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BackspaceStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%position_specs%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE backspace_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of BackspaceStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE backspace_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BackspaceStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%position_specs%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE backspace_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of BackspaceStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'backspace-stmt'.
  
  FUNCTION backspace_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'backspace-stmt'
    
  END FUNCTION backspace_What
  
END MODULE BackspaceStmts
