! $Id: ComplexLiteralConstants.f90 2798 2019-03-22 17:18:41Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ComplexLiteralConstants module.


!*******************************************************************************
!!
!> Procedures for parsing complex literal constants.

MODULE ComplexLiteralConstants
  
  USE ParseNodes
  USE Tokens
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: IsComplexLiteralConstant
  PUBLIC :: ParseComplexLiteralConstant
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a complex literal constant.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: ComplexLiteralConstant
    !> Sign of the /real-part/.
    !!
    !! If not allocated then the /real-part/ had no sign.
    TYPE(Token), ALLOCATABLE :: real_sign
    !> "Meat" of the /real-part/.
    TYPE(Token) :: real_part
    !> Sign of the /imag-part/.
    !!
    !! If not allocated then the /imag-part/ had no sign.
    TYPE(Token), ALLOCATABLE :: imag_sign
    !> "Meat" of the /imag-part/.
    TYPE(Token) :: imag_part
  CONTAINS
    PROCEDURE, NOPASS :: What => cmplx_What
    PROCEDURE :: Write => cmplx_Write
    PROCEDURE :: Query => cmplx_Query
    PROCEDURE :: Modify => cmplx_Modify
    PROCEDURE :: Advertise => cmplx_Advertise
    PROCEDURE :: GetLocation => cmplx_GetLocation
  END TYPE ComplexLiteralConstant
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Test whether a token sequence is likely to be a 
  !! /complex-literal-constant/.
  !!
  !! @param[in]     tlist             The sequence of tokens, from the 
  !! token immediately after the opening parenthesis, to the token immediately 
  !! before the closing parenthesis.
  !!
  !! @returns True if the token sequence looks like a complex literal 
  !! constant, false otherwise.
  
  FUNCTION IsComplexLiteralConstant(tlist) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = FindBare(tlist, ittComma, scck_',') /= 0
    
  END FUNCTION IsComplexLiteralConstant
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /complex-literal-constant/.
  !!
  !! @param[in]     tlist             The sequence of tokens in the statement 
  !! with the complex literal constant to be parsed.
  !!
  !! @param[in]     start             The starting index of the constant in 
  !! \a tlist.  This should be the token after the opening parenthesis.
  !!
  !! @param[in]     finish            The finishing index of the constant in 
  !! \a tlist.  This should be the token immediately before the closing 
  !! parenthesis.
  !!
  !! @param[out]    constant          The results of parsing the constant.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseComplexLiteralConstant( tlist, start, finish,  &
      constant, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(ComplexLiteralConstant), INTENT(OUT) :: constant
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Syntax rule for a /complex-literal-constant/.
    CHARACTER(*), PARAMETER :: comp = 'R417'
    
    INTEGER :: it             ! Token index.
    TYPE(Error) :: err        ! Expect error.
    INTEGER :: opt            ! Expect option.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /complex-literal-constant/.' )
      RETURN
    END IF
    
    it = start
    
    !---------------------------------------------------------------------------
    ! The real part
    
    CALL Expect(  &
        tlist(:finish), it,  &
        [ittIntegerLiteral, ittRealLiteral, ittOperator, ittOperator],  &
        [scck_' ', scck_' ', scck_'+', scck_'-'],  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF ((opt == 3) .OR. (opt == 4)) THEN
      ! Optional leading sign.
      constant%real_sign = tlist(it)
      it = it + 1
      CALL Expect(  &
          tlist(:finish), it,  &
          [ittIntegerLiteral, ittRealLiteral],  &
          [scck_' ', scck_' '],  &
          opt, err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    END IF
    
    constant%real_part = tlist(it)
    it = it + 1
    
    !---------------------------------------------------------------------------
    ! The separating comma
    
    CALL Expect(tlist(:finish), it, ittComma, scck_',', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    it = it + 1
    
    !---------------------------------------------------------------------------
    ! The imaginary part
    
    CALL Expect(  &
        tlist(:finish), it,  &
        [ittIntegerLiteral, ittRealLiteral, ittOperator, ittOperator],  &
        [scck_' ', scck_' ', scck_'+', scck_'-'],  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF ((opt == 3) .OR. (opt == 4)) THEN
      ! Optional leading sign.
      constant%imag_sign = tlist(it)
      it = it + 1
      CALL Expect(  &
          tlist(:finish), it,  &
          [ittIntegerLiteral, ittRealLiteral],  &
          [scck_' ', scck_' '],  &
          opt, err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    END IF
    
    constant%imag_part = tlist(it)
    it = it + 1
    
    !---------------------------------------------------------------------------
    ! The end!
    
    CALL ExpectEnd(tlist(:finish), it, comp, err_list)
    
  END SUBROUTINE ParseComplexLiteralConstant
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComplexLiteralConstant%What - get a text description 
  !! of the relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'complex-literal-constant'
  
  FUNCTION cmplx_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'complex-literal-constant'
    
  END FUNCTION cmplx_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComplexLiteralConstant%Write - write a debugging 
  !! representation of the literal to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  !!
  !! Unlike the parse procedure, this procedure includes the parentheses.
  
  SUBROUTINE cmplx_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComplexLiteralConstant), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('( ')", ADVANCE='NO')
    IF (ALLOCATED(pnode%real_sign)) THEN
      WRITE (unit, "(A,' ')", ADVANCE='NO') QueryValue(pnode%real_sign)
    END IF
    WRITE (unit, "(A,', ')", ADVANCE='NO') QueryValue(pnode%real_part)
    IF (ALLOCATED(pnode%imag_sign)) THEN
      WRITE (unit, "(A,' ')", ADVANCE='NO') QueryValue(pnode%imag_sign)
    END IF
    WRITE (unit, "(A,' )')", ADVANCE='NO') QueryValue(pnode%imag_part)
    
  END SUBROUTINE cmplx_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComplexLiteralConstant%GetLocation - get the location in 
  !! source of the start of the expression.
  !!
  !! @param[in]     pnode             The expression.
  !!
  !! @returns The source location for the start of the expression.
  !!
  !! We just forward the call to the first stored token.  This skips the 
  !! opening parenthesis, but I don't care, do you?
  !!
  !! @todo Fix.
  
  FUNCTION cmplx_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Charateristics
    
    CLASS(ComplexLiteralConstant), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%real_sign)) THEN
      loc = QueryLocation(pnode%real_sign)
    ELSE
      loc = QueryLocation(pnode%real_part)
    END IF
    
  END FUNCTION cmplx_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComplexLiteralConstant%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE cmplx_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComplexLiteralConstant), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cmplx_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComplexLiteralConstant%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE cmplx_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComplexLiteralConstant), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cmplx_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComplexLiteralConstant%Advertise - advertise 
  !! the identifiers referenced or defined by the expression.
  !!
  !! @param[in]     pnode             The expression to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE cmplx_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComplexLiteralConstant), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    CALL advertise_token(ident_list, pnode%real_part)
    CALL advertise_token(ident_list, pnode%imag_part)
    
  END SUBROUTINE cmplx_Advertise
  
  
  !*****************************************************************************
  !!
  !> Advertise a token that might appear as the real or imaginary part in 
  !! a /complex-literal/ or as some sort of other literal constant.
  !!
  !! @param[in,out] ident_list        The list of identifier natures.
  !!
  !! @param[in]     tok               The token to analyse.
  
  SUBROUTINE advertise_token(ident_list, tok)
    
    USE CharacterTypes
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IdentifierNature), INTENT(INOUT), ALLOCATABLE :: ident_list(:)
    TYPE(Token), INTENT(IN) :: tok
    
    !***************************************************************************
    
    IF (tok == ittName) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=QueryValue(tok),  &
            NATURES=[NatureRefDataObject],  &
            LOCATIONS=[QueryLocation(tok)] ) )
    ELSE IF (ANY(tok == [ittIntegerLiteral, ittRealLiteral])) THEN
      CALL advertise_kind(ident_list, tok)
    END IF
    
  END SUBROUTINE advertise_token
  
  
  !*****************************************************************************
  !!
  !> Advertise a kind specification in a token that some sort of literal 
  !! constant.
  !!
  !! @param[in,out] ident_list        The list of identifier natures.
  !!
  !! @param[in]     tok               The token to process.
  
  SUBROUTINE advertise_kind(ident_list, tok)
    
    USE CharacterTypes
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IdentifierNature), INTENT(INOUT), ALLOCATABLE :: ident_list(:)
    TYPE(Token), INTENT(IN) :: tok
    
    !---------------------------------------------------------------------------
    ! Locals.
    
    ! Kind specifier on a value
    CHARACTER(:), ALLOCATABLE :: kind_text
    
    !***************************************************************************
    
    ! Perhaps the literal has a kind specifier that is a named constant.  
    IF (HasKind(tok)) THEN
      kind_text = QueryKind(tok)
      ! Such a constant will start with a letter.
      IF (IsLetterChar(kind_text(1:1))) THEN
        ! Kind specification would appear to be a named constant.  The 
        ! location will point to the start of the token, rather than the 
        ! kind text proper, but so be it.
        CALL MergeNatures(  &
            ident_list,  &
            IdentifierNature(  &
              NAME=kind_text,  &
              NATURES=[NatureRefDataObject],  &
              LOCATIONS=[QueryLocation(tok)] ) )
      END IF
    END IF
    
  END SUBROUTINE advertise_kind
  
END MODULE ComplexLiteralConstants
