! $Id: DimensionStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the DimensionStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /dimension-stmt/'s.

MODULE DimensionStmts
  
  USE CompilerKinds
  USE identifiers
  USE ParseNodes
  USE SubscriptLists
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseDimensionStmt
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a single declaration in a dimension stmt.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: DimensionDecl
    !> The name of the object being declared.
    TYPE(Identifier) :: name
    !> The /array-spec/.
    !!
    !! Always allocated, always size greater than zero.
    TYPE(DimSpec), ALLOCATABLE :: array_spec(:)
  CONTAINS
    PROCEDURE :: Write => decl_Write
    PROCEDURE :: Query => decl_Query
    PROCEDURE :: Modify => decl_Modify
    PROCEDURE, NOPASS :: What => decl_What
    PROCEDURE :: Advertise => decl_Advertise
    PROCEDURE :: GetLocation => decl_GetLocation
  END TYPE DimensionDecl
  
  !> Stores the results of parsing a /dimension-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: DimensionStmt
    !> The declarations in the statement.
    TYPE(DimensionDecl), ALLOCATABLE :: decls(:)
  CONTAINS
    PROCEDURE :: Dump => dim_Dump
    PROCEDURE :: Query => dim_Query
    PROCEDURE :: Modify => dim_Modify
    PROCEDURE, NOPASS :: What => dim_What
    ! Nothing to advertise.
  END TYPE DimensionStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /dimension-stmt/ (R545).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    dim_stmt          The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseDimensionStmt(tlist, dim_stmt, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DimensionStmt), INTENT(OUT) :: dim_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /dimension-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R545'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: list_count     ! Number of items in the declaration list.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /dimension-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'DIMENSION', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! option :: before name list.
    
    CALL Expect(  &
        tlist, 2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  & 
        '"::" or /array-name/',  &
        opt,  &
        err )   
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    
    !---------------------------------------------------------------------------
    ! The list of declarations.
    
    list_count = ListItemCount(tlist(it:))
    ALLOCATE(dim_stmt%decls(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, SIZE(tlist), item_start, item_finish)
      CALL parse_dimension_decl(  &
          tlist,  &
          item_start,  &
          item_finish,  &
          dim_stmt%decls(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! That's it!
  
  END SUBROUTINE ParseDimensionStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a single declaration in a
  !! /dimension-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! that is the start of the declaration.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! that is the end of the declaration.
  !!
  !! @param[out]    decl              Results of parsing the declaration.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_dimension_decl(tlist, start, finish, decl, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(DimensionDecl), INTENT(OUT) :: decl
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err          ! Error object for Expect calls.
    INTEGER :: closing          ! Index of closing parenthesis.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ExpectName(tlist, start, '/array-name/', err)
    CALL Add(err_list, err)
    IF (Failure(err)) RETURN
    
    decl%name = IdentifierIfortBug(tlist(start))
    
    !---------------------------------------------------------------------------
    ! ( array-spec )
    
    CALL Expect(tlist, start + 1, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, start + 1, closing, err)
    CALL Add(err_list, err)
    IF (Failure(err)) RETURN
    
    CALL ParseArraySpec(  &
        tlist(start+2:closing-1),  &
        .FALSE.,  &
        decl%array_spec,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    IF (Failure(sub_err_list)) RETURN
    
    ! That's all folks.
    IF (finish >= closing + 1) THEN
      CALL Add( err_list,  &
          CODE=errUnexpectedExtraTokens,  &
          LOCATION=QueryLocation(tlist(closing+1)),  &
          MSG='Unexpected extra tokens following ")".' )
    END IF
    
  END SUBROUTINE parse_dimension_decl
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in DimensionStmt.
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimensionStmt%Dump - write a debugging representation 
  !! of the results of parsing a dimension statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE dim_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimensionStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Declaration list index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(pnode%decls)
      WRITE (unit, "(A)", ADVANCE='NO') pfx
      CALL pnode%decls(i)%Write(unit)
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE dim_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimensionStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE dim_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimensionStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Declaration index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dim_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimensionStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE dim_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimensionStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Declaration index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dim_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimensionStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'dimension-stmt'.
  
  FUNCTION dim_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'dimension-stmt'
    
  END FUNCTION dim_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in DimensionDecl.
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimensionDecl%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !1 @param[in]     pnode             The node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE decl_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimensionDecl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('/array-name/ = ',A,' (')",  ADVANCE='NO') &
          .NameOf. pnode%name
    CALL WriteDimSpec(unit, pnode%array_spec)
    WRITE (unit, "(')')", ADVANCE='NO')
    
  END SUBROUTINE decl_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimensionDecl%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE decl_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimensionDecl), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Array spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%array_spec)
      CALL pnode%array_spec(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE decl_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimensionDecl%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE decl_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimensionDecl), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Array spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%array_spec)
      CALL pnode%array_spec(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE decl_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimensionDecl%Advertise - advertise the 
  !! identifiers in a DimensionDecl that are not child nodes.
  !!
  !! @param[in]     pnode             The statement to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE decl_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimensionDecl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=[NatureDefDimension],  &
        LOCATIONS=[.LocationOf. pnode%name] )
    
  END SUBROUTINE decl_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimensionDecl%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns '(dimension-decl)'.
  
  FUNCTION decl_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = '(dimension-decl)'
    
  END FUNCTION decl_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimensionDecl%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION decl_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(DimensionDecl), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION decl_GetLocation
  
END MODULE DimensionStmts
