! $Id: EntryStmts.f90 2929 2019-04-27 22:38:46Z ian $
!> @file
!! Defines the EntryStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /entry-stmt/'s.
!!
!! Dummy arg list processing is provided by a procedure in the 
!! SubroutineStmts module.

MODULE EntryStmts
  
  USE Identifiers
  USE LanguageBindingSpecs
  USE ParseNodes
  USE SubroutineStmts
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseEntryStmt
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing an /entry-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: EntryStmt
    !> The /entry-name/.
    TYPE(Identifier) :: entry_name
    !> Dummy arguments.
    !!
    !! Always allocated but may be zero length.
    TYPE(DummyArg), ALLOCATABLE :: args(:)
    !> Function result name.  May not be allocated.
    TYPE(Identifier), ALLOCATABLE :: result_name
    !> Language binding spec.  May not be allocated.
    TYPE(LangSpec), ALLOCATABLE :: ls
  CONTAINS
    PROCEDURE :: Dump => entry_Dump
    PROCEDURE :: Query => entry_Query
    PROCEDURE :: Modify => entry_Modify
    PROCEDURE, NOPASS :: What => entry_What
    PROCEDURE :: Advertise => entry_Advertise
  END TYPE EntryStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /entry-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    entry_stmt        The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEntryStmt(tlist, entry_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE FunctionStmts
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(EntryStmt), INTENT(OUT) :: entry_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /entry-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1240'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err      ! Expect error object.
    INTEGER :: opt          ! Expect option.
    INTEGER :: it           ! Current token index
    INTEGER :: it_end       
    INTEGER :: closing      ! Location of closing parenthesis.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /entry-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'ENTRY', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! /entry-name/
    
    CALL ExpectName(tlist, 2, '/entry-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    entry_stmt%entry_name = IdentifierIfortBug(tlist(2))
    
    !---------------------------------------------------------------------------
    ! [ ( [ /dummy-arg-list/ ] ) ]
    
    ! Note that unlike the function-stmt the parentheses are optional.
    
    CALL Expect(  &
        tlist,  &
        3,  &
        [ittEndStatement, ittDelimiter],  &
        [scck_' ', scck_'('],  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! No parentheses means we are done (means we cannot have a 
    ! /proc-language-binding-spec/ either.)
    IF (opt == 1) RETURN
        
    CALL ExpectClosing(tlist, 3, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! The argument list is optional.
    IF (closing - 1 > 3) THEN
      CALL ParseDummyArgList(  &
          tlist,  &
          4,  &
          closing - 1,  &
          entry_stmt%args,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(err)) RETURN
    ELSE
      ALLOCATE(entry_stmt%args(0))
    END IF
    it = closing + 1
    
    !---------------------------------------------------------------------------
    ! suffix
    
    CALL Expect(  &
        tlist,  &
        it,  &
        [ittEndStatement, ittName, ittName],  &
        [scck_'      ', scck_'RESULT', scck_'BIND  '],  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) RETURN
    
    ! The suffix for an entry statement has the same possibilities as a 
    ! function statement.
    CALL ParseSuffix(  &
        tlist,  &
        it,  &
        entry_stmt%result_name,  &
        entry_stmt%ls,  &
        it_end,  &
        sub_err_list )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    CALL ExpectEOS(tlist, it_end, comp, err_list)
    
  END SUBROUTINE ParseEntryStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of EntryStmt%Dump - write a debugging representation 
  !! of the results of parsing an entry statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE entry_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EntryStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Argument list index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/entry-name/ = ',A)") pfx,  &
        .NameOf. pnode%entry_name
    WRITE (unit, "(A,'/dummy-arg-list/')") pfx
    DO i = 1, SIZE(pnode%args)
      IF (pnode%args(i)%is_star) THEN
        WRITE (unit, "(A,'*')") pfx // '  '
      ELSE
        WRITE (unit, "(A,A)") pfx // '  ', .NameOf. pnode%args(i)%name
      END IF
    END DO
    
  END SUBROUTINE entry_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of EntryStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE entry_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EntryStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%ls)) CALL pnode%ls%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE entry_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of EntryStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in]     pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE entry_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EntryStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%ls)) CALL pnode%ls%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE entry_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of EntryStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'write-stmt'.
  
  FUNCTION entry_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'entry-stmt'
    
  END FUNCTION entry_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of EntryStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The statement to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE entry_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EntryStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! arg index.
    INTEGER :: result         ! Identifiers in the result name.
    
    !***************************************************************************
    
    ! @todo identifiers in any decl spec.
    ! @todo identifiers in any binding label.
    
    IF (ALLOCATED(pnode%result_name)) THEN
      result = 1
    ELSE
      result = 0
    END IF
    
    ALLOCATE(ident_list(1 + SIZE(pnode%args) + result))
    ident_list(1) = IdentifierNature(  &
        .NameOf. pnode%entry_name,  &
        [NatureDefEntry],  &
        [.LocationOf. pnode%entry_name] )
    
    ! Identifiers in the arguments.
    DO i = 1, SIZE(pnode%args)
      ident_list(i+1) = IdentifierNature(  &
          .NameOf. pnode%args(i)%name,  &
          [NatureDefDummy],  &
          [.LocationOf. pnode%args(i)%name] )
    END DO
    
    IF (ALLOCATED(pnode%result_name)) THEN
      ident_list(i+1) = IdentifierNature(  &
          .NameOf. pnode%result_name,  &
          [NatureDefResult],  &
          [.Locationof. pnode%result_name] )
    END IF
    
  END SUBROUTINE entry_Advertise
  
END MODULE EntryStmts
