! $Id: EquivalenceStmts.f90 2189 2016-06-06 10:42:45Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the EquivalenceStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /equivalence-stmt/'s (R565).

MODULE EquivalenceStmts
  
  USE CompilerKinds
  USE Designators
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseEquivalenceStmt
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing a single /equivalence-set/.
  TYPE, PUBLIC :: EquivalenceSet
    !> The objects in the set.  Always of size two or greater.
    !!
    !! A designator is the general case - what is valid as an equivalence 
    !! object is more restricted.
    TYPE(Designator), ALLOCATABLE :: object_list(:)
  END TYPE EquivalenceSet
  
  !> The results of parsing an equivalence statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: EquivalenceStmt
    !> The equivalence sets in the statement.
    TYPE(EquivalenceSet), ALLOCATABLE :: sets(:)
  CONTAINS
    PROCEDURE :: Dump => equiv_Dump
    PROCEDURE :: Query => equiv_Query
    PROCEDURE :: Modify => equiv_Modify
    PROCEDURE, NOPASS :: What => equiv_What
  END TYPE EquivalenceStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /equivalence-stmt/ (R565).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    equiv_stmt        The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEquivalenceStmt(tlist, equiv_stmt, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(EquivalenceStmt), INTENT(OUT) :: equiv_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /equivalence-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R565'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: list_count     ! Number of items in the /target-decl-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /equivalence-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! EQUIVALENCE...
    
    CALL ExpectKeyword(tlist, 1, scck_'EQUIVALENCE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! /equivalence-set-list/
    
    list_count = ListItemCount(tlist(2:))
    ALLOCATE(equiv_stmt%sets(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 2, SIZE(tlist), item_start, item_finish)
      CALL parse_equivalence_set( tlist, item_start, item_finish,  &
          equiv_stmt%sets(item), sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseEquivalenceStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /equivalence-set/ (R566).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the start of the 
  !! /equivalence-set/ in @a tlist.
  !!
  !! @param[in]     start             Index of the end of the 
  !! /equivalence-set/ in @a tlist.
  !!
  !! @param[out]    set               The results of parsing the 
  !! /equivalence-set/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_equivalence_set(tlist, start, finish, set, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(EquivalenceSet), INTENT(OUT) :: set
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /equivalence-set/.
    CHARACTER(*), PARAMETER :: comp = 'R566'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: list_count     ! Number of items in the /target-decl-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL Expect(tlist, start, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, start, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist(start+1:closing-1))
    IF (list_count < 2) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
!          LOCATION=tlist(closing-1)%location,  
          MSG='There must be at least two objects in an /equivalence-set/.' )
      ! Carry on.
    END IF
    
    ALLOCATE(set%object_list(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem( tlist, start + 1, closing - 1, item_start,  &
          item_finish )
      CALL parse_equivalence_object( tlist, item_start, item_finish,  &
          set%object_list(item), sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! Check nothing after the closing parenthesis.
    CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
    
  END SUBROUTINE parse_equivalence_set
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /equivalence-object/ (R567).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the start of the 
  !! /equivalence-object/ in @a tlist.
  !!
  !! @param[in]     start             Index of the end of the 
  !! /equivalence-object/ in @a tlist.
  !!
  !! @param[out]    object            The results of parsing the 
  !! /equivalence-object/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! An /equivalence-object/ can be:
  !! - a /variable-name/
  !! - an /array-element/
  !! - a substring.
  !!
  !! These are all objects for a /designator/, so for convenience we use 
  !! that to store the results of parsing.  Later processing needs to filter 
  !! out the options that are not valid.
  
  SUBROUTINE parse_equivalence_object(tlist, start, finish, object, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(Designator), INTENT(OUT) :: object
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /equivalence-object/.
    CHARACTER(*), PARAMETER :: comp = 'R567'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ParseDesignator(tlist(start:finish), object, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE parse_equivalence_object
  
  
  !*****************************************************************************
  !!
  !> Implementation of EquivalenceStmt%Dump - write a debugging representation 
  !! of the results of parsing an equivalence statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE equiv_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EquivalenceStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: is             ! Set index.
    INTEGER :: io             ! Object index.
    
    !***************************************************************************
    
    DO is = 1, SIZE(pnode%sets)
      WRITE (unit, "(A,'equivalence set')") pfx
      DO io = 1, SIZE(pnode%sets(is)%object_list)
        WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
        CALL WriteDesignator(unit, pnode%sets(is)%object_list(io))
        WRITE (unit, "()")
      END DO
    END DO
    
  END SUBROUTINE equiv_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of EquivalenceStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE equiv_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EquivalenceStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: is             ! Equivalence set index.
    INTEGER :: io             ! Object index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO is = 1, SIZE(pnode%sets)
      DO io = 1, SIZE(pnode%sets(is)%object_list)
        CALL pnode%sets(is)%object_list(io)%Query(visitor)
      END DO
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE equiv_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of EquivalenceStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE equiv_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EquivalenceStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: is             ! Equivalence set index.
    INTEGER :: io             ! Object index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO is = 1, SIZE(pnode%sets)
      DO io = 1, SIZE(pnode%sets(is)%object_list)
        CALL pnode%sets(is)%object_list(io)%Modify(visitor)
      END DO
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE equiv_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of EquivalenceStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'equivalence-stmt'.
  
  FUNCTION equiv_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'equivalence-stmt'
    
  END FUNCTION equiv_What
  
END MODULE EquivalenceStmts
