! $Id: FinalProcedureStmts.f90 2161 2016-06-01 23:18:33Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the FinalProcedureStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /final-procedure-stmt/'s.

MODULE FinalProcedureStmts
  
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseFinalProcedureStmt
  
  !> Stores the results of parsing a final procedure statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: FinalProcedureStmt
    !> The list of final subroutine names.
    TYPE(Identifier), ALLOCATABLE :: finals(:)
  CONTAINS
    PROCEDURE :: Dump => fin_Dump
    PROCEDURE :: Query => fin_Query
    PROCEDURE :: Modify => fin_Modify
    PROCEDURE, NOPASS :: What => fin_What
    PROCEDURE :: Advertise => fin_Advertise
  END TYPE FinalProcedureStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /final-procedure-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    fin_stmt          The results of processing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseFinalProcedureStmt(tlist, fin_stmt, err_list)
    
    USE Strings
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(FinalProcedureStmt), INTENT(OUT) :: fin_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /final-procedure-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R452'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err            ! Error object for Expect calls.
    INTEGER :: opt                ! Expect option.
    INTEGER :: it                 ! Current token index.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /final-procedure-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! FINAL 
    
    CALL ExpectKeyword(tlist, 1, scck_'FINAL', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional ::, or w can straight into the name list.
    
    CALL Expect(  &
        tlist, 2,  &
        [ittName, ittDeclare],  &
        [scck_'  ', scck_'::'],  &
        '/final-subroutine-name/ or "::"',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Set it to be the start of the /final-subroutine-name-list/.
    IF (opt == 2) THEN
      it = 3
    ELSE
      it = 2
    END IF
    
    !---------------------------------------------------------------------------
    ! /final-subroutine-name-list/
    
    CALL ParseNameList(  &
        tlist,  &
        it,  &
        SIZE(tlist),  &
        '/final-subroutine-name/',  &
        fin_stmt%finals,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    ! Done.
    
  END SUBROUTINE ParseFinalProcedureStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of FinalProcedureStmt%Dump - write a debugging 
  !! representation of the results of parsing the statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE fin_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FinalProcedureStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Binding list index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/final-subroutine-name-list/')") pfx
    DO i = 1, SIZE(pnode%finals)
      WRITE (unit, "(A,A)") pfx // '  ', .NameOf. pnode%finals(i)
    END DO
    
  END SUBROUTINE fin_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of FinalProcedureStmt%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE fin_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FinalProcedureStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE fin_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of FinalProcedureStmt%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE fin_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FinalProcedureStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE fin_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of FinalProcedureStmt%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'final-procedure-stmt'.
  
  FUNCTION fin_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'final-procedure-stmt'
    
  END FUNCTION fin_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of FinalProcedureStmt%Advertise - advertise the 
  !! identifiers referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE fin_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FinalProcedureStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Specific binding index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(SIZE(pnode%finals)))
    DO i = 1, SIZE(pnode%finals)
      ident_list(i) = IdentifierNature(  &
          NAME=.NameOf. pnode%finals(i),  &
          NATURES=[NatureRefSpecBinding],  &
          LOCATIONS=[.Locationof. pnode%finals(i)] )
    END DO
    
  END SUBROUTINE fin_Advertise
  
END MODULE FinalProcedureStmts
