! $Id: Formats.f90 2800 2019-03-22 18:13:05Z ian $
!> @file
!! Defines the Formats module.


!*******************************************************************************
!!
!> Types and procedures for handling /format/'s.

MODULE Formats
  
  USE Expressions
  USE Labels
  USE ParseNodes
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseFormat
  
  !-----------------------------------------------------------------------------
  
  !> Describes a format specification.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: Format
    !> If true, then the format spec was '*'.
    LOGICAL :: is_star
    !> If is_star is true, the location of the star.
    !!
    !! Otherwise not allocated.
    TYPE(SourceLocation), ALLOCATABLE :: star_location
    !> The format spec, if it was a label.
    !!
    !! Not allocated if the format spec was not a label.
    TYPE(Label), ALLOCATABLE :: label_spec
    !> Format spec was a character expression.
    !!
    !! Not allocated if @a is_star or @a label is true.
    CLASS(Expression), ALLOCATABLE :: expr
  CONTAINS
    PROCEDURE :: Write => format_Write
    PROCEDURE :: Query => format_Query
    PROCEDURE :: Modify => format_Modify
    PROCEDURE, NOPASS :: What => format_What
    PROCEDURE :: GetLocation => format_GetLocation
  END TYPE Format
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse sequence of tokens that is a /format/ (R915).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist of 
  !! the /format/.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist of 
  !! the /format/.
  !!
  !! @param[out]    fmt_spec          Results of parsing the /format/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! If a /format/ is an expression then it must be of type character.  For 
  !! some trivial cases (the expression is a real literal constant, for 
  !! example) we could test that here, but we defer all testing to a later 
  !! pass.
  
  SUBROUTINE ParseFormat(tlist, start, finish, fmt_spec, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Scanner
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(Format), INTENT(OUT) :: fmt_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /format/.
    CHARACTER(*), PARAMETER:: comp = 'R915'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /format/.')
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    IF (IsToken(tlist(start), ittOperator, scck_'*')) THEN
      CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
      fmt_spec%is_star = .TRUE.
      fmt_spec%star_location = QueryLocation(tlist(start))
      RETURN
    END IF
    
    IF ( (tlist(start) == ittIntegerLiteral)  &
        .AND. (.NOT. HasKind(tlist(start)))  &
        .AND. (finish == start) ) THEN
      ALLOCATE(fmt_spec%label_spec)
      CALL ParseLabel(  &
          tlist,  &
          start,  &
          finish,  &
          fmt_spec%label_spec,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      RETURN
    END IF
    
    ! We must have an expression
    CALL ParseExpr(tlist, start, finish, fmt_spec%expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseFormat
  
  
  !*****************************************************************************
  !!
  !> Implementation of Format%Write - write a text representation of a 
  !! Format to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE format_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Format), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    IF (pnode%is_star) THEN
      WRITE (unit, "('*')", ADVANCE='NO')
    ELSE IF (ALLOCATED(pnode%label_spec)) THEN
      WRITE (unit, "(I0)", ADVANCE='NO') .ValueOf. pnode%label_spec
    ELSE
      CALL pnode%expr%Write(unit)
    END IF
    
  END SUBROUTINE format_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of Format%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE format_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Format), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%expr)) CALL pnode%expr%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE format_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of Format%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE format_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Format), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%expr)) CALL pnode%expr%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE format_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of Format%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION format_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'format'
    
  END FUNCTION format_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of Format%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION format_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Format), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (pnode%is_star) THEN
      loc = pnode%star_location
    ELSE IF (ALLOCATED(pnode%label_spec)) THEN
      loc = .LocationOf. pnode%label_spec
    ELSE
      loc = .LocationOf. pnode%expr
    END IF
    
  END FUNCTION format_GetLocation
  
END MODULE Formats
