! $Id: Identifiers.f90 2922 2019-04-26 22:40:09Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the Identifiers module.


!*******************************************************************************
!!
!> Defines a type that can act as common storage for identifier like things 
!! in parse nodes.
!!
!! Previously we used to just store the identifier name directly in each 
!! parse node.  But this resulted in too much information loss for later 
!! passes.

MODULE Identifiers
  
  USE IdentifierUtils
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: Identifier              ! From IdentifierUtils
  PUBLIC :: IdentifierIfortBug      ! From IdentifierUtils
  PUBLIC :: ParseNameList
  
  !> Parse a list of comma separated names.
  !!
  !! This is a generic interface as a transition from the previous string 
  !! based name list, which is in the ParseUtils module.
  INTERFACE ParseNameList
    MODULE PROCEDURE ParseNameList_
  END INTERFACE ParseNameList
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a comma separated list of names.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     list_start        Start index of the list in @a tlist.
  !!
  !! @param[in]     list_finish       End index of the list in @a tlist.
  !!
  !! @param[in]     syntax_rule       The syntax term for each name, used for 
  !! error reporting.  This is used as the @a msg parameter in a child call 
  !! to Expect_name.  Include the trailing and leading slashes.
  !!
  !! @param[out]    name_list         List of names extracted from @a tlist.
  !!
  !! @param[out]    err_list          List of errors.  No specific component 
  !! is assigned.
  !!
  !! The syntax of a comma separated list is defined by syntax rule R101 
  !! for a /list/, but we don't explicitly refer to that in any 
  !! generated error messages.
  
  SUBROUTINE ParseNameList_( tlist, list_start, list_finish, syntax_rule,  &
      name_list, err_list )
    
    USE Strings
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: list_start
    INTEGER, INTENT(IN) :: list_finish
    CHARACTER(*), INTENT(IN) :: syntax_rule
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: name_list(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: item_count     ! Number of items in the list.
    INTEGER :: iname          ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    TYPE(Error) :: err        ! Error code for procedure calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    item_count = ListItemCount(tlist(list_start:list_finish))
    ALLOCATE(name_list(item_count))
    item_finish = 0
    DO iname = 1, item_count
      CALL GetNextListItem(  &
          tlist,  &
          list_start,  &
          list_finish,  &
          item_start,  &
          item_finish )
      CALL ExpectName(tlist, item_start, syntax_rule, err)
      CALL Add(err_list, err)
      IF (Failure(err)) RETURN
      ! This assignment may also do kind conversion from scck to default char.
      name_list(iname) = IdentifierIfortBug(tlist(item_start))
      IF (item_finish > item_start) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            LOCATION=QueryLocation(tlist(list_start+1)),  &
            MSG='Unexpected token "' // QueryRaw(tlist(item_start+1)) //'".' )
        RETURN
      END IF
    END DO
    
  END SUBROUTINE ParseNameList_
  
END MODULE Identifiers
