! $Id: IfConstructs.f90 2878 2019-04-05 22:15:49Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the IfConstructs module.


!*******************************************************************************
!!
!> Procedures and types for parsing /if-construct/'s.

MODULE IfConstructs
  
  USE Expressions
  USE Identifiers
  USE Labels
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseIfConstruct
  PUBLIC :: ParseIfStmtToConstruct
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing a particular "block" (as it is termed here) 
  !! of an if-construct.
  !!
  !! @todo Extend from BlockStorage?
  TYPE :: sub_block
    !> The expression in the opening statement of the block.
    CLASS(Expression), ALLOCATABLE :: ex
    !> The block of executable statements following the initial statement.
    !!
    !! Always allocated, but maybe zero size.
    TYPE(SCParseNodeElement), ALLOCATABLE :: children(:)
  END TYPE sub_block
  
  !> The results of parsing an if-construct.
  !!
  !! Layout is similar to that for the corresponding statement tree node.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: IfConstruct
    !> The construct name, if present.
    TYPE(Identifier), ALLOCATABLE :: construct_name
    !> The if or else if blocks.
    !!
    !! Always allocated and size greater than zero.
    TYPE(sub_block), ALLOCATABLE :: blocks(:)
    !> The children (executable only) of the else part.
    !!
    !! Not allocated if there was no else part.  Zero sized if there 
    !! was an else part, but that then had no child statements.
    TYPE(SCParseNodeElement), ALLOCATABLE :: else_children(:)
    !> The label on the end if statement, if present, not allocated otherwise.
    TYPE(Label), ALLOCATABLE :: end_label
  CONTAINS
    PROCEDURE :: Dump => ifc_Dump
    PROCEDURE, NOPASS :: What => ifc_What
    PROCEDURE :: Query => ifc_Query
    PROCEDURE :: Modify => ifc_Modify
    PROCEDURE :: Advertise => ifc_Advertise
  END TYPE IfConstruct
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse an /if-construct/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! construct.
  !!
  !! @param[out]    construct         The results of parsing the construct.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseIfConstruct(st, construct, err_list)
    
    USE IfConstructStNodes
    USE IfStmts
    USE Errors
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IfConstructStNode), INTENT(INOUT) :: st
    TYPE(IfConstruct), INTENT(OUT) :: construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: ib             ! Block index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! There is a one to one mapping of statement tree blocks across to 
    ! parsed result blocks.
    ALLOCATE(construct%blocks(SIZE(st%blocks)))
    
    !---------------------------------------------------------------------------
    ! The initial /if-then-stmt/.
    
    CALL ParseIfThenStmt(  &
        st%blocks(1)%initial%tlist,  &
        construct%blocks(1)%ex,  &
        construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    ! Executable children of that initial part.
    CALL ParseBlock(  &
        st%blocks(1)%exec,  &
        construct%blocks(1)%children,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Subsequent /else-if-stmt/'s.
    
    DO ib = 2, SIZE(st%blocks)
      CALL ParseElseIfStmt(  &
          st%blocks(ib)%initial%tlist,  &
          construct%construct_name,  &
          construct%blocks(ib)%ex, sub_err_list )
      CALL Add(err_list, sub_err_list)
      
      ! Executable children of that subsequent part.
      CALL ParseBlock(  &
          st%blocks(ib)%exec,  &
          construct%blocks(ib)%children,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
    END DO
    
    !---------------------------------------------------------------------------
    ! The else block.
    
    IF (ALLOCATED(st%else_block)) THEN
      CALL ParseElseStmt(  &
          st%else_block%initial%tlist,  &
          construct%construct_name,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      
      CALL ParseBlock(  &
          st%else_block%exec,  &
          construct%else_children,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
    END IF
    
    !---------------------------------------------------------------------------
    ! The end-if statement.
    
    CALL ParseEndIfStmt(st%last%tlist, construct%construct_name, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    ! Save the statement label of the last statement, if any, in case 
    ! it is referenced by other statements (which is valid if those 
    ! statements are within the executable children of the construct).
    IF (ALLOCATED(st%last%statement_label)) THEN
      construct%end_label = st%last%statement_label
    END IF
    
END SUBROUTINE ParseIfConstruct
  
  
  !*****************************************************************************
  !!
  !> Implementation of IfConstruct%Dump - write a debugging 
  !! representation of the ParseNode to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE ifc_Dump(pnode, unit, pfx)
    
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IfConstruct), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: ib             ! Child block index.
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%construct_name))  &
        WRITE (unit, "(A,'construct_name = ',A)") pfx,  &
          .NameOf. pnode%construct_name
    
    WRITE (unit, "(A,'initial expression = ')", ADVANCE='NO') pfx
    CALL WriteExpr(unit, pnode%blocks(1)%ex)
    WRITE (unit, "()")
    
    WRITE (unit, "(A,'initial child statements:')") pfx
    CALL DumpBlock(pnode%blocks(1)%children, unit, pfx // '  ')
    
    DO ib = 2, SIZE(pnode%blocks)
      WRITE (unit, "(A,'else-if expression:')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%blocks(ib)%ex)
      WRITE (unit, "()")
      
      WRITE (unit, "(A,'else-if child statements:')") pfx
      CALL DumpBlock(pnode%blocks(ib)%children, unit, pfx // '  ')
    END DO
    
    IF (ALLOCATED(pnode%else_children)) THEN
      WRITE (unit, "(A,'else child statements:')") pfx
      CALL DumpBlock(pnode%else_children, unit, pfx // '  ')
    END IF
    
  END SUBROUTINE ifc_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of IfConstruct%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'if-construct'.
  
  FUNCTION ifc_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'if-construct'
    
  END FUNCTION ifc_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of IfConstruct%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE ifc_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IfConstruct), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child block index.
    INTEGER :: j              ! Child block child statement index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%blocks)
      ! The expression being tested.
      CALL pnode%blocks(i)%ex%Query(visitor)
      ! The child statements of the block.
      DO j = 1, SIZE(pnode%blocks(i)%children)
        CALL pnode%blocks(i)%children(j)%Query(visitor)
      END DO
    END DO
    
    DO i = 1, SIZE(pnode%else_children)
      CALL pnode%else_children(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ifc_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of IfConstruct%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE ifc_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IfConstruct), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child block index.
    INTEGER :: j              ! Child block child statement index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%blocks)
      ! The expression being tested.
      CALL pnode%blocks(i)%ex%Modify(visitor)
      ! The child statements of the block.
      DO j = 1, SIZE(pnode%blocks(i)%children)
        CALL pnode%blocks(i)%children(j)%Modify(visitor)
      END DO
    END DO
    
    DO i = 1, SIZE(pnode%else_children)
      CALL pnode%else_children(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ifc_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of IfConstruct%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE ifc_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IfConstruct), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=.NameOf. pnode%construct_name,  &
            NATURES=[NatureDefConstruct], &
            LOCATIONS=[.LocationOf. pnode%construct_name] ) )
    END IF
    
  END SUBROUTINE ifc_Advertise
  
  
  !*****************************************************************************
  !!
  !> Parse an IfStmt through to an IfConstruct structure.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    if_construct      The results of parsing the /if-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseIfStmtToConstruct(tlist, if_construct, err_list)
    
    USE BuildTreePass
    USE Classifier
    USE Errors
    USE ErrorCodes
    USE ExecStNodes
    USE ExecutionParts
    USE IfStmts
    USE SingleStNodes
    USE Statements
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(INOUT) :: tlist(:)
    TYPE(IfConstruct), INTENT(OUT) :: if_construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /if-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R837'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Tokens that make up the child /action-stmt/ of the /if-stmt/.
    TYPE(Token), ALLOCATABLE :: child_tlist(:)
    
    ! Action statement tree node.
    TYPE(ExecStNode) :: exec_st
    
    INTEGER :: ist            ! Child statement number.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! Parse the leading if-stmt.
    
    ! We know that we will only have one block.
    ALLOCATE(if_construct%blocks(1))
    
    CALL ParseIfStmt(  &
        tlist,  &
        if_construct%blocks(1)%ex,  &
        child_tlist, sub_err_list )
    CALL Add(err_list, sub_err_list)
    ! No point continuing if we can't parse the front bit.
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Work out what the action statement is.
    
    CALL Classify(child_tlist, [INTEGER ::], ist)
    SELECT CASE (ist)
    CASE (istUnclassified)
      CALL Add( err_list,  &
          CODE=errUnclassifiedStatement,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(child_tlist(1)),  &
          MSG='Unclassifiable /action-stmt/.' )
      
    CASE ( istAllocate, istAssignment, istBackspace, istCall, istClose,  &
        istContinue, istCycle, istDeallocate, istEndfile, istErrorStop,  &
        istExit, istFlush, istForall, istGoto, istInquire,  &
        istLock, istNullify, istOpen, istPointerAssignment, istPrint,  &
        istRead, istReturn, istRewind, istStop, istSyncAll, istSyncImages,  &
        istSyncMemory, istUnlock, istWait, istWhere, istWrite,  &
        istArithmeticIf, istComputedGoto )
      ! Parse it.
      CALL MakeExecStNode(child_tlist, ist, exec_st)
      ! Failure not covered by ConstructParsingTests.
      CALL DoBuildTreePassCheck(exec_st, err_list)
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
      
      CALL ParseBlock(exec_st, if_construct%blocks(1)%children, sub_err_list)
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
      
    CASE DEFAULT
      ! This deals with statements that are not /action-stmt/s too.
      CALL Add( err_list,  &
          CODE=errIfBadChildStmt,  &
          COMPONENT='C828 via ' // comp,  &  
          LOCATION=QueryLocation(child_tlist(1)),  &
          MSG='A /' // GetStName(ist) // '/ is not a permitted &
            &/action-stmt/ for an /if-stmt/.' )
      RETURN
      
    END SELECT
    
  END SUBROUTINE ParseIfStmtToConstruct
  
END MODULE IfConstructs
