! $Id: IntentSpecs.f90 2925 2019-04-27 00:21:42Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the IntentSpecs module.


!*******************************************************************************
!!
!> Procedures for parsing /intent-spec/'s and some constants to describe the 
!! result of that parsing.
!!
!! An /intent-spec/ is defined by R523 as:
!! - IN
!! - OUT
!! - INOUT (which can also be IN OUT as per 3.3.2.2)

MODULE IntentSpecs
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseIntentSpec
  
  !-----------------------------------------------------------------------------
  
  !> The intent has not been specified for an entity.
  INTEGER, PARAMETER, PUBLIC :: iintNone = 0
  !> The intent for an entity has been specified as IN.
  INTEGER, PARAMETER, PUBLIC :: iintIn = 1
  !> The intent for an entity has been specified as OUT.
  INTEGER, PARAMETER, PUBLIC :: iintOut = 2
  !> The intent for an entity has been specified as INOUT.
  INTEGER, PARAMETER, PUBLIC :: iintInOut = 3
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an intent-spec/ (R523).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /intent-spec/ (the bit that is inside the parentheses.
  !!
  !! @param[out]    intent            The iint* constant that results from 
  !! the parsing.  Not defined on error.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseIntentSpec(tlist, intent, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(OUT) :: intent
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /intent-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R523'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /intent-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Parse the keywords that make up the spec.
    !
    ! We do this manually (rather than using test procedures) for simplicity.
    
    it = 1
    IF (Match(tlist, it, ittName, scck_'IN')) THEN
      it = it + 1
      ! IN OUT is permissible in free form source by 3.3.2.2.
      IF (Match(tlist, it, ittName, scck_'OUT')) THEN
        intent = iintInOut
        it = it + 1
      ELSE
        intent = iintIn
      END IF
    ELSE IF (Match(tlist, it, ittName, scck_'OUT')) THEN
      intent = iintOut
      it = it + 1
    ELSE IF (Match(tlist, it, ittName, scck_'INOUT')) THEN
      intent = iintInOut
      it = it + 1
    ELSE
      ! Empty length test at the start guarantees that SIZE(tlist) >= 1.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='"' // QueryRaw(tlist(it)) // '" is not a valid &
            &/intent-spec/.' )
      RETURN
    END IF
    
    ! it here points to the token after the keyword(s) for the /intent-spec/.
    
    ! Make sure nothing superfluous follows
    CALL ExpectEnd(tlist, it, comp, err_list)
    
  END SUBROUTINE ParseIntentSpec
  
END MODULE IntentSpecs
