! $Id: ModuleStmts.f90 2161 2016-06-01 23:18:33Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ModuleStmts module.


!*******************************************************************************
!!
!> Procedures for parsing module statements (/module-stmt/'s).
!!
!! R1105: /module-stmt/ is MODULE /module-name/
!!
!! Depedency analysis uses procedures defined by this module to extract 
!! a module name.

MODULE ModuleStmts
  
  USE ModuleStmtsUtils
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !----------------------------------------------------------------------------
  ! Expose module procedures and interfaces
  
  PUBLIC :: GetModuleName     ! From ModuleStmtsUtils
  PUBLIC :: ParseModuleStmt
  PUBLIC :: ParseEndModuleStmt
  
  !> Stores the results of parsing a module statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: ModuleStmt
    !> The name of the module.
    TYPE(Identifier) :: module_name
  CONTAINS
    PROCEDURE :: Dump => mod_Dump
    PROCEDURE :: Query => mod_Query
    PROCEDURE :: Modify => mod_Modify
    PROCEDURE, NOPASS :: What => mod_What
    PROCEDURE :: Advertise => mod_Advertise
  END TYPE ModuleStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /module-stmt/ (R1105).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    mod_stmt          The results of parsing the 
  !! /module-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseModuleStmt(tlist, mod_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
  
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ModuleStmt), INTENT(OUT) :: mod_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /module-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1105'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /module-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'MODULE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectName(tlist, 2, '/module-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    mod_stmt%module_name = IdentifierIfortBug(tlist(2))
    
    CALL ExpectEOS(tlist, 3, comp, err_list)
    
  END SUBROUTINE ParseModuleStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-module-stmt/ (R1106).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     name              The /module-name/ from the statement, 
  !! or not allocated if there was no name.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndModuleStmt(tlist, name, err_list)
    
    USE CompilerKinds
    USE EndProgramUnitStmts
    USE Errors
    USE Identifiers
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL ParseEndProgramUnitStmt(  &
        tlist,  &
        scck_'MODULE',  &
        'R1106',  &
        '/end-module-stmt/',  &
        '/module-name/',  &
        name,  &
        err_list )
    
  END SUBROUTINE ParseEndModuleStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ModuleStmt%Dump - write a debugging 
  !! representation of the results of parsing the separate module procedure 
  !! to a logical unit.
  !!
  !! @param[in]     pnode             The FunctionSubprogram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  ! As separate module procedures cannot contain separate module procedures, 
  ! this doesn't need to be recursive.
  SUBROUTINE mod_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ModuleStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/module-name/ = ',A)") pfx, .NameOf. pnode%module_name
    
  END SUBROUTINE mod_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ModuleStmt%Query - traverse the node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE mod_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ModuleStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE mod_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ModuleStmt%Modify - traverse the node 
  !! with a modify operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE mod_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ModuleStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE mod_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ModuleStmt%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'module-stmt'
  
  FUNCTION mod_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'module-stmt'
    
  END FUNCTION mod_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ModuleStmt%Advertise - advertise the 
  !! identifiers referenced or defined by the separate subprogram
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  
  SUBROUTINE mod_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ModuleStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%module_name,  &
        NATURES=[NatureDefModule],  &
        LOCATIONS=[.Locationof. pnode%module_name] )
    
  END SUBROUTINE mod_Advertise
  
END MODULE ModuleStmts
