! $Id: ParameterStmts.f90 2907 2019-04-12 22:44:35Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ParameterStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /parameter-stmts/.

MODULE ParameterStmts
  
  USE CompilerKinds
  USE Expressions
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseParameterStmt
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing a single item in the /named-constant-def-list/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: NamedConstantDef
    !> The name of the constant.
    TYPE(Identifier) :: name
    
    !> The value of the constant.
    !!
    !! Always allocated.
    CLASS(Expression), ALLOCATABLE :: expr
  CONTAINS
    PROCEDURE :: Write => ncd_Write
    PROCEDURE :: Query => ncd_Query
    PROCEDURE :: Modify => ncd_Modify
    PROCEDURE, NOPASS :: What => ncd_What
    PROCEDURE :: Advertise => ncd_Advertise
    PROCEDURE :: GetLocation => ncd_GetLocation
  END TYPE NamedConstantDef
  
  !> Stores the results of parsing a /parameter-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: ParameterStmt
    !> The result of parsing the declarations in the /parameter-stmt/.
    !!
    !! Not allocated on error, otherwise always allocated, maybe size zero.
    TYPE(NamedConstantDef), ALLOCATABLE :: decls(:)
  CONTAINS
    PROCEDURE :: Dump => param_Dump
    PROCEDURE :: Query => param_Query
    PROCEDURE :: Modify => param_Modify
    PROCEDURE, NOPASS :: What => param_What
  END TYPE ParameterStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /parameter-stmt/ (R548).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    param_stmt        The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseParameterStmt(tlist, param_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ParameterStmt), INTENT(OUT) :: param_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /parameter-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R548'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing parenthesis.
    INTEGER :: list_count     ! Number of items in the list of constants.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for chlid procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /parameter-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! PARAMETER...
    
    CALL ExpectKeyword(tlist, 1, scck_'PARAMETER', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! ( xxx )
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The /named-constant-def-list/.
    
    list_count = ListItemCount(tlist(3:closing-1))
    ALLOCATE(param_stmt%decls(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 3, closing - 1, item_start, item_finish)
      CALL parse_named_constant_def(  &
          tlist,  &
          item_start, item_finish,  &
          param_stmt%decls(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! That's it!
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseParameterStmt
  
  
  !*****************************************************************************
  !!
  !> Parsed a sequence of tokens that is a /named-constant-def/ (R549).
  !!
  !! @param[in]     tlist             The sequence of tokens that makes up 
  !! the statement.
  !!
  !! @param[in]     start             The index of the first token in the 
  !! /named-constant-def/ to be parsed.
  !!
  !! @param[out]    finish            The index of the last token in the 
  !! /named-constant-def/ to be parsed.
  !!
  !! @param[out]    decl              The results of parsing the 
  !! /named-constant-def/.  Not defined on error.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_named_constant_def(tlist, start, finish, decl, err_list)
    
     USE Errors
     USE ErrorCodes
     USE ParseUtils
     USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(NamedConstantDef), INTENT(OUT) :: decl
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /named-constant-def/.
    CHARACTER(*), PARAMETER :: comp = 'R549'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The name of the constant.
    
    CALL ExpectName(tlist, start, '/named-constant/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    decl%name = IdentifierIfortBug(tlist(start))
    
    !---------------------------------------------------------------------------
    ! = 
    
    CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The expression for the value.
    
    CALL ParseExpr(tlist, start + 2, finish, decl%expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
  END SUBROUTINE parse_named_constant_def
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParameterStmt%Dump - write a debugging 
  !! representation of the results of parsing the statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE param_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParameterStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Parameter declaration index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(pnode%decls)
      WRITE (unit, "(A,'/named-constant/ = ',A)") pfx,  &
          .NameOf. pnode%decls(i)%name
      WRITE (unit, "(A,'/constant-expr/ = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%decls(i)%expr)
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE param_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParameterStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE param_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParameterStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Parameter declaration index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE param_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParameterStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE param_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParameterStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Parameter declaration index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE param_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParameterStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'parameter-stmt'.
  
  FUNCTION param_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'parameter-stmt'
    
  END FUNCTION param_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in NamedConstantDef
  
  
  !*****************************************************************************
  !!
  !> Implementation of NamedConstantDef%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !1 @param[in]     pnode             The node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE ncd_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NamedConstantDef), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE ( unit,  &
        "('/named-constant/ = ',A,', /constant-expr/ = ')",   &
        ADVANCE='NO')  &
        .NameOf. pnode%name
    CALL pnode%expr%Write(unit)
    
  END SUBROUTINE ncd_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of NamedConstantDef%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE ncd_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NamedConstantDef), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ncd_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of NamedConstantDef%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE ncd_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NamedConstantDef), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ncd_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of NamedConstantDef%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns '(common-block-set)'.
  
  FUNCTION ncd_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'named-constant-def'
    
  END FUNCTION ncd_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of NamedConstantDef%Advertise - advertise the 
  !! identifiers in a DimensionDecl that are not child nodes.
  !!
  !! @param[in]     pnode             The statement to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE ncd_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NamedConstantDef), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=[NatureDefParameter],  &
        LOCATIONS=[.LocationOf. pnode%name] )
    
  END SUBROUTINE ncd_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of NamedConstantDef%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION ncd_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(NamedConstantDef), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION ncd_GetLocation
  
END MODULE ParameterStmts
