! $Id: ParseDeclarationConstructListExternal.f90 2908 2019-04-13 00:28:11Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ParseDeclarationConstructListExternal procedure.


!*******************************************************************************
!!
!> Worker routine for ParseSpecificationPart to handle the children 
!! component of the statement tree node for the specification part.
!!
!! @param[in]     stmts             The array of statement tree nodes for the 
!! specification part.  All the nodes must be declaration constructs.
!!
!! @param[out]    children          The results of parsing the array of 
!! statement tree nodes.
!!
!! @param[in,out] err_list          List of errors.

RECURSIVE SUBROUTINE ParseDeclarationConstructListExternal( stmts,  &
    children, err_list )
  
  USE Errors
  USE ParseNodes
  USE StNodes
  
  !-----------------------------------------------------------------------------
  ! Arguments
  
  TYPE(StNodeList), INTENT(INOUT) :: stmts(:)
  TYPE(SCParseNodeElement), INTENT(INOUT), ALLOCATABLE :: children(:)
  TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
  
  !***************************************************************************
  
  CALL execute(stmts, children, err_list)
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> The main execution sequence for the subroutine.  Split out to avoid 
  !! accidental re-use of host associated names.
  !!
  !! Arguments have the same meaning as for the host.
  
  RECURSIVE SUBROUTINE execute(stmts, children, err_list)
    
    USE Statements
    
    USE SingleStNodes
    USE InterfaceStNodes
    USE DerivedTypeDefStNodes
    USE EnumDefStNodes
    
    USE DerivedTypeDefs
    USE InterfaceBlocks
    USE EnumDefs
    
    USE AccessStmts
    USE AllocatableStmts
    USE AsynchronousStmts
    USE BindStmts
    USE CodimensionStmts
    USE CommonStmts
    USE ContiguousStmts
    USE DataStmts
    USE DimensionStmts
    USE EntryStmts
    USE EquivalenceStmts
    USE ExternalStmts
    USE FormatStmts
    USE IntentStmts
    USE IntrinsicStmts
    USE NamelistStmts
    USE OptionalStmts
    USE ParameterStmts
    USE PointerStmts
    USE ProcedureDeclarationStmts
    USE ProtectedStmts
    USE SaveStmts
    USE StmtFunctionStmts
    USE TargetStmts
    USE ValueStmts
    USE VolatileStmts
    
    USE TypeDeclarationStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(StNodeList), INTENT(INOUT) :: stmts(:)
    TYPE(SCParseNodeElement), INTENT(INOUT), ALLOCATABLE :: children(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: i              ! Statement tree node index.
    
    ! Generic parse node temporary.
    CLASS(SCParseNode), ALLOCATABLE :: pnode
    
    ! Results structure for a /derived-type-def/.
    TYPE(DerivedTypeDef), ALLOCATABLE :: dt_def
    
    ! Results structure for an /interface-block/.
    TYPE(InterfaceBlock), ALLOCATABLE :: intf_block
    
    ! Results structure for an /enum-def/.
    TYPE(EnumDef), ALLOCATABLE :: enum_def
    
    ! Results structure for a type declaration statement.
    TYPE(TypeDeclarationStmt), ALLOCATABLE :: type_declaration
    
    ! Results structure for an access statement.
    TYPE(AccessStmt), ALLOCATABLE :: access_stmt
    
    ! Results structure for an allocatable statement.
    TYPE(AllocatableStmt), ALLOCATABLE :: alloc_stmt
    
    ! Results structure for an asynchronous statement.
    TYPE(AsynchronousStmt), ALLOCATABLE :: async_stmt
    
    ! ParseNode for a bind statement.
    TYPE(BindStmt), ALLOCATABLE :: bind_stmt
    
    ! ParseNode for a codimension statement.
    TYPE(CodimensionStmt), ALLOCATABLE :: codim_stmt
    
    ! ParseNode for a contiguous statement.
    TYPE(ContiguousStmt), ALLOCATABLE :: contig_stmt
    
    ! ParseNode for a common statement.
    TYPE(CommonStmt), ALLOCATABLE :: common_stmt
    
    ! ParseNode for a data statement.
    TYPE(DataStmt), ALLOCATABLE :: data_stmt
    
    ! ParseNode for a dimension statement.
    TYPE(DimensionStmt), ALLOCATABLE :: dim_stmt
    
    ! ParseNode for an entry statement.
    TYPE(EntryStmt), ALLOCATABLE :: entry_stmt
    
    ! ParseNode for an equivalence statement.
    TYPE(EquivalenceStmt), ALLOCATABLE :: equiv_stmt
    
    ! ParseNode for an external statement.
    TYPE(ExternalStmt), ALLOCATABLE :: ex_stmt
    
    ! ParseNode for a format statement.
    TYPE(FormatStmt), ALLOCATABLE :: fmt_stmt
    
    ! ParseNode for an intent statement.
    TYPE(IntentStmt), ALLOCATABLE :: intent_stmt
    
    ! ParseNode for an intrinsic statement.
    TYPE(IntrinsicStmt), ALLOCATABLE :: intr_stmt
    
    ! ParseNode for a namelist statement.
    TYPE(NamelistStmt), ALLOCATABLE :: nml_stmt
    
    ! ParseNode for an optional statement.
    TYPE(OptionalStmt), ALLOCATABLE :: opt_stmt
    
    ! ParseNode for a parameter statement.
    TYPE(ParameterStmt), ALLOCATABLE :: param_stmt
    
    ! ParseNode for a pointer statement.
    TYPE(PointerStmt), ALLOCATABLE :: ptr_stmt
    
    ! ParseNode for a procedure declaration statement.
    TYPE(ProcedureDeclarationStmt), ALLOCATABLE :: procdecl_stmt
    
    ! ParseNode for a protected statement.
    TYPE(ProtectedStmt), ALLOCATABLE :: prot_stmt
    
    ! ParseNode for a save statement.
    TYPE(SaveStmt), ALLOCATABLE :: save_stmt
    
    ! ParseNode for a stmt-function-stmt.
    TYPE(StmtFunctionStmt), ALLOCATABLE :: stmtfun_stmt
    
    ! ParseNodes for a target statement.
    TYPE(TargetStmt), ALLOCATABLE :: targ_stmt
    
    ! ParseNode for a volatile statement.
    TYPE(VolatileStmt), ALLOCATABLE :: vol_stmt
    
    ! ParseNode for a value statement.
    TYPE(ValueStmt), ALLOCATABLE :: val_stmt
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    DO i = 1, SIZE(stmts)
      ! At the end of this outer select type construct, sub_err_list must 
      ! have been appropriately defined and, if successful, pnode allocated 
      ! to the appropriate dynamic type for the child node.
      SELECT TYPE (item => stmts(i)%item)
      TYPE IS (SingleStNode)
        SELECT CASE (item%stmt%ist)
        CASE (istTypeDeclaration)
          ALLOCATE(type_declaration)
          CALL ParseTypeDeclarationStmt(  &
              item%stmt%tlist,  &
              type_declaration,  &
              sub_err_list )
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(type_declaration)
          ELSE
            CALL MOVE_ALLOC(type_declaration, pnode)
          END IF
          
        CASE (istAccess)
          ALLOCATE(access_stmt)
          CALL ParseAccessStmt(item%stmt%tlist, access_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(access_stmt)
          ELSE
            CALL MOVE_ALLOC(access_stmt, pnode)
          END IF
          
        CASE (istAllocatable)
          ALLOCATE(alloc_stmt)
          CALL ParseAllocatableStmt(  &
              item%stmt%tlist,  &
              alloc_stmt,  &
              sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(alloc_stmt)
          ELSE
            CALL MOVE_ALLOC(alloc_stmt, pnode)
          END IF
          
        CASE (istAsynchronous)
          ALLOCATE(async_stmt)
          CALL ParseAsynchronousStmt(  &
              item%stmt%tlist,  &
              async_stmt,  &
              sub_err_list )
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(async_stmt)
          ELSE
            CALL MOVE_ALLOC(async_stmt, pnode)
          END IF
          
        CASE (istBind)
          ALLOCATE(bind_stmt)
          CALL ParseBindStmt(item%stmt%tlist, bind_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(bind_stmt)
          ELSE
            CALL MOVE_ALLOC(bind_stmt, pnode)
          END IF
          
        CASE (istCodimension)
          ALLOCATE(codim_stmt)
          CALL ParseCodimensionStmt(item%stmt%tlist, codim_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(codim_stmt)
          ELSE
            CALL MOVE_ALLOC(codim_stmt, pnode)
          END IF
          
        CASE (istContiguous)        ! Not listed as a specification statement, 
                                    ! but presumably this is just an oversight.
          ALLOCATE(contig_stmt)
          CALL ParseContiguousStmt(item%stmt%tlist, contig_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(contig_stmt)
          ELSE
            CALL MOVE_ALLOC(contig_stmt, pnode)
          END IF
          
        CASE (istCommon)
          ALLOCATE(common_stmt)
          CALL ParseCommonStmt(item%stmt%tlist, common_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(common_stmt)
          ELSE
            CALL MOVE_ALLOC(common_stmt, pnode)
          END IF
          
        CASE (istData)
          ALLOCATE(data_stmt)
          CALL ParseDataStmt(item%stmt%tlist, data_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(data_stmt)
          ELSE
            CALL MOVE_ALLOC(data_stmt, pnode)
          END IF
          
        CASE (istDimension)
          ALLOCATE(dim_stmt)
          CALL ParseDimensionStmt(item%stmt%tlist, dim_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(dim_stmt)
          ELSE
            CALL MOVE_ALLOC(dim_stmt, pnode)
          END IF
          
        CASE (istEntry)
          ALLOCATE(entry_stmt)
          CALL ParseEntryStmt(item%stmt%tlist, entry_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(entry_stmt)
          ELSE
            CALL MOVE_ALLOC(entry_stmt, pnode)
          END IF
          
        CASE (istEquivalence)
          ALLOCATE(equiv_stmt)
          CALL ParseEquivalenceStmt(item%stmt%tlist, equiv_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(equiv_stmt)
          ELSE
            CALL MOVE_ALLOC(equiv_stmt, pnode)
          END IF
          
        CASE (istExternal)
          ALLOCATE(ex_stmt)
          CALL ParseExternalStmt(item%stmt%tlist, ex_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(ex_stmt)
          ELSE
            CALL MOVE_ALLOC(ex_stmt, pnode)
          END IF
          
        CASE (istFormat)
          ALLOCATE(fmt_stmt)
          CALL ParseFormatStmt(item%stmt%tlist, fmt_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(fmt_stmt)
          ELSE
            CALL MOVE_ALLOC(fmt_stmt, pnode)
          END IF
          
        CASE (istIntent)
          ALLOCATE(intent_stmt)
          CALL ParseIntentStmt(item%stmt%tlist, intent_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(intent_stmt)
          ELSE
            CALL MOVE_ALLOC(intent_stmt, pnode)
          END IF
          
        CASE (istIntrinsic)
          ALLOCATE(intr_stmt)
          CALL ParseIntrinsicStmt(item%stmt%tlist, intr_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(intr_stmt)
          ELSE
            CALL MOVE_ALLOC(intr_stmt, pnode)
          END IF
          
        CASE (istNamelist)
          ALLOCATE(nml_stmt)
          CALL ParseNamelistStmt(item%stmt%tlist, nml_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(nml_stmt)
          ELSE
            CALL MOVE_ALLOC(nml_stmt, pnode)
          END IF
          
        CASE (istOptional)
          ALLOCATE(opt_stmt)
          CALL ParseOptionalStmt(item%stmt%tlist, opt_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(opt_stmt)
          ELSE
            CALL MOVE_ALLOC(opt_stmt, pnode)
          END IF
          
        CASE (istParameter)
          ALLOCATE(param_stmt)
          CALL ParseParameterStmt(item%stmt%tlist, param_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(param_stmt)
          ELSE
            CALL MOVE_ALLOC(param_stmt, pnode)
          END IF
          
        CASE (istPointer)
          ALLOCATE(ptr_stmt)
          CALL ParsePointerStmt(item%stmt%tlist, ptr_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(ptr_stmt)
          ELSE
            CALL MOVE_ALLOC(ptr_stmt, pnode)
          END IF
          
        CASE (istProcedureDeclaration)
          ALLOCATE(procdecl_stmt)
          CALL ParseProcedureDeclarationStmt(  &
              item%stmt%tlist,  &
              procdecl_stmt,  &
              sub_err_list )
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(procdecl_stmt)
          ELSE
            CALL MOVE_ALLOC(procdecl_stmt, pnode)
          END IF
          
        CASE (istProtected)
          ALLOCATE(prot_stmt)
          CALL ParseProtectedStmt(item%stmt%tlist, prot_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(prot_stmt)
          ELSE
            CALL MOVE_ALLOC(prot_stmt, pnode)
          END IF
          
        CASE (istSave)
          ALLOCATE(save_stmt)
          CALL ParseSaveStmt(item%stmt%tlist, save_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(save_stmt)
          ELSE
            CALL MOVE_ALLOC(save_stmt, pnode)
          END IF
          
        CASE (istStmtFunction)
          ALLOCATE(stmtfun_stmt)
          CALL ParseStmtFunctionStmt(  &
              item%stmt%tlist,  &
              stmtfun_stmt,  &
              sub_err_list )
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(stmtfun_stmt)
          ELSE
            CALL MOVE_ALLOC(stmtfun_stmt, pnode)
          END IF
          
        CASE (istTarget)
          ALLOCATE(targ_stmt)
          CALL ParseTargetStmt(item%stmt%tlist, targ_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(targ_stmt)
          ELSE
            CALL MOVE_ALLOC(targ_stmt, pnode)
          END IF
          
        CASE (istValue)
          ALLOCATE(val_stmt)
          CALL ParseValueStmt(item%stmt%tlist, val_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(val_stmt)
          ELSE
            CALL MOVE_ALLOC(val_stmt, pnode)
          END IF
          
        CASE (istVolatile)
          ALLOCATE(vol_stmt)
          CALL ParseVolatileStmt(item%stmt%tlist, vol_stmt, sub_err_list)
          IF (Failure(sub_err_list)) THEN
            DEALLOCATE(vol_stmt)
          ELSE
            CALL MOVE_ALLOC(vol_stmt, pnode)
          END IF
          
        CASE DEFAULT
          PRINT "('The unknown statement is a ',A)", GetStName(item%stmt%ist)
          STOP 'Internal error in ParseDeclarationConstructListExternal, &
              &unknown SingleStNode.'
          
        END SELECT
        
      TYPE IS (DerivedTypeDefStNode)
        ALLOCATE(dt_def)
        CALL ParseDerivedTypeDef(item, dt_def, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(dt_def)
        ELSE
          CALL MOVE_ALLOC(dt_def, pnode)
        END IF
        
      TYPE IS (InterfaceStNode)
        ALLOCATE(intf_block)
        CALL ParseInterfaceBlock(item, intf_block, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(intf_block)
        ELSE
          CALL MOVE_ALLOC(intf_block, pnode)
        END IF
        
      TYPE IS (EnumDefStNode)
        ALLOCATE(enum_def)
        CALL ParseEnumDef(item, enum_def, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(enum_def)
        ELSE
          CALL MOVE_ALLOC(enum_def, pnode)
        END IF
        
      CLASS DEFAULT
        PRINT "('The unknown statement tree node is for a ',A)",  &
            GetPartName(item)
        STOP 'Internal error in ParseDeclarationConstructListExternal, &
            &unknown StNode.'
        
      END SELECT
      CALL Add(err_list, sub_err_list)
      IF (.NOT. Failure(sub_err_list)) THEN
        CALL pnode%Init(stmts(i)%item)
        CALL AddPNode(children, pnode)
      END IF
    END DO
    
  END SUBROUTINE execute
  
END SUBROUTINE ParseDeclarationConstructListExternal
