! $Id: ParseExecutionPartConstructListExternal.f90 2911 2019-04-16 11:28:52Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ParseExecutionPartConstructListExternal procedure.


!*******************************************************************************
!!
!> Procedure that carries out parsing of the syntax term:
!!  [ /execution-part-construct/ ] ...
!!
!! which is the whole of an execution-part, and also features in any 
!! executable construct that contains a /block/.
!!
!! Client code shouldn't be calling this - use the equivalent module procedure 
!! from ExecutionParts instead - that way we don't have to worry about calls 
!! missing explicit interfaces.
!!
!! Why didn't we call this ParseBlockExternal?
!!
!! If the interface of this procedure is changed, then make sure that 
!! the corresponding interface body in ExecutionParts is updated!

SUBROUTINE ParseExecutionPartConstructListExternal( exec_st, children,  &
    err_list )
  
  USE ExecStNodes
  USE Errors
  USE ParseNodes
  USE StNodes
  
  USE AssociateConstructStNodes
  USE AssociateConstructs
  USE BlockConstructStNodes
  USE BlockConstructs
  USE CaseConstructStNodes
  USE CaseConstructs
  USE CriticalConstructStNodes
  USE CriticalConstructs
  USE DoConstructStNodes
  USE DoConstructs
  USE ForallConstructStNodes
  USE ForallConstructs
  USE IfConstructStNodes
  USE IfConstructs
  USE SelectTypeConstructStNodes
  USE SelectTypeConstructs
  USE WhereConstructStNodes
  USE WhereConstructs
  
  USE SingleStNodes
  USE Statements
  USE AllocateStmts
  USE AssignmentStmts
  USE BackspaceStmts
  USE CallStmts
  USE CloseStmts
  USE ContinueStmts
  USE DataStmts
  USE DeallocateStmts
  USE DoStmts                 ! For CycleStmt
  USE EndfileStmts
  USE EntryStmts
  USE ExitStmts
  USE FlushStmts
  USE FormatStmts
  USE IfStmts
  USE InquireStmts
  USE GotoStmts
  USE LockStmts
  USE NullifyStmts
  USE OpenStmts
  USE PointerAssignmentStmts
  USE PrintStmts
  USE ReadStmts
  USE ReturnStmts
  USE RewindStmts
  USE StopStmts
  USE SyncStmts
  USE WaitStmts
  USE WriteStmts
  
  IMPLICIT NONE
  
  !-----------------------------------------------------------------------------
  ! Arguments
  
  TYPE(ExecStNode), INTENT(INOUT) :: exec_st
  TYPE(SCParseNodeElement), INTENT(OUT), ALLOCATABLE :: children(:)
  TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
  
  !---------------------------------------------------------------------------
  ! Local variables
  
  INTEGER :: i              ! Child index.
  
  ! Error list for child procedure calls.
  TYPE(Error), ALLOCATABLE :: sub_err_list(:)
  
  ! Statement nodes.  For each type of node we have an allocatable object of 
  ! the correct type.  When processing a particular st node we allocate that 
  ! object, construct the object using the appropriate ParseXxxx routine then 
  ! MOVE_ALLOC the allocatable object into the list of child parse nodes.
  
  ! ParseNode for an allocate statement.
  TYPE(AllocateStmt), ALLOCATABLE :: allocate_stmt
  
  ! ParseNode for an arithmetic if statement.
  TYPE(ArithmeticIfStmt), ALLOCATABLE :: arith_stmt
  
  ! PNode for an assignment statement.
  TYPE(AssignmentStmt), ALLOCATABLE :: assignment
  
  ! ParseNode for a backspace statement.
  TYPE(BackspaceStmt), ALLOCATABLE :: backspace_stmt
  
  ! ParseNode for a call statement
  TYPE(CallStmt), ALLOCATABLE :: call_stmt
  
  ! ParseNode for a computed goto statement.
  TYPE(ComputedGotoStmt), ALLOCATABLE :: cgoto_stmt
  
  ! ParseNode for a continue statement.
  TYPE(ContinueStmt), ALLOCATABLE :: continue_stmt
  
  ! ParseNode for a nullify statement.
  TYPE(NullifyStmt), ALLOCATABLE :: nullify_stmt
  
  ! PNode for an open statement.
  TYPE(OpenStmt), ALLOCATABLE :: open_stmt
  
  ! PNode for a close statement.
  TYPE(CloseStmt), ALLOCATABLE :: close_stmt
  
  ! PNode for a cycle statement.
  TYPE(CycleStmt), ALLOCATABLE :: cycle_stmt
  
  ! PNode for a data statement.
  TYPE(DataStmt), ALLOCATABLE :: data_stmt
  
  ! PNode for a deallocate statement.
  TYPE(DeallocateStmt), ALLOCATABLE :: dealloc_stmt
  
  ! PNode for an endfile statement.
  TYPE(EndfileStmt), ALLOCATABLE :: endfile_stmt
  
  ! PNode for an entry statement.
  TYPE(EntryStmt), ALLOCATABLE :: entry_stmt
  
  ! PNode for an exit statement.
  TYPE(ExitStmt), ALLOCATABLE :: exit_stmt
  
  ! PNode for a flush statement.
  TYPE(FlushStmt), ALLOCATABLE :: flush_stmt
  
  ! PNode for a goto statement.
  TYPE(GotoStmt), ALLOCATABLE :: goto_stmt
  
  ! PNode for a lock statement.
  TYPE(LockStmt), ALLOCATABLE :: lock_stmt
  
  ! PNode for a stop statement.
  TYPE(StopStmt), ALLOCATABLE :: stop_stmt
  
  ! PNode for an error stop statement.
  TYPE(ErrorStopStmt), ALLOCATABLE :: errstop_stmt
  
  ! PNode for an inquire statement.
  TYPE(InquireStmt), ALLOCATABLE :: inquire_stmt
  
  ! PNode for a format statement.
  TYPE(FormatStmt), ALLOCATABLE :: fmt_stmt
  
  ! PNode for a return statement.
  TYPE(ReturnStmt), ALLOCATABLE :: return_stmt
  
  ! PNode for a rewindstatement.
  TYPE(RewindStmt), ALLOCATABLE :: rewind_stmt
  
  ! PNode for a pointer assignment statement.
  TYPE(PointerAssignmentStmt), ALLOCATABLE :: ptrassign_stmt
  
  ! PNode for a print statement.
  TYPE(PrintStmt), ALLOCATABLE :: print_stmt
  
  ! ParseNode for a read statement.
  TYPE(ReadStmt), ALLOCATABLE :: read_stmt
  
  ! PNode for a sync all statement.
  TYPE(SyncAllStmt), ALLOCATABLE :: syncall_stmt
  
  ! PNode for a sync images statement.
  TYPE(SyncImagesStmt), ALLOCATABLE :: syncimages_stmt
  
  ! PNode for a sync memory statement.
  TYPE(SyncMemoryStmt), ALLOCATABLE :: syncmem_stmt
  
  ! PNode for an unlock statement.
  TYPE(UnlockStmt), ALLOCATABLE :: unlock_stmt
  
  ! PNode for a wait statement.
  TYPE(WaitStmt), ALLOCATABLE :: wait_stmt
  
  ! PNode for a write statement.
  TYPE(WriteStmt), ALLOCATABLE :: write_stmt
  
  ! Generic parse node temporary.
  CLASS(SCParseNode), ALLOCATABLE :: pnode
  
  ! Construct nodes.
  
  ! ParseNode for an associate construct.
  TYPE(AssociateConstruct), ALLOCATABLE :: assoc_construct
  
  ! Parse node for a block construct.
  TYPE(BlockConstruct), ALLOCATABLE :: block_construct
  
  ! Parse node for a critical-construct.
  TYPE(CriticalConstruct), ALLOCATABLE :: crit_construct
  
  ! Parse node for a critical-construct.
  TYPE(CaseConstruct), ALLOCATABLE :: case_construct
  
  ! Parse node for a do-construct.
  TYPE(DoConstruct), ALLOCATABLE :: do_construct
  
  ! Parse node for an if-construct.  Also used for the single statement 
  ! /if-stmt/ variant.
  TYPE(IfConstruct), ALLOCATABLE :: if_construct
  
  ! Parse node for a forall construct.  Also used for the single statement 
  ! /forall-stmt/ variant.
  TYPE(ForallConstruct), ALLOCATABLE :: forall_construct
  
  ! Parse node for a select type construct.
  TYPE(SelectTypeConstruct), ALLOCATABLE :: st_construct
  
  ! Parse node for a where construct.  Also used for the single statement 
  ! /where-stmt/ variant.
  TYPE(WhereConstruct), ALLOCATABLE :: where_construct
  
  !*****************************************************************************
  
  ALLOCATE(err_list(0))
  
  !-----------------------------------------------------------------------------
  
  ALLOCATE(children(0))
  
  DO i = 1, SIZE(exec_st%children)
    SELECT TYPE (item => exec_st%children(i)%item)
    TYPE IS (SingleStNode)
      SELECT CASE (item%stmt%ist)
      CASE (istAllocate)
        ALLOCATE(allocate_stmt)
        CALL ParseAllocateStmt(item%stmt%tlist, allocate_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(allocate_stmt)
        ELSE
          CALL MOVE_ALLOC(allocate_stmt, pnode)
        END IF
        
      CASE (istArithmeticIf)
        ALLOCATE(arith_stmt)
        CALL ParseArithmeticIfStmt(item%stmt%tlist, arith_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(arith_stmt)
        ELSE
          CALL MOVE_ALLOC(arith_stmt, pnode)
        END IF
        
      CASE (istAssignment)
        ALLOCATE(assignment)
        CALL ParseAssignmentStmt(item%stmt%tlist, assignment, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(assignment)
        ELSE
          CALL MOVE_ALLOC(assignment, pnode)
        END IF
        
      CASE (istBackspace)
        ALLOCATE(backspace_stmt)
        CALL ParseBackspaceStmt(item%stmt%tlist, backspace_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(backspace_stmt)
        ELSE
          CALL MOVE_ALLOC(backspace_stmt, pnode)
        END IF
        
      CASE (istCall)
        ALLOCATE(call_stmt)
        CALL ParseCallStmt(item%stmt%tlist, call_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(call_stmt)
        ELSE
          CALL MOVE_ALLOC(call_stmt, pnode)
        END IF
        
      CASE (istClose)
        ALLOCATE(close_stmt)
        CALL ParseCloseStmt(item%stmt%tlist, close_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(close_stmt)
        ELSE
          CALL MOVE_ALLOC(close_stmt, pnode)
        END IF
        
      CASE (istComputedGoto)
        ALLOCATE(cgoto_stmt)
        CALL ParseComputedGotoStmt(item%stmt%tlist, cgoto_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(cgoto_stmt)
        ELSE
          CALL MOVE_ALLOC(cgoto_stmt, pnode)
        END IF
        
      CASE (istContinue)
        ALLOCATE(continue_stmt)
        CALL ParseContinueStmt(item%stmt%tlist, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(continue_stmt)
        ELSE
          CALL MOVE_ALLOC(continue_stmt, pnode)
        END IF
        
      CASE (istCycle)
        ALLOCATE(cycle_stmt)
        CALL ParseCycleStmt(item%stmt%tlist, cycle_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(cycle_stmt)
        ELSE
          CALL MOVE_ALLOC(cycle_stmt, pnode)
        END IF
        
      CASE (istData)
        ALLOCATE(data_stmt)
        CALL ParseDataStmt(item%stmt%tlist, data_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(data_stmt)
        ELSE
          CALL MOVE_ALLOC(data_stmt, pnode)
        END IF
        
      CASE (istDeallocate)
        ALLOCATE(dealloc_stmt)
        CALL ParseDeallocateStmt(item%stmt%tlist, dealloc_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(dealloc_stmt)
        ELSE
          CALL MOVE_ALLOC(dealloc_stmt, pnode)
        END IF
        
      CASE (istEndfile)
        ALLOCATE(endfile_stmt)
        CALL ParseEndfileStmt(item%stmt%tlist, endfile_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(endfile_stmt)
        ELSE
          CALL MOVE_ALLOC(endfile_stmt, pnode)
        END IF
        
      CASE (istEntry)
        ALLOCATE(entry_stmt)
        CALL ParseEntryStmt(item%stmt%tlist, entry_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(entry_stmt)
        ELSE
          CALL MOVE_ALLOC(entry_stmt, pnode)
        END IF
        
      CASE (istExit)
        ALLOCATE(exit_stmt)
        CALL ParseExitStmt(item%stmt%tlist, exit_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(exit_stmt)
        ELSE
          CALL MOVE_ALLOC(exit_stmt, pnode)
        END IF
        
      CASE (istFlush)
        ALLOCATE(flush_stmt)
        CALL ParseFlushStmt(item%stmt%tlist, flush_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(flush_stmt)
        ELSE
          CALL MOVE_ALLOC(flush_stmt, pnode)
        END IF
        
      CASE (istForall)
        ALLOCATE(forall_construct)
        CALL ParseForallStmtToConstruct( item%stmt%tlist, forall_construct,  &
            sub_err_list )
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(forall_construct)
        ELSE
          CALL MOVE_ALLOC(forall_construct, pnode)
        END IF
        
      CASE (istFormat)
        ALLOCATE(fmt_stmt)
        CALL ParseFormatStmt(item%stmt%tlist, fmt_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(fmt_stmt)
        ELSE
          CALL MOVE_ALLOC(fmt_stmt, pnode)
        END IF
        
      CASE (istGoto)
        ALLOCATE(goto_stmt)
        CALL ParseGotoStmt(item%stmt%tlist, goto_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(goto_stmt)
        ELSE
          CALL MOVE_ALLOC(goto_stmt, pnode)
        END IF
        
      CASE (istIf)
        ALLOCATE(if_construct)
        CALL ParseIfStmtToConstruct( item%stmt%tlist, if_construct,  &
            sub_err_list )
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(if_construct)
        ELSE
          CALL MOVE_ALLOC(if_construct, pnode)
        END IF
        
      CASE (istInquire)
        ALLOCATE(inquire_stmt)
        CALL ParseInquireStmt(item%stmt%tlist, inquire_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(inquire_stmt)
        ELSE
          CALL MOVE_ALLOC(inquire_stmt, pnode)
        END IF
        
      CASE (istLock)
        ALLOCATE(lock_stmt)
        CALL ParseLockStmt(item%stmt%tlist, lock_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(lock_stmt)
        ELSE
          CALL MOVE_ALLOC(lock_stmt, pnode)
        END IF
        
      CASE (istNullify)
        ALLOCATE(nullify_stmt)
        CALL ParseNullifyStmt(item%stmt%tlist, nullify_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(nullify_stmt)
        ELSE
          CALL MOVE_ALLOC(nullify_stmt, pnode)
        END IF
        
      CASE (istOpen)
        ALLOCATE(open_stmt)
        CALL ParseOpenStmt(item%stmt%tlist, open_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(open_stmt)
        ELSE
          CALL MOVE_ALLOC(open_stmt, pnode)
        END IF
        
      CASE (istPointerAssignment)
        ALLOCATE(ptrassign_stmt)
        CALL ParsePointerAssignmentStmt( item%stmt%tlist, ptrassign_stmt,  &
            sub_err_list )
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(ptrassign_stmt)
        ELSE
          CALL MOVE_ALLOC(ptrassign_stmt, pnode)
        END IF
        
      CASE (istPrint)
        ALLOCATE(print_stmt)
        CALL ParsePrintStmt(item%stmt%tlist, print_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(print_stmt)
        ELSE
          CALL MOVE_ALLOC(print_stmt, pnode)
        END IF
        
      CASE (istRead)
        ALLOCATE(read_stmt)
        CALL ParseReadStmt(item%stmt%tlist, read_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(read_stmt)
        ELSE
          CALL MOVE_ALLOC(read_stmt, pnode)
        END IF
        
      CASE (istReturn)
        ALLOCATE(return_stmt)
        CALL ParseReturnStmt(item%stmt%tlist, return_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(return_stmt)
        ELSE
          CALL MOVE_ALLOC(return_stmt, pnode)
        END IF
        
      CASE (istRewind)
        ALLOCATE(rewind_stmt)
        CALL ParseRewindStmt(item%stmt%tlist, rewind_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(rewind_stmt)
        ELSE
          CALL MOVE_ALLOC(rewind_stmt, pnode)
        END IF
        
      CASE (istStop)
        ALLOCATE(stop_stmt)
        CALL ParseStopStmt(item%stmt%tlist, stop_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(stop_stmt)
        ELSE
          CALL MOVE_ALLOC(stop_stmt, pnode)
        END IF
        
      CASE (istErrorStop)
        ALLOCATE(errstop_stmt)
        CALL ParseErrorStopStmt(item%stmt%tlist, errstop_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(errstop_stmt)
        ELSE
          CALL MOVE_ALLOC(errstop_stmt, pnode)
        END IF
        
      CASE (istSyncAll)
        ALLOCATE(syncall_stmt)
        CALL ParseSyncAllStmt(item%stmt%tlist, syncall_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(syncall_stmt)
        ELSE
          CALL MOVE_ALLOC(syncall_stmt, pnode)
        END IF
        
      CASE (istSyncImages)
        ALLOCATE(syncimages_stmt)
        CALL ParseSyncImagesStmt(item%stmt%tlist, syncimages_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(syncimages_stmt)
        ELSE
          CALL MOVE_ALLOC(syncimages_stmt, pnode)
        END IF
        
      CASE (istSyncMemory)
        ALLOCATE(syncmem_stmt)
        CALL ParseSyncMemoryStmt(item%stmt%tlist, syncmem_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(syncmem_stmt)
        ELSE
          CALL MOVE_ALLOC(syncmem_stmt, pnode)
        END IF
        
      CASE (istUnLock)
        ALLOCATE(unlock_stmt)
        CALL ParseUnlockStmt(item%stmt%tlist, unlock_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(unlock_stmt)
        ELSE
          CALL MOVE_ALLOC(unlock_stmt, pnode)
        END IF
        
      CASE (istWait)
        ALLOCATE(wait_stmt)
        CALL ParseWaitStmt(item%stmt%tlist, wait_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(wait_stmt)
        ELSE
          CALL MOVE_ALLOC(wait_stmt, pnode)
        END IF
        
      CASE (istWhere)
        ALLOCATE(where_construct)
        CALL ParseWhereStmtToConstruct( item%stmt%tlist, where_construct,  &
            sub_err_list )
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(where_construct)
        ELSE
          CALL MOVE_ALLOC(where_construct, pnode)
        END IF
        
      CASE (istWrite)
        ALLOCATE(write_stmt)
        CALL ParseWriteStmt(item%stmt%tlist, write_stmt, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(write_stmt)
        ELSE
          CALL MOVE_ALLOC(write_stmt, pnode)
        END IF
        
      CASE DEFAULT
        PRINT "('The unknown statement is a ',A)", GetStName(item%stmt%ist)
        STOP 'Internal error in ParseExecutionPartConstructListExternal, &
            &unknown SingleStNode.'
        
      END SELECT
      
    TYPE IS (AssociateConstructStNode)
      ALLOCATE(assoc_construct)
      CALL ParseAssociateConstruct(item, assoc_construct, sub_err_list)
      IF (Failure(sub_err_list)) THEN
        DEALLOCATE(assoc_construct)
      ELSE
        CALL MOVE_ALLOC(assoc_construct, pnode)
      END IF
      
    TYPE IS (BlockConstructStNode)
      ALLOCATE(block_construct)
      CALL ParseBlockConstruct(item, block_construct, sub_err_list)
      IF (Failure(sub_err_list)) THEN
        DEALLOCATE(block_construct)
      ELSE
        CALL MOVE_ALLOC(block_construct, pnode)
      END IF
      
    TYPE IS (CaseConstructStNode)
      ALLOCATE(case_construct)
      CALL ParseCaseConstruct(item, case_construct, sub_err_list)
      IF (Failure(sub_err_list)) THEN
        DEALLOCATE(case_construct)
      ELSE
        CALL MOVE_ALLOC(case_construct, pnode)
      END IF
      
    TYPE IS (CriticalConstructStNode)
      ALLOCATE(crit_construct)
      CALL ParseCriticalConstruct(item, crit_construct, sub_err_list)
      IF (Failure(sub_err_list)) THEN
        DEALLOCATE(crit_construct)
      ELSE
        CALL MOVE_ALLOC(crit_construct, pnode)
      END IF
      
    TYPE IS (DoConstructStNode)
      ALLOCATE(do_construct)
      CALL ParseDoConstruct(item, do_construct, sub_err_list)
      IF (Failure(sub_err_list)) THEN
        DEALLOCATE(do_construct)
      ELSE
        CALL MOVE_ALLOC(do_construct, pnode)
      END IF
      
    TYPE IS (ForallConstructStNode)
      ALLOCATE(forall_construct)
      CALL ParseForallConstruct(item, forall_construct, sub_err_list)
      IF (Failure(sub_err_list)) THEN
        DEALLOCATE(forall_construct)
      ELSE
        CALL MOVE_ALLOC(forall_construct, pnode)
      END IF
      
    TYPE IS (IfConstructStNode)
      ALLOCATE(if_construct)
      CALL ParseIfConstruct(item, if_construct, sub_err_list)
      IF (Failure(sub_err_list)) THEN
        DEALLOCATE(if_construct)
      ELSE
        CALL MOVE_ALLOC(if_construct, pnode)
      END IF
      
    TYPE IS (SelectTypeConstructStNode)
      ALLOCATE(st_construct)
      CALL ParseSelectTypeConstruct(item, st_construct, sub_err_list)
      IF (Failure(sub_err_list)) THEN
        DEALLOCATE(st_construct)
      ELSE
        CALL MOVE_ALLOC(st_construct, pnode)
      END IF
      
    TYPE IS (WhereConstructStNode)
      ALLOCATE(where_construct)
      CALL ParseWhereConstruct(item, where_construct, sub_err_list)
      IF (Failure(sub_err_list)) THEN
        DEALLOCATE(where_construct)
      ELSE
        CALL MOVE_ALLOC(where_construct, pnode)
      END IF
      
    CLASS DEFAULT
      PRINT "('The unknown statement tree node is for a ',A)",  &
          GetPartName(item)
      STOP 'Internal error in ParseExecutionPartConstructListExternal, &
          &unknown StNode.'
      
    END SELECT
    CALL Add(err_list, sub_err_list)
    IF (.NOT. Failure(sub_err_list)) THEN
      CALL pnode%Init(exec_st%children(i)%item)
      CALL AddPNode(children, pnode)
    END IF
  END DO
  
END SUBROUTINE ParseExecutionPartConstructListExternal
