! $Id: PositionSpecs.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the PositionSpecs module.


!*******************************************************************************
!!
!> A type and procedures for parsing /position-spec-list/'s and 
!! /position-spec/'s.
!!
!! These may appear in the second forms of each of /backspace-stmt/'s, 
!! /endfile-stmt/'s and /rewind-stmt/'s.

MODULE PositionSpecs
  
  USE Expressions
  USE Labels
  USE ParseNodes
  USE SourceLocations
  USE Variables
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParsePositionSpecList
  PUBLIC :: DumpPositionSpecList
  PUBLIC :: TestPositionStmtFirstForm
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing an /position-spec-list/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: PositionSpecList
    !> The location of the position spec list.
    !!
    !! This is held explicitly to overcome any issues with an unspecified 
    !! unit.
    TYPE(SourceLocation) :: location
    !> The UNIT specification.
    !!
    !! This is a /file-unit-number/.
    !!
    !! This should always be allocated AFTER successful checks.
    CLASS(Expression), ALLOCATABLE :: unit
    !> The IOMSG specification
    !!
    !! This is an /iomsg-variable/.
    CLASS(Variable), ALLOCATABLE :: iomsg
    !> The IOSTAT specification.
    CLASS(Variable), ALLOCATABLE :: iostat
    !> The ERR specification.
    !!
    !! Not allocated if there was no ERR specifier.
    TYPE(Label), ALLOCATABLE :: err
  CONTAINS
    PROCEDURE :: Write => psl_Write
    PROCEDURE :: Query => psl_Query
    PROCEDURE :: Modify => psl_Modify
    PROCEDURE, NOPASS :: What => psl_What
    PROCEDURE :: GetLocation => psl_GetLocation
  END TYPE PositionSpecList
  
  !-----------------------------------------------------------------------------
  
  !> Syntax rule number for a /position-spec/.
  CHARACTER(*), PARAMETER :: comp = 'R927'
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /position-spec-list/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist of 
  !! the list.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist of 
  !! the list.
  !!
  !! @param[out]    position_specs    The results of parsing the list.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! The token sequence for the list does not include the delimiting "(" and 
  !! ")".
  
  SUBROUTINE ParsePositionSpecList( tlist, start, finish, position_specs,  &
      err_list )
    
    USE Errors
    USE ErrorCodes
    USE Scanner
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(PositionSpecList), INTENT(OUT) :: position_specs
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the /close-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.   
    LOGICAL :: no_spec        ! Flag to indicate UNIT= absent.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      ! Note no component - that will be provided by the parent syntax rule.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          MSG='Expected /position-spec-list/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Save location.
    position_specs%location = QueryLocation(tlist(start))
    
    list_count = ListItemCount(tlist(start:finish))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      CALL parse_position_spec(  &
          tlist,  &
          item_start, item_finish,  &
          position_specs,  &
          no_spec,  &
          sub_err_list )
      ! Note no component chaining.
      CALL Add(err_list, sub_err_list)
      ! Continue on error, but don't rely upon parse_position_spec output.
      IF (.NOT. Failure(sub_err_list)) THEN
        IF ((item /= 1) .AND. no_spec) THEN
          ! No syntax rule.
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C942',  &
  !            LOCATION=tlist(item_start)%location,  
              MSG='If the optional "UNIT=" is omitted, the /file-unit-number/ &
                &shall be the first item in the /position-spec-list/.' )
          ! Continue on error.
        END IF
      END IF
    END DO
        
    !---------------------------------------------------------------------------
    ! Check constraints
    
    IF (.NOT. ALLOCATED(position_specs%unit)) THEN
      ! No syntax rule.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C942',  &
!          LOCATION=tlist(start)%location,  
          MSG='A /file-unit-number/ shall be specified in a &
            &/position-spec-list/.' )
      ! Continue on error.
    END IF
    
  END SUBROUTINE ParsePositionSpecList
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /position-spec/ (R927).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist of 
  !! the item.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist of 
  !! the item.
  !!
  !! @param[out]    position_specs    The results of parsing the spec.  Note 
  !! that this is an INTENT(INOUT) argument - each spec sets a particular 
  !! component.
  !!
  !! @param[out]    no_spec           Flag to indicate that the "XXX=" part of 
  !! the spec was absent.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_position_spec( tlist, start, finish, position_specs,  &
      no_spec, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE IoControlSpecs, ONLY: ParseFileUnitNumber
    USE MatchUtils
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
      
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(PositionSpecList), INTENT(INOUT) :: position_specs
    LOGICAL, INTENT(OUT) :: no_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Keywords for each /position-spec/.
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(4) = [  &
        scck_'UNIT  ', scck_'IOMSG ', scck_'IOSTAT', scck_'ERR   ' ]
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option for the keyword test.
    INTEGER :: assign         ! Index of the =.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /position-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Do we have an  argument keyword?
    
    assign = FindBare(tlist(:finish), ittAssign, scck_'=', start)
    IF (assign == 0) THEN
      ! No.
      no_spec = .TRUE.
      opt = 1
    ELSE
      ! Yes.  What was the keyword?
      no_spec = .FALSE.
      CALL ExpectKeyword(tlist, start, keywords, '/position-spec/', opt, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    
      CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    SELECT CASE (opt)
    CASE (1)      ! UNIT
      IF (ALLOCATED(position_specs%unit)) THEN
        CALL already('UNIT')
      ELSE
        IF (assign == 0) THEN
          CALL ParseFileUnitNumber( tlist, start, finish,  &
              position_specs%unit, sub_err_list)
        ELSE
          CALL ParseFileUnitNumber( tlist, assign + 1, finish,  &
              position_specs%unit, sub_err_list )
        END IF
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (2)      ! IOMSG
      IF (ALLOCATED(position_specs%iomsg)) THEN
        CALL already('IOMSG')
      ELSE
        ! The syntax term is actually an /iomsg-variable/.
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            position_specs%iomsg,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (3)      ! IOSTAT
      IF (ALLOCATED(position_specs%iostat)) THEN
        CALL already('IOSTAT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            position_specs%iostat,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (4)      ! ERR
      IF (ALLOCATED(position_specs%err)) THEN
        CALL already('ERR')
      ELSE
        ALLOCATE(position_specs%err)
        CALL ParseLabel(  &
            tlist,  &
            assign + 1,  &
            finish,  &
            position_specs%err,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE DEFAULT
      STOP 'Unknown /position-spec/.'
      
    END SELECT
    
  CONTAINS
    
    SUBROUTINE already(text)
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list, CODE=errSyntax, COMPONENT='C941',  &
!          LOCATION=tlist(start)%location,
          MSG=text // ' has already been specified for this statement.' )
    END SUBROUTINE already
    
  END SUBROUTINE parse_position_spec
  
  
  !*****************************************************************************
  !!
  !> Test the form of a position statement (a statement defined in section 
  !! 9.8, such as a /backspace-stmt/, /endfile-stmt/ or /rewind-stmt/).
  !!
  !! @param[in]     tlist             The sequence of tokens to test.  This 
  !! sequence should start with the first token after the keyword (in the 
  !! case of the /endfile-stmt/ this could be either the second or third 
  !! token in the statement.
  !!
  !! @returns .TRUE. If the statement appears to be of the first form, 
  !! where there is no /position-spec-list/ and the expression for the 
  !! /file-unit-number/ directly follows the keyword, .FALSE. if the statement 
  !! appears to be of the second form, where there is a /position-spec-list/
  !! in parentheses after the keyword.
  !!
  !! It is possible to write syntax that fits both forms - have an expression 
  !! that is a vlid unit number that is then pointlessly prenthesised.  In 
  !! this case we take that as the second form, as that is what the reader 
  !! would typically expect.
  !!
  !! Note that the first and second forms are the other way around from those 
  !! of a /read-stmt/.
  
  FUNCTION TestPositionStmtFirstForm(tlist) RESULT(first_form)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result
    LOGICAL :: first_form
    
    !---------------------------------------------------------------------------
    ! Local variables.
    
    INTEGER :: closing        ! Index of the closing parenthesis.
    INTEGER :: finish         ! End of the parenthesised bit.
    
    !***************************************************************************
    
    ! Second form must have "(" after the keyword.
    first_form = .NOT. Match(tlist, 1, ittDelimiter, scck_'(')
    IF (first_form) RETURN    ! No "(" - must be first form.
    
    closing = FindClosingParens(tlist, 1)
    IF (closing == 0) THEN
      ! Tolerate a mismatch and see if we can deduce the programmers intent.
      finish = SIZE(tlist)
    ELSE
      finish = closing - 1
    END IF
    
    ! First form must not have a bare comma in the parenthesised bits.
    first_form = FindBare(tlist(2:finish), ittComma, scck_',') == 0
    IF (.NOT. first_form) RETURN    ! There was a comma - must be second form.
    
    ! First form must not have argument keywords.
    first_form = FindBare(tlist(3:closing-1), ittComma, scck_'=') == 0
    IF (.NOT. first_form) RETURN    ! There was a "=" - must be second form.
    
    IF (closing /= 0) THEN
      ! Closing parenthesis must be the last token in the statement 
      ! for a valid second form.
      first_form = (closing /= SIZE(tlist))
      IF (first_form) RETURN    ! Closing parenthesis not the last.
      
      ! Here we have tlist being something like ( xxx ).  It could be either 
      ! form - we'll take it as the second form.
      first_form = .FALSE.
      RETURN
    END IF
    
    ! Mismatched parenthesis and no comma here.  We'll assume the second 
    ! form.
    first_form = .FALSE.
    
  END FUNCTION TestPositionStmtFirstForm
  
  
  !*****************************************************************************
  !!
  !> Write a text representation of a PositionSpecList to a logical unit.
  !!
  !! @param[in]     log_unit          Logical unit connected for formatted 
  !! output for the text represenation.
  !!
  !! @param[in]     spec              The /io-control-spec-list/.
  !!
  !! @param[in]     indent            Left hand side indent in characters.
  !!
  !! Unlike many of the other Write* routines in this module this procedure 
  !! uses non-advancing output and writes one item per record.
  
  SUBROUTINE DumpPositionSpecList(log_unit, spec, indent)
    
    USE Expressions
    USE Variables
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(PositionSpecList), INTENT(IN) :: spec
    INTEGER, INTENT(IN) :: indent
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Indent prefix.
    CHARACTER(indent) :: prefix
    
    !***************************************************************************
    
    prefix = ''   ! Fixed length string - set to all blanks.
    
    IF (ALLOCATED(spec%unit)) THEN
      WRITE (log_unit, "(A,'UNIT = ')", ADVANCE='NO') prefix
      CALL WriteExpr(log_unit, spec%unit)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%iomsg)) THEN
      WRITE (log_unit, "(A,'IOMSG = ')", ADVANCE='NO') prefix
      CALL WriteVariable(log_unit, spec%iomsg)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%iostat)) THEN
      WRITE (log_unit, "(A,'IOSTAT = ')", ADVANCE='NO') prefix
      CALL WriteVariable(log_unit, spec%iostat)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%err)) THEN
      WRITE (log_unit, "(A,'ERR = ',I0)") prefix, .ValueOf. spec%err
    END IF
    
  END SUBROUTINE DumpPositionSpecList
  
  
  !*****************************************************************************
  !!
  !> Implementation of PositionSpecList%Write - write a text representation of 
  !! the parse node to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE psl_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PositionSpecList), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Indicates that a particular spec is the first.  This is used for comma 
    ! output control.
    LOGICAL :: first
    
    !***************************************************************************
    
    ! This is set to false in the do_sep internal procedure.
    first = .TRUE.
    
    IF (ALLOCATED(pnode%unit)) THEN
      CALL do_sep
      
      WRITE (unit, "('UNIT = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%unit)
    END IF
    
    IF (ALLOCATED(pnode%iomsg)) THEN
      CALL do_sep
      
      WRITE (unit, "('IOMSG = ')", ADVANCE='NO')
      CALL WriteVariable(unit, pnode%iomsg)
    END IF
    
    IF (ALLOCATED(pnode%iostat)) THEN
      CALL do_sep
      
      WRITE (unit, "('IOSTAT = ')", ADVANCE='NO')
      CALL WriteVariable(unit, pnode%iostat)
    END IF
    
    IF (ALLOCATED(pnode%err)) THEN
      CALL do_sep
      
      WRITE (unit, "('ERR = ',I0)", ADVANCE='NO') .ValueOf. pnode%err
    END IF
    
  CONTAINS
    
    ! Test/sets the first flag, and if necessary outputs the comma that 
    ! separates list items.
    SUBROUTINE do_sep
      IF (first) THEN
        first = .FALSE.
      ELSE
        WRITE (unit, "(', ')", ADVANCE='NO')
      END IF
    END SUBROUTINE do_sep
  END SUBROUTINE psl_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of PositionSpecList%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE psl_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(PositionSpecList), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%unit)) CALL pnode%unit%Query(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Query(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Query(visitor)
    ! err.
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE psl_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of PositionSpecList%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE psl_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(PositionSpecList), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%unit)) CALL pnode%unit%Modify(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Modify(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Modify(visitor)
    ! err.
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE psl_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of PositionSpecList%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION psl_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'position-spec-list'
    
  END FUNCTION psl_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of PositionSpecList%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION psl_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(PositionSpecList), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = pnode%location
    
  END FUNCTION psl_GetLocation
  
END MODULE PositionSpecs
