! $Id: ProgramStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ProgramStmts module.


!*******************************************************************************
!!
!> Procedures for parsing program statements (/program-stmt/'s).
!!
!! A /program-stmt/ is defined by R1102 to be:
!!
!! PROGRAM /program-name/
!!
!! An /end-program-stmt/ is defined by R1103 to be:
!!
!! END [ PROGRAM [ /program-name/ ] ]
!!

MODULE ProgramStmts
  
  USE Identifiers
  USE ParseNodes
  USE ProgramStmtsUtils
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  ! Expose module procedures and interfaces
  
  PUBLIC :: GetProgramName        ! From ProgramStmtsUtils
  
  PUBLIC :: ParseProgramStmt
  PUBLIC :: ParseEndProgramStmt
  
  !> Stores the results of parsing a program statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: ProgramStmt
    !> The name of the program.
    TYPE(Identifier) :: name
  CONTAINS
    PROCEDURE :: Dump => program_Dump
    PROCEDURE :: Query => program_Query
    PROCEDURE :: Modify => program_Modify
    PROCEDURE, NOPASS :: What => program_What
    PROCEDURE :: Advertise => program_Advertise
  END TYPE ProgramStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /program-stmt/ (R1102).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    name              The name specified on the 
  !! /program-stmt/.  Not allocated on error.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseProgramStmt(tlist, prog_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ProgramStmt), INTENT(OUT) :: prog_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /program-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1102'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /program-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! PROGRAM ...
    CALL ExpectKeyword(tlist, 1, scck_'PROGRAM', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! name
    CALL ExpectName(tlist, 2, '/program-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    prog_stmt%name = IdentifierIfortBug(tlist(2))
    
    IF (SIZE(tlist) > 2) THEN
      CALL Add( err_list, CODE=errUnexpectedExtraTokens,  &
          COMPONENT=comp,  LOCATION=QueryLocation(tlist(3)),  &
          MSG='Unexpected tokens following '  &
            // QueryErrorText(tlist(2)) // '.' )
    END IF
    
  END SUBROUTINE ParseProgramStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-program-stmt/ (R1103).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     name              The /program-name/ provided on the 
  !! opening /program-stmt/, or not allocated if there was no /program-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndProgramStmt(tlist, name, err_list)
    
    USE CharUtils
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /end-program-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1103'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-program-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! END [ PROGRAM [ name ] ] or ENDPROGRAM [ name ]
    
    CALL ExpectKeyword(  &
        tlist,  &
        1,  &
        [scck_'END       ', scck_'ENDPROGRAM'],  &
        '"END" or "ENDPROGRAM"',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      ! PROGRAM could be next, or the statement could be an END on its own.
      CALL Expect(  &
          tlist,  &
          2,  &
          [ittName, ittEndStatement],  &
          [scck_'PROGRAM', scck_'       '],  &
          opt,  &
          err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      IF (opt == 2) RETURN    ! end-of-statement - we are done.
      it = 2
    ELSE
      it = 1
    END IF
    ! Name is optional.
    IF (SIZE(tlist) == it) RETURN
    
    it = it + 1
    
    CALL ExpectName(tlist, it, '/program-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ALLOCATE(name)
    name = IdentifierIfortBug(tlist(it))
    
    CALL ExpectEOS(tlist, it + 1, comp, err_list)
    
  END SUBROUTINE ParseEndProgramStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProgramStmt%Dump - write a debugging 
  !! representation of the results of parsing the program statement 
  !! to a logical unit.
  !!
  !! @param[in]     pnode             The FunctionSubprogram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  ! Recursive because function subprograms can have internal functions.
  RECURSIVE SUBROUTINE program_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProgramStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/program-name/ = ',A)") pfx, .NameOf. pnode%name
    
  END SUBROUTINE program_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProgramStmt%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE program_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProgramStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE program_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProgramStmt%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE program_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProgramStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE program_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProgramStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'function-subprogram'
  
  FUNCTION program_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'program-stmt'
    
  END FUNCTION program_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProgramStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  
  SUBROUTINE program_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProgramStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        .NameOf. pnode%name,  &
        [NatureDefMainProgram],  &
        [.Locationof. pnode%name] )
    
  END SUBROUTINE program_Advertise
  
END MODULE ProgramStmts
