! $Id: SubmoduleStmts.f90 2385 2016-08-24 05:41:29Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the SubmoduleStmts module.


!*******************************************************************************
!!
!> Procedures for parsing submodule statements (/submodule-stmt/'s).
!!
!! R1117: SUBMODULE ( /parent-identifier/ ) /submodule-name/
!!
!! R1118: /parent-identifier/ is /ancestor-module-name/ 
!!                                  [ : /parent-submodule-name/ ]
!!
!! Some terminology:
!! - the ancestor name is the parent module of the submodule - use 
!!   GetAncestorName to extract this from the submodule-stmt.
!! - the parent identifier is as described above - it is a colon separated 
!!   list of the ancestor module followed by the parent submodule of 
!!   the particular submodule.  Use GetParentIdentifier to extract this 
!!   from the submodule-stmt.
!! - the identifier of the module is the ancestor module name appended with 
!!   another separating colon and the name of the submodule - it is the 
!!   parent-identifier for any children of the submodule - handy for 
!!   building dependency trees.  Use GetIdentifier to extract this from 
!!   the submodule-stmt.
!!
!! The Get* procedures are intended to be called from early parsing code and 
!! do not validate all of the respective syntax rules.  The Parse procedure 
!! carries out the complete syntax parse.
!!
!! There's some duplication of logic in this module between the two groups 
!! of procedures, but you get that.

MODULE SubmoduleStmts
  
  USE Identifiers
  USE ParseNodes
  USE SubmoduleStmtsUtils
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  ! Expose module procedures and interfaces
  
  PUBLIC :: ParseSubmoduleStmt
  PUBLIC :: ParseEndSubmoduleStmt
  
  !> Stores the results of parsing a /submodule-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: SubmoduleStmt
    !> The name of the submodule itself.
    TYPE(Identifier) :: submodule_name
    
    !> The name of the submodule's ancestor module.
    TYPE(Identifier) :: ancestor_module_name
    
    !> The name of the parent submodule.
    !!
    !! Not allocated if the submodule had no parent submodule.
    TYPE(Identifier), ALLOCATABLE :: parent_submodule_name
  CONTAINS
    PROCEDURE :: Dump => sms_Dump
    PROCEDURE :: Query => sms_Query
    PROCEDURE :: Modify => sms_Modify
    PROCEDURE, NOPASS :: What => sms_What
    PROCEDURE :: Advertise => sms_Advertise
  END TYPE SubmoduleStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /submodule-stmt/ (R1117).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    ancestor          Name of the ancestor module.
  !!
  !! @param[out]    parent            Name of the parent submodule.  Not 
  !! allocated if no parent submodule is specified.
  !!
  !! @param[out]    name              Name of this submodule.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! @todo Are we forgetting the complete parent identifer in all of this?
  
  SUBROUTINE ParseSubmoduleStmt(tlist, submod_stmt, err_list)
    
    USE Strings
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(SubmoduleStmt), INTENT(OUT) :: submod_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /submodule-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1117'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing parenthesis.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /submodule-stmt/.' )
      RETURN
    END IF
      
    !---------------------------------------------------------------------------
    ! SUBMODULE...
    
    CALL ExpectKeyword(tlist, 1, scck_'SUBMODULE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The /parent-identifier/, in parens.
    
    ! SUBMODULE(...)
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Pulled in from SubmoduleStmtsUtils (parent identifier extraction uses 
    ! the same procedure).
    CALL ParseParentIdentifier(  &
        tlist,  &
        3, closing - 1,  &
        submod_stmt%ancestor_module_name,  &
        submod_stmt%parent_submodule_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! submodule name and then we're done!
    
    CALL ExpectName(tlist, closing + 1, '/submodule-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    submod_stmt%submodule_name = IdentifierIfortBug(tlist(closing+1))
    
    CALL ExpectEnd(tlist, closing + 2, comp, err_list)
    
  END SUBROUTINE ParseSubmoduleStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /parent-identifier/ (R1118).
  !!
  !! @param[in]     tlist             The sequence of tokens for the statement.
  !!
  !! @param[in]     start             Index in @a tlist of the first token 
  !! of the /parent-identifier/.
  !!
  !! @param[in]     finish            Index in @a tlist of the last token 
  !! of the /parent-identifier/.
  !!
  !! @param[out]    ancestor          The name of the ancestor module.
  !!
  !! @param[out]    parent            The name of the parent submodule 
  !! if one is specified, or not allocated if one is not specified.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseParentIdentifier( tlist, start, finish, ancestor,  &
      parent, err_list )
    
    USE CompilerKinds
    USE Errors
    USE Strings
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(Identifier), INTENT(OUT) :: ancestor
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: parent
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /parent-identifier/.
    CHARACTER(*), PARAMETER :: comp = 'R1118'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The ancestor module
    
    ! Everywhere else we have a list of something the language uses a comma.  
    ! Here we use a colon.  Go figure.
    
    ! There must be at least one - for the ancestor module.
    CALL ExpectName(tlist, start, '/ancestor-module-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ancestor = IdentifierIfortBug(tlist(start))
    
    !---------------------------------------------------------------------------
    ! The : /parent-submodule-name/ is optional.
    
    IF (finish == start) RETURN
    
    CALL Expect(tlist, start + 1, ittColon, scck_':', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectName(tlist, start + 2, '/parent-submodule-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !===========================================================================
    ! ifort 15.0.2 bug?
    !parent = IdentifierIfortBug(tlist(start + 2))
    !===========================================================================
    ALLOCATE(parent, SOURCE=IdentifierIfortBug(tlist(start + 2)))
    !===========================================================================
    
    CALL ExpectEnd(tlist(:finish), start + 3, comp, err_list)
    
  END SUBROUTINE ParseParentIdentifier
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-submodule-stmt/ (R1119).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     name              The /submodule-name/ from the statement, 
  !! or not allocated if there was no name.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndSubmoduleStmt(tlist, name, err_list)
    
    USE CompilerKinds
    USE Identifiers
    USE Errors
    USE EndProgramUnitStmts
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL ParseEndProgramUnitStmt(  &
        tlist,  &
        scck_'SUBMODULE',  &
        'R1119',  &
        '/end-submodule-stmt/',  &
        '/submodule-name/',  &
        name,  &
        err_list )
    
  END SUBROUTINE ParseEndSubmoduleStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubmoduleStmt%Dump - write a debugging 
  !! representation of the results of parsing the separate module procedure 
  !! to a logical unit.
  !!
  !! @param[in]     pnode             The FunctionSubprogram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  ! As separate module procedures cannot contain separate module procedures, 
  ! this doesn't need to be recursive.
  SUBROUTINE sms_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SubmoduleStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'ancestor ',A)") pfx, .NameOf. pnode%ancestor_module_name
    IF (ALLOCATED(pnode%parent_submodule_name)) THEN
        WRITE (unit, "(A,'parent ',A)") pfx,  &
            .NameOf. pnode%parent_submodule_name
    END IF
    WRITE (unit, "(A,'/submodule-name/ = ',A)") pfx,  &
        .NameOf. pnode%submodule_name
    
  END SUBROUTINE sms_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubmoduleStmt%Query - traverse the node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE sms_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SubmoduleStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sms_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubmoduleStmt%Modify - traverse the node 
  !! with a modify operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE sms_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SubmoduleStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sms_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubmoduleStmt%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'submodule-stmt'
  
  FUNCTION sms_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'submodule-stmt'
    
  END FUNCTION sms_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubmoduleStmt%Advertise - advertise the 
  !! identifiers referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  
  SUBROUTINE sms_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubmoduleStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%parent_submodule_name)) THEN
      ALLOCATE(ident_list(3))
    ELSE
      ALLOCATE(ident_list(2))
    END IF
    
    ! The submodule.
    ident_list(1) = IdentifierNature(  &
        .NameOf. pnode%ancestor_module_name // ':'  &
          // .NameOf. pnode%submodule_name,  &
        [NatureDefSubmodule],  &
        [.Locationof. pnode%ancestor_module_name] )
    
    ! The parent module.
    ident_list(2) = IdentifierNature(  &
        .NameOf. pnode%ancestor_module_name,  &
        [NatureRefModule],  &
        [.LocationOf. pnode%ancestor_module_name] )
    
    IF (ALLOCATED(pnode%parent_submodule_name)) THEN
      ! The parent submodule.
      ident_list(3) = IdentifierNature(  &
          .NameOf. pnode%ancestor_module_name // ':'  &
            // .NameOf. pnode%parent_submodule_name,  &
          [NatureRefSubmodule],  &
          [.LocationOf. pnode%ancestor_module_name] )
    END IF
    
  END SUBROUTINE sms_Advertise
  
END MODULE SubmoduleStmts
