! $Id: SubscriptLists.f90 2858 2019-03-29 20:31:00Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the SubscriptLists module


!*******************************************************************************
!!
!> Type for storing the result of parsing a subscript lists.
!!
!! *Spec things are those that appear in declarations.  
!! ParseSectionSubscriptList parses things that appear in expressions.
!!
!! We provide routines for parsing the specific types of array specifications, 
!! even though a sequential parse may not know what the appropriate array 
!! specification is for an entity or declaration attribute.  For example, 
!! if an entity is allocatable or a pointer then the array specification 
!! must be /deferred-shape-spec-list/, but it is possible for the 
!! DIMENSION attribute with that specification to appear in a completely 
!! separate statement to where the POINTER or ALLOCATABLE attribute is 
!! specified.
!!
!! In the general case the ParseArraySpec procedure can be used.  At a higher 
!! level than statement parsing the Is*SpecList functions can be used 
!! to confirm that the array-spec is valid.
!!
!! @todo Some parts of this were written early on, and could do with a 
!! review.  Comma separated list handling is out of date.
!!
!! Note that corrigenda III changed things for implied shape specs and 
!! assumed size specs slightly.

MODULE SubscriptLists
  
  USE Expressions
  USE ParseNodes
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  !----------------------------------------------------------------------------- 
  
  PUBLIC :: Move
  PUBLIC :: ParseArraySpec
  PUBLIC :: ParseCoarraySpec
  PUBLIC :: ParseSectionSubscriptList
  
  PUBLIC :: IsDeferredShapeSpecList
  PUBLIC :: IsAssumedShapeSpecList
  PUBLIC :: IsAssumedSizeSpec
  PUBLIC :: IsImpliedShapeSpec
  PUBLIC :: IsExplicitShapeSpecList
  PUBLIC :: IsDeferredCoshapeSpecList
  PUBLIC :: ClassifyArraySpec
  
  PUBLIC :: ParseExplicitShapeSpecList
  PUBLIC :: ParseExprBasedShapeSpecList
  PUBLIC :: ParseDeferredShapeSpecList
  PUBLIC :: ParseDeferredCoshapeSpecList
  PUBLIC :: ParseAllocateCoarraySpec
  
  PUBLIC :: VerifyAssumedShapeSpecList
  PUBLIC :: VerifyDeferredShapeSpecList
  PUBLIC :: VerifyExplicitShapeSpecList
  PUBLIC :: VerifyImpliedShapeSpec
  PUBLIC :: VerifyAssumedSizeSpec
  
  PUBLIC :: IsSubstringRangeLike
  PUBLIC :: IsNotFunctionArgList
  
  PUBLIC :: WriteDimSpec
  
  !----------------------------------------------------------------------------- 
  
  !> Represents the specification of a single dimension, or a single 
  !! subscript in a subscript list or a single section-subscript.  For 
  !! a dimension specification this can be for an array or a coarray.
  !!
  !! See R515 for a list of what can be in an array spec in the dimension 
  !! attribute.
  !!
  !! Some dimension attribute declaration examples (using a single rank 
  !! array for clarity):
  !!
  !! - "(a)" (R516: explicit-shape-spec with no lower bound): finish is a, 
  !! everything else .FALSE. or not allocated.
  !!
  !! - "(a:b)" (R516: explicit-shape-spec with both lower and upper bounds): 
  !! start is a, colon is .TRUE., finish is b, everything else .FALSE. or 
  !! not allocated.
  !!
  !! - "(:)" (R519: assumed-shape-spec with no lower bound, or R520: deferred 
  !! shape spec): colon is .TRUE., everything else .FALSE. or not allocated.
  !!
  !! - "(a:)" (R519: assumed-shape-spec with lower bound): start is a, colon 
  !! is .TRUE., everything else .FALSE. or not allocated.
  !!
  !! - "(*)" (R521: assumed-size-spec with no lower bound): star is .TRUE., 
  !! everything else is .FALSE. or not allocated.  As of corrigenda 
  !! three this is now considered an /assumed-implied-spec/.
  !!
  !! - "(a:*)" (R521: assumed-size-spec with lower bound): star is .TRUE., 
  !! start is a, everything else is .FALSE. or not allocated.  As of 
  !! corrigenda three this is now considered an /assumed-implied-spec/.
  !!
  !! See R509 for a list of what can be in a coarray-spec.  Some examples: 
  !!
  !! - "(:)" (R510: deferred-coshape-spec): colon is .TRUE., everything 
  !! else .FALSE. or not allocated.
  !!
  !! - "(a:b)" (R511: explicit-coshape-spec with lower bound): colon is 
  !! .TRUE., start is a, finish is b, everything else .FALSE. or unallocated.
  !!
  !! - "(a)" (R511: explicit-coshape-spec with no lower bound): finish is 
  !! a, everything else .FALSE. or unallocated.
  !!
  !! - "(*)" (R511: explicit-coshape-spec, last dimension): star is .TRUE., 
  !! everything else is .FALSE. or unallocated.
  !! 
  !! Some data-ref examples, where we are storing a section-subscript-list: 
  !!
  !! - "(a)" (R619 if a is scalar, R623 if not): finish is a, everything 
  !! else .FALSE. or not allocated.
  !!
  !! - "(:)" (R621 without any subscripts): colon is .TRUE., everything else 
  !! .FALSE. or not allocated.
  !!
  !! - "(:a)" (R621 without the start or stride subscripts): colon is .TRUE., 
  !! finish is a, everything else .FALSE. or not allocated.
  !!
  !! - "(a:)" (R621 without the finish or stride subscripts): colon is .TRUE., 
  !! start is a, everything else .FALSE. or not allocated.
  !!
  !! - "(a:b)" (R621 without the stride subscripts): colon is .TRUE., 
  !! start is a, finish is b, everything else .FALSE. or not allocated.
  !!
  !! - "(a:b:c)" (R621 with everything): colon is .TRUE., start is a, 
  !! finish is b, stride is c, star (the only thing left) is .FALSE..
  !!
  !! - "(::c)" (R621 with only the stride subscript): colon is .TRUE., 
  !! stride is c, everything else is .FALSE. or not allocated.
  !!
  !! Routines for parsing DimSpec's (which have accidentally also been called 
  !! "Subscripts") are in the ParseExpressions module until submodule support 
  !! is available.
  !!
  !! If additional components are added then the Move_ procedure needs to be 
  !! updated.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: DimSpec
    !> The dimension specification contained at least one colon.
    LOGICAL :: colon   ! :
    
    !> The location of the colon.
    !!
    !! Only allocated if colon is true.
    TYPE(SourceLocation), ALLOCATABLE :: colon_location
    
    !> The specification contained a star.
    !!
    !! This is only relevant for dimension specifications.
    LOGICAL :: star    ! *
    
    !> The location of the star.
    !!
    !! Only allocated if star is true.
    TYPE(SourceLocation), ALLOCATABLE :: star_location
    
    !> The start of a subscript-triplet, or the lower bound of a dimension 
    !! attribute in a declaration.
    !!
    !! For a triplet, if this is not allocated then the subscript-triplet 
    !! starts from the start of the array.  For an assumed shape declaration 
    !! the starting index is one.
    CLASS(Expression), ALLOCATABLE :: start
    
    !> A scalar subscript, a vector subscript, the end of a subscript triplet 
    !! or the upper bound or extend of a rank declaration.
    !!
    !! For a triplet, if this is not allocated then the subscript-triplet 
    !! ends at the end of the array.  
    !!
    !! This must be allocated for an explicit-shape declaration.  This must 
    !! not be allocated for assumed size declaration in the last dimension.
    CLASS(Expression), ALLOCATABLE :: finish
    
    !> The stride in a subscript-triplet.
    !!
    !! If not allocated then the stride is one.  If allocated then start and 
    !! finish refer to a triplet.  If allocated then there will also be 
    !! a colon (so the colon component will be true).
    CLASS(Expression), ALLOCATABLE :: stride
  CONTAINS
    PROCEDURE :: Write => ds_Write
    PROCEDURE :: Query => ds_Query
    PROCEDURE :: Modify => ds_Modify
    PROCEDURE, NOPASS :: What => ds_What
    PROCEDURE :: Advertise => ds_Advertise
    PROCEDURE :: GetLocation => ds_GetLocation
  END TYPE DimSpec
  
  !-----------------------------------------------------------------------------
  ! Interfaces
  
  INTERFACE Move
    MODULE PROCEDURE Move_
  END INTERFACE Move
  
  !> Parse an array-spec.
  INTERFACE ParseArraySpec
    !> Specific form for array-spec-lists of various sorts.
    MODULE PROCEDURE ParseArraySpec_list
    !> Specific form for a single spec in a array-spec list of some sort.
    MODULE PROCEDURE ParseArraySpec_single
  END INTERFACE ParseArraySpec
  
  !> Parse a coshape-spec of some description.
  INTERFACE ParseCoarraySpec
    MODULE PROCEDURE ParseCoshapeSpec_list
    MODULE PROCEDURE ParseCoshapeSpec_single
  END INTERFACE ParseCoarraySpec
  
  !> Is it a coshape-spec or a coarray-spec?  We couldn't decide!
  INTERFACE ParseCoshapeSpec
    MODULE PROCEDURE ParseCoshapeSpec_list
    MODULE PROCEDURE ParseCoshapeSpec_single
  END INTERFACE ParseCoshapeSpec
  
  !-----------------------------------------------------------------------------
  ! Spec types returned by ClassifyArraySpec
  
  !> We don't know what sort of spec we have.
  INTEGER, PARAMETER :: iasNone = 0
  !> We have an explicit shape spec.
  INTEGER, PARAMETER :: iasExplicitShape = 1
  !> We have a deferred shape spec.
  INTEGER, PARAMETER :: iasDeferredShape = 2
  !> We have an assumed shape spec.
  INTEGER, PARAMETER :: iasAssumedShape = 3
  !> We have an assumed size spec.
  INTEGER, PARAMETER :: iasAssumedSize = 4
  !> We have an implied shape spec.
  INTEGER, PARAMETER :: iasImpliedShape = 5
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Move a DimSpec from one variable to another.
  !!
  !! @param[in,out] from              The variable to move information from.
  !!
  !! @param[out]    to                The entity to receive the contents of 
  !! @a from.
  !!
  !! @a from needs to be defined again before it can be re-used.
  
  SUBROUTINE Move_(from, to)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DimSpec), INTENT(INOUT) :: from
    TYPE(DimSpec), INTENT(OUT) :: to
    
    !***************************************************************************
    
    to%colon = from%colon
    CALL MOVE_ALLOC(from%colon_location, to%colon_location)
    to%star = from%star
    CALL MOVE_ALLOC(from%star_location, to%star_location)
    CALL MOVE_ALLOC(from%start, to%start)
    CALL MOVE_ALLOC(from%finish, to%finish)
    CALL MOVE_ALLOC(from%stride, to%stride)
    
  END SUBROUTINE Move_
  
  
  !*****************************************************************************
  !!
  !> Parse an /array-spec/ (R515) or a /component-array-spec/ (R439).
  !!
  !! An /array-spec/ is the form used in /type-declaration-stmt/'s.  It can 
  !! use "*" in the list for a dimension to mean assumed-size and items 
  !! in the list cannot have a stride.
  !!
  !! A /component-array-spec/ excludes the array spec varieties with a *.
  !! It also excludes /assumed-shape-spec/'s, but these are not generally 
  !! distinguishable from /deferred-shape-spec/'s without more context.
  !!
  !! @param[out]    ss_list           The list of subscripts.
  !!
  !! @param[in]     tlist             The list of tokens that make up the 
  !! subscript list.  The first token should be the token that comes 
  !! immediately after the opening parentheses, and the last token should be 
  !! the token immediately before the closing parentheses.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseArraySpec_list(tlist, component, ss_list, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    LOGICAL, INTENT(IN) :: component
    TYPE(DimSpec), INTENT(OUT), ALLOCATABLE :: ss_list(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the /implicit-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    INTEGER :: i              ! Current token
    
    ! Error lists from procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      IF (component) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='R439',  &
            MSG='Expected /component-array-spec/.' )
      ELSE
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='R515',  &
            MSG='Expected /array-spec/.' )
      END IF
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist)
    ALLOCATE(ss_list(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 1, SIZE(tlist), item_start, item_finish)
      CALL ParseArraySpec( tlist(item_start:item_finish), ss_list(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=my_comp())
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    !---------------------------------------------------------------------------
    ! Check that we've not got something daft like (:,*) or (*,*)
    
    IF (component) THEN
      ! /assumed-size-spec/ and /implied-shape/ (the forms that use *) are
      ! not permitted.  Also /assumed-shape-spec/ is not permitted - this 
      ! means that any spec with a colon must have both start and finish 
      ! or neither start and finish.
      DO i = SIZE(ss_list), 1, -1
        IF (ss_list(i)%star) THEN
          CALL Add( err_list, CODE=errSyntax, COMPONENT='R439',  &
  !            LOCATION=???        
              MSG='This is not a valid specification for a &
                &/component-array-spec/.' )
          RETURN
        END IF
        IF (ss_list(i)%colon .AND. ( ALLOCATED(ss_list(i)%start)  &
            .NEQV. ALLOCATED(ss_list(i)%finish) )) THEN
          CALL Add( err_list, CODE=errSyntax, COMPONENT='R439',  &
  !            LOCATION=???        
              MSG='This is not a valid specification for a &
                &/component-array-spec/.' )
          RETURN
        END IF
        IF (ALLOCATED(ss_list(i)%stride)) THEN
          CALL Add( err_list, CODE=errSyntax, COMPONENT='R439',  &
  !            LOCATION=???        
              MSG='This is not a valid specification for a &
                &/component-array-spec/.' )
          RETURN
        END IF
      END DO
    ELSE
      ! No array specification permits a stride.
      IF (ANY(has_stride(ss_list))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT='R515',  &
!            LOCATION=???        
            MSG='A stride is not permitted in an /array-spec/.' )
        RETURN
      END IF
      
      IF (ANY(ss_list%star)) THEN
        ! Things with stars.
        !
        ! Either only the last must be a star, or all must have a star.  If 
        ! everything but the last has a finish expression, then we know the 
        ! last is a star because of the ANY test above.
        IF (.NOT. ( ALL(ss_list%star)  &
            .OR. ALL(has_finish(ss_list(:SIZE(ss_list)-1))) )) THEN
          CALL Add( err_list, CODE=errSyntax, COMPONENT='R515',  &
!              LOCATION=???, 
              MSG='Either all specifications must be &
                &an /assumed-implied-spec/, or only the last specification &
                &must be an /assumed-implied-spec/ with earlier &
                &specifications to be an /explicit-shape-spec/.' )
          RETURN
        END IF
      ELSE
        ! Things with no stars.
        IF (ANY(.NOT. has_finish(ss_list))) THEN
          ! If no finish specified on one spec, then no finish can be 
          ! specified on any specs.
          IF (ANY(has_finish(ss_list))) THEN
            CALL Add( err_list, CODE=errSyntax, COMPONENT='R515',  &
!               LOCATION=???
                MSG='An upper bound cannot be specified in a &
                  &/deferred-shape-spec-list/ or an &
                  &/assumed-shape-spec-list/.' )
            RETURN
          END IF
        END IF
      END IF
      
      
    END IF
    
  CONTAINS
  
    !***************************************************************************
    
    FUNCTION my_comp()
      CHARACTER(4) :: my_comp
      !*************************************************************************
      IF (component) THEN
        my_comp = 'R439'
      ELSE
        my_comp = 'R515'
      END IF
    END FUNCTION my_comp
  
  END SUBROUTINE ParseArraySpec_list
  
  
  !*****************************************************************************
  !!
  !> Classifies an array specification to its likely type.
  !!
  !! @param[in]     ss_list           The array specification.
  !!
  !! @param[out]    spec_type         Inferred type of the specification.
  
  SUBROUTINE ClassifyArraySpec(ss_list, spec_type)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DimSpec), INTENT(IN) :: ss_list(:)
    INTEGER, INTENT(OUT) :: spec_type
    
    !***************************************************************************
    
    IF (IsAssumedSizeSpec(ss_list)) THEN
      spec_type = iasAssumedSize
      RETURN
    END IF
    
    IF (IsImpliedShapeSpec(ss_list)) THEN
      spec_type = iasImpliedShape
      RETURN
    END IF
    
    IF (IsDeferredShapeSpecList(ss_list)) THEN
      spec_type = iasDeferredShape
      RETURN
    END IF
    
    IF (IsAssumedShapeSpecList(ss_list)) THEN
      spec_type = iasAssumedShape
      RETURN
    END IF
    
    IF (IsExplicitShapeSpecList(ss_list)) THEN
      spec_type = iasExplicitShape
      RETURN
    END IF
    
    spec_type = iasNone
    
  END SUBROUTINE ClassifyArraySpec
  
  
  !*****************************************************************************
  !!
  !> Tests whether a array specification for a single dimension is an 
  !! /assumed-implied-spec/.
  !!
  !! @param[in]   spec                The array spec to test.
  !!
  !! @returns True if the array spec matches the pattern 
  !! `[ /lower-bound/ :] *`, false otherwise.
  !!
  !! This is a syntax term introduced with corrigendum three.
  
  ELEMENTAL FUNCTION IsAssumedImpliedSpec(spec) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DimSpec), INTENT(IN) :: spec
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    ! Must have a star.
    b = spec%star
    IF (.NOT. b) RETURN
    
    ! Don't care whether it has a lower bound.  If it has a lower bound, 
    ! parsing code should ensure that it has the necessary colon.
    
  END FUNCTION IsAssumedImpliedSpec
  
  
  !*****************************************************************************
  !!
  !> Tests whether an array specification is a /assumed-size-spec/.
  !!
  !! If there is only one dimension specification, then this could also be 
  !! an /implied-shape-spec/.
  !!
  !! @param[in]     ss_list           The array spec to test.  This should
  !! have been previously constructed using ParseArraySpec.
  !
  !! @returns .TRUE. If the array spec is suitable as a 
  !! /assumed-size-spec/, .FALSE. otherwise.
  
  PURE FUNCTION IsAssumedSizeSpec(ss_list) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DimSpec), INTENT(IN) :: ss_list(:)
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    ! The first part of the array spec must look like an explicit shape spec.
    IF (SIZE(ss_list) > 1) THEN
      b = IsExplicitShapeSpecList(ss_list(:SIZE(ss_list)-1))
      IF (.NOT. b) RETURN
    END IF
    
    ! The last spec must contain a * and an optional lower bound, but no 
    ! upper bound or stride.
    b = IsAssumedImpliedSpec(ss_list(SIZE(ss_list)))
    
  END FUNCTION IsAssumedSizeSpec
  
  
  !*****************************************************************************
  !!
  !> Tests whether an array specification is a /implied-shape-spec/.
  !!
  !! If there is only one element in the specification, then an 
  !! /implied-shape-spec/ can also be a /assumed-size-spec/.
  !!
  !! @param[in]     ss_list           The array spec to test.  This should
  !! have been previously constructed using ParseArraySpec.
  !
  !! @returns .TRUE. If the array spec is suitable as a 
  !! /implied-shape-spec-list/, .FALSE. otherwise.
  !!
  !! As published, F2008 called this an /implied-shape-spec-list/.  This was 
  !! changed with corrigenda three.
  
  PURE FUNCTION IsImpliedShapeSpec(ss_list) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DimSpec), INTENT(IN) :: ss_list(:)
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    ! All specs must be of the form `[lower :] *`
    b = ALL(IsAssumedImpliedSpec(ss_list))
    
  END FUNCTION IsImpliedShapeSpec
  
  
  !*****************************************************************************
  !!
  !> Tests whether an array specification is a /deferred-shape-spec-list/.
  !!
  !! @param[in]     ss_list           The array spec to test.  This should
  !! have been previously constructed using ParseArraySpec.
  !
  !! @returns .TRUE. If the array spec is suitable as a 
  !! /deferred-shape-spec-list/, .FALSE. otherwise.  Note that 
  !! /deferred-shape-spec-list/ array specifications can also be 
  !! /assumed-shape-spec-list/ array specifications.
  
  PURE FUNCTION IsDeferredShapeSpecList(ss_list) RESULT(is_deferred)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DimSpec), INTENT(IN) :: ss_list(:)
    
    ! Function result
    LOGICAL :: is_deferred
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i
    
    !***************************************************************************
    
    is_deferred = .FALSE.
    
    DO i = 1, SIZE(ss_list)
      ! Must have a colon.
      IF (.NOT. ss_list(i)%colon) RETURN
      ! Must not have a star.
      IF (ss_list(i)%star) RETURN
      ! Must not have any lower, upper or stride expressions.
      IF (ALLOCATED(ss_list(i)%start)) RETURN
      IF (ALLOCATED(ss_list(i)%finish)) RETURN
    END DO
    
    is_deferred = .TRUE.
    
  END FUNCTION IsDeferredShapeSpecList
  
  
  !*****************************************************************************
  !!
  !> Tests whether an array specification is a /deferred-shape-spec-list/.
  !!
  !! @param[in]     ss_list           The array spec to test.  This should
  !! have been previously constructed using ParseArraySpec.
  !
  !! @returns .TRUE. If the array spec is suitable as a 
  !! /deferred-shape-spec-list/, .FALSE. otherwise.  Note that 
  !! /deferred-shape-spec-list/ array specifications can also be 
  !! /assumed-shape-spec-list/ array specifications.
  
  PURE FUNCTION IsAssumedShapeSpecList(ss_list) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DimSpec), INTENT(IN) :: ss_list(:)
    
    ! Function result
    LOGICAL :: b
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Spec index.
    
    !***************************************************************************
    
    b = .FALSE.
    
    DO i = 1, SIZE(ss_list)
      ! Must have a colon.
      IF (.NOT. ss_list(i)%colon) RETURN
      ! Must not have a star.
      IF (ss_list(i)%star) RETURN
      ! Must not have any upper or stride expressions.
      IF (ALLOCATED(ss_list(i)%finish)) RETURN
    END DO
    
    b = .TRUE.
    
  END FUNCTION IsAssumedShapeSpecList
  
  
  !*****************************************************************************
  !!
  !> Tests whether an array specification is an /explicit-shape-spec-list/.
  !!
  !! @param[in]     ss_list           The array spec to test.  This should
  !! have been previously constructed using ParseArraySpec.
  !
  !! @returns .TRUE. If the array spec is suitable as a 
  !! /explicit-shape-spec-list/, .FALSE. otherwise.
  
  PURE FUNCTION IsExplicitShapeSpecList(ss_list) RESULT(is_explicit)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DimSpec), INTENT(IN) :: ss_list(:)
    
    ! Function result
    LOGICAL :: is_explicit
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Array specification index.
    
    !***************************************************************************
    
    is_explicit = .FALSE.
    
    DO i = 1, SIZE(ss_list)
      IF (ss_list(i)%star) RETURN
      IF (.NOT. ALLOCATED(ss_list(i)%finish)) RETURN
      IF (ss_list(i)%colon .AND. (.NOT. ALLOCATED(ss_list(i)%start))) RETURN
    END DO
    
    is_explicit = .TRUE.
  
  END FUNCTION IsExplicitShapeSpecList
  
  
  !*****************************************************************************
  !!
  !> Tests whether an coarray specification is a /deferred-coshape-spec-list/.
  !!
  !! @param[in]     ss_list           The coarray spec to test.  This should
  !! have been previously constructed using ParseCoarraySpec.
  !
  !! @returns .TRUE. If the array spec is suitable as a 
  !! /deferred-coshape-spec-list/, .FALSE. otherwise.
  !!
  !! The syntax is the same as for a normal /deferred-shape-spec-list/, so 
  !! we just hand over to IsDeferredShapeSpecList.
  
  PURE FUNCTION IsDeferredCoshapeSpecList(ss_list) RESULT(is_deferred)
  
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DimSpec), INTENT(IN) :: ss_list(:)
    
    ! Function result
    LOGICAL :: is_deferred
    
    !***************************************************************************
    
    is_deferred = IsDeferredShapeSpecList(ss_list)
    
  END FUNCTION IsDeferredCoshapeSpecList
  
  
  !*****************************************************************************
  !!
  !> Verify that a previously parsed /array-spec/ is an 
  !! /assumed-shape-spec-list/.
  !!
  !! @param[in]     specs             The array spec.
  !!
  !! @param[in]     err_list          List of errors.
  
  SUBROUTINE VerifyAssumedShapeSpecList(specs, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DimSpec), INTENT(IN) :: specs(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Spec index.
    
    ! Syntax rule for an /assumed-shape-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R519'
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    DO i = 1, SIZE(specs)
      IF (has_finish(specs(i))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='An upper bound specification is not permitted in an &
              &/assumed-shape-spec/.' )
        CYCLE
      END IF
      
      IF (specs(i)%star) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='"*" is not permitted in an /assumed-shape-spec/.' )
        CYCLE
      END IF
      
      IF (has_stride(specs(i))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='A stride specification is not permitted in an &
              &/assumed-shape-spec/.' )
        CYCLE
      END IF
      
      IF (.NOT. specs(i)%colon) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='":" is required in an /assumed-shape-spec/.' )
      END IF
    END DO
    
  END SUBROUTINE VerifyAssumedShapeSpecList
  
  
  !*****************************************************************************
  !!
  !> Verify that a previously parsed /array-spec/ is an 
  !! /deferred-shape-spec-list/.
  !!
  !! @param[in]     specs             The array spec.
  !!
  !! @param[in]     err_list          List of errors.
  
  SUBROUTINE VerifyDeferredShapeSpecList(specs, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DimSpec), INTENT(IN) :: specs(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Spec index.
    
    ! Syntax rule for an /deferred-shape-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R520'
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    DO i = 1, SIZE(specs)
      IF (has_finish(specs(i))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='An upper bound specification is not permitted in a &
              &/deferred-shape-spec/.' )
        CYCLE
      END IF
      
      IF (specs(i)%star) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='"*" is not permitted in a /deferred-shape-spec/.' )
        CYCLE
      END IF
      
      IF (has_stride(specs(i))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='A stride specification is not permitted in a &
              &/deferred-shape-spec/.' )
        CYCLE
      END IF
      
      IF (.NOT. specs(i)%colon) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='":" is required in a /deferred-shape-spec/.' )
      END IF
      
      IF (has_start(specs(i))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='A lower bound specification is not permitted in a &
              &/deferred-shape-spec/.' )
      END IF
    END DO
    
  END SUBROUTINE VerifyDeferredShapeSpecList
  
  
  !*****************************************************************************
  !!
  !> Verify that a previously parsed /array-spec/ is an 
  !! /explicit-shape-spec-list/.
  !!
  !! @param[in]     specs             The array spec.
  !!
  !! @param[in]     err_list          List of errors.
  
  SUBROUTINE VerifyExplicitShapeSpecList(specs, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DimSpec), INTENT(IN) :: specs(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Spec index.
    
    ! Syntax rule for an /explicit-shape-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R516'
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    DO i = 1, SIZE(specs)
      IF (.NOT. has_finish(specs(i))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='An upper bound specification required in an &
              &/explicit-shape-spec/.' )
        CYCLE
      END IF
      
      IF (specs(i)%star) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='"*" is not permitted in an /explicit-shape-spec/.' )
        CYCLE
      END IF
      
      IF (has_stride(specs(i))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='A stride specification is not permitted in an &
              &/explicit-shape-spec/.' )
        CYCLE
      END IF
    END DO
    
  END SUBROUTINE VerifyExplicitShapeSpecList
  
  
  !*****************************************************************************
  !!
  !> Verify that a previously parsed /array-spec/ is an 
  !! /implied-shape-spec/.
  !!
  !! @param[in]     specs             The array spec.
  !!
  !! @param[in]     err_list          List of errors.
  
  SUBROUTINE VerifyImpliedShapeSpec(specs, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DimSpec), INTENT(IN) :: specs(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Spec index.
    
    ! Syntax rule for an /implied-shape-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R522'
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    DO i = 1, SIZE(specs)
      IF (has_finish(specs(i))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='An upper bound specification is not permitted in an &
              &/implied-shape-spec/.' )
        CYCLE
      END IF
      
      IF (.NOT. specs(i)%star) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='"*" is requried in an /implied-shape-spec/.' )
        CYCLE
      END IF
      
      IF (has_stride(specs(i))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='A stride specification is not permitted in an &
              &/implied-shape-spec/.' )
        CYCLE
      END IF
    END DO
    
  END SUBROUTINE VerifyImpliedShapeSpec
  
  
  !*****************************************************************************
  !!
  !> Verify that a previously parsed /array-spec/ is an 
  !! /implied-shape-spec/.
  !!
  !! @param[in]     specs             The array spec.
  !!
  !! @param[in]     err_list          List of errors.
  
  SUBROUTINE VerifyAssumedSizeSpec(specs, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DimSpec), INTENT(IN) :: specs(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Spec index.
    
    ! Syntax rule for an /assumed-size-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R521'
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    DO i = 1, SIZE(specs) - 1
      IF (.NOT. has_finish(specs(i))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='An upper bound specification is required here in an &
              &/assumed-size-spec/.' )
        CYCLE
      END IF
      
      IF (specs(i)%star) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='"*" is not permitted here in an /assumed-size-spec/.' )
        CYCLE
      END IF
      
      IF (has_stride(specs(i))) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            MSG='A stride specification is not permitted in an &
              &/assumed-size-spec/.' )
        CYCLE
      END IF
    END DO
    
    IF (has_finish(specs(i))) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='An upper bound specification is not permitted here in an &
            &/assumed-size-spec/.' )
    ELSE IF (.NOT. specs(i)%star) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='"*" is required here in an /assumed-size-spec/.' )
    ELSE IF (has_stride(specs(i))) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='A stride specification is not permitted in an &
            &/assumed-size-spec/.' )
    END IF
    
  END SUBROUTINE VerifyAssumedSizeSpec
  
  
  !*****************************************************************************
  !!
  !> Test whether an array spec has a stride.
  !!
  !! @param[in]     spec              The array spec to test.
  !!
  !! @return true if the spec has a stride.
  
  ELEMENTAL FUNCTION has_stride(spec) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DimSpec), INTENT(IN) :: spec
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = ALLOCATED(spec%stride)
    
  END FUNCTION has_stride
  
  
  !*****************************************************************************
  !!
  !> Test whether an array spec has a finish expression.
  !!
  !! @param[in]     spec              The array spec to test.
  !!
  !! @return true if the spec has a finish expression.
  
  ELEMENTAL FUNCTION has_finish(spec) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DimSpec), INTENT(IN) :: spec
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = ALLOCATED(spec%finish)
    
  END FUNCTION has_finish
  
  
  !*****************************************************************************
  !!
  !> Test whether an array spec has a finish expression.
  !!
  !! @param[in]     spec              The array spec to test.
  !!
  !! @return true if the spec has a finish expression.
  
  ELEMENTAL FUNCTION has_start(spec) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DimSpec), INTENT(IN) :: spec
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = ALLOCATED(spec%start)
    
  END FUNCTION has_start
  
  
  !*****************************************************************************
  !!
  !> Parse a single specified in an array-spec (R515).
  !!
  !! @param[out]    sub               The subscript (item in the array-spec
  !! for a particular dimension).
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[out]    sub_err_list      List of errors.
  !!
  !! Note that we never allocate @a sub%stride - as that is not a valid
  !! part of an array-spec.
  
  SUBROUTINE ParseArraySpec_single(tlist, sub, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DimSpec), INTENT(OUT) :: sub
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
        
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! List of errors from procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Position of the colon, if present.
    INTEGER :: part_end
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    ! We don't know what sort of /array-spec/ we are processing, therefore 
    ! we don't know what to expect for a particular item.
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          MSG='Expected an array shape specification.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
  
    ! See if we have a lower bound.  If we have a lower bound then there 
    ! will be a : in the tokens.
    part_end = FindBare(tlist, ittColon, scck_':')  
    IF (part_end /= 0) THEN
      ! We might have a lower bound
      sub%colon = .TRUE.
      sub%colon_location = QueryLocation(tlist(part_end))
      IF (part_end == 1) THEN
        ! No lower bound.  Might be assumed-shape-spec (R519) or 
        ! deferred-shape-spec (R520), in which case there is no 
        ! upper bound.
        IF (SIZE(tlist) > 1) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='R519 or R520',  &
              MSG='Syntax error - "' // QueryRaw(tlist(2))  &
                // '" was unexpected' )
          RETURN
        END IF
        sub%star = .FALSE.
        ! Everything else unallocated.
        RETURN
      END IF
      
      ! Get the lower bound
      CALL ParseExpr(tlist, 1, part_end - 1, sub%start, sub_err_list)
      ! lower-bound is R517.  Could be part of R518/R519/R521.
      CALL Add(err_list, sub_err_list, COMPONENT='R517' )
      IF (Failure(sub_err_list)) RETURN
      
      ! Is there an upper bound?
      IF (part_end == SIZE(tlist)) THEN
        ! No.  We have a assumed-shape-spec
        sub%star = .FALSE.
        ! start will be allocated, rest not.
        RETURN
      END IF
      
      IF (IsToken(tlist(part_end+1), ittOperator, scck_'*')) THEN
        ! Assumed size (R521) with lower bound.
        IF (SIZE(tlist) > part_end + 1) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='R521',  &
!              LOCATION=???,  &
              MSG='Syntax error - "'  &
                // QueryRaw(tlist(part_end+2)) // '" was unexpected' )
          RETURN
        END IF
        sub%star = .TRUE.
        sub%star_location = QueryLocation(tlist(part_end+1))
        ! start will be allocated, rest not.
        RETURN
      END IF
      
      ! Anything else must be the upper bound of an explicit-shape-spec (R516).
      CALL ParseExpr(  &
          tlist,  &
          part_end + 1, SIZE(tlist),  &
          sub%finish,  &
          sub_err_list )
      ! R518 is for upper-bound.
      CALL Add(err_list, sub_err_list, COMPONENT='R518')
      sub%star = .FALSE.
      RETURN
    END IF
    
    sub%colon = .FALSE.
    
    ! Do we just have a star - an assumed-size-spec ?
    IF (IsToken(tlist(1), ittOperator, scck_'*')) THEN
      ! Assumed size spec with no lower bound.
      sub%star = .TRUE.
      sub%star_location = QueryLocation(tlist(1))
      ! everything else unallocated
      CALL ExpectEnd(tlist, 2, 'R522', err_list)
      RETURN
    END IF
    
    ! We must just have an upper bound of an explicit-shape-spec.
    CALL ParseExpr(tlist, 1, SIZE(tlist), sub%finish, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT='R518 from R516')
    sub%star = .FALSE.
    
  END SUBROUTINE ParseArraySpec_single
  
  
  !*****************************************************************************
  !!
  !> Parse an /explicit-shape-spec-list/.
  !!
  !! An /explicit-shape-spec-list/ is one that has the size of each dimension 
  !! explicitly specified (via an /explicit-shape-spec/).  A dimension may also 
  !! optionally have the lower bound specified.
  !!
  !! These specifications can appear in all types of declaration statement, 
  !! including:
  !! - function results
  !! - common block variables
  !! - module variables
  !! - local variables
  !! - data components
  !! - dummy arguments
  !!
  !! The syntax of an /explicit-shape-spec-list/ is the same as the syntax of 
  !! a /allocate-shape-spec/, except the latter is not constrained to having 
  !! specification expressions for the bounds.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    ss_list           The list of subscripts.
  !!
  !! @param[out]    err_list          List of errors
  
  SUBROUTINE ParseExplicitShapeSpecList(tlist, ss_list, err_list)
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DimSpec), INTENT(OUT), ALLOCATABLE :: ss_list(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL ParseExprBasedShapeSpecList(  &
        tlist,  &
        1, SIZE(tlist),  &
        'R516', '/explicit-shape-spec/',  &
        'R517', '/lower-bound/',  &
        'R518', '/upper-bound/',  &
        ss_list,  &
        err_list )
    
  END SUBROUTINE ParseExplicitShapeSpecList
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an array shape spec list that is 
  !! comprised of upper bound and optional lower bound expressions only.
  !!
  !! This is the worker routine for parsing and /explicit-shape-spec-list/.  
  !! It is also used for parsing an /allocate-shape-spec-list/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     comp              The syntax rule number for the shape 
  !! spec, including the leading 'R'.  This is used as the error message 
  !! component.
  !!
  !! @param[in]     rule              The syntax rule name for the shape 
  !! spec.
  !!
  !! @param[in]     lower_comp        The syntax rule number for the 
  !! lower bound expression.
  !!
  !! @param[in]     lower_name        The syntax rlue name for the lower 
  !! bound expression.
  !!
  !! @param[in]     upper_comp        The syntax rule number for the 
  !! upper bound expression.
  !!
  !! @param[in]     upper_name        The syntax rlue name for the upper 
  !! bound expression.
  !!
  !! @param[out]    ss_list           The list of subscripts.
  !!
  !! @param[out]    err_list          List of errors
  
  SUBROUTINE ParseExprBasedShapeSpecList( tlist, start, finish, comp, rule,  &
      lower_comp, lower_rule, upper_comp, upper_rule, ss_list, err_list )
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CHARACTER(*), INTENT(IN) :: comp
    CHARACTER(*), INTENT(IN) :: rule
    CHARACTER(*), INTENT(IN) :: lower_comp
    CHARACTER(*), INTENT(IN) :: lower_rule
    CHARACTER(*), INTENT(IN) :: upper_comp
    CHARACTER(*), INTENT(IN) :: upper_rule
    TYPE(DimSpec), INTENT(OUT), ALLOCATABLE :: ss_list(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the /implicit-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist(start:finish))
    ALLOCATE(ss_list(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      CALL parse_expr_based_shape_spec(  &
          tlist,  &
          item_start, item_finish,  &
          comp, rule,  &
          lower_comp, lower_rule,  &
          upper_comp, upper_rule,  &
          ss_list(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseExprBasedShapeSpecList
  
  
  !*****************************************************************************
  !!
  !> Parse an /deferred-shape-spec-list/.
  !!
  !! A /deferred-shape-spec-list/ is one that has : for each dimension.  
  !!
  !! These specifications are used for allocatable and pointer variables and 
  !! components.  They also look exactly like an /assumed-shape-spec-list/ 
  !! where none of the specs have a lower bound.
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[out]    ss_list           The list of subscripts.
  !!
  !! @param[out]    err_list          List of errors
  
  SUBROUTINE ParseDeferredShapeSpecList(tlist, ss_list, err_list)
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DimSpec), INTENT(OUT), ALLOCATABLE :: ss_list(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the /implicit-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist)
    ALLOCATE(ss_list(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 1, SIZE(tlist), item_start, item_finish)
      CALL parse_deferred_shape_spec(  &
          tlist(item_start:item_finish),  &
          ss_list(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseDeferredShapeSpecList
  
  
  !*****************************************************************************
  !!
  !> Parse a deferred-coshape-spec-list or an explicit-coshape-spec.
  !!
  !! @param[out]    ss_list           The list of subscripts.
  !!
  !! @param[in]     tlist             The list of tokens that make up the 
  !! subscript list.  The first token should be the token that comes 
  !! immediately after the opening parentheses, and the last token should be 
  !! the token immediately before the closing parentheses.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseCoshapeSpec_list(tlist, ss_list, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DimSpec), INTENT(OUT), ALLOCATABLE :: ss_list(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
  
    !---------------------------------------------------------------------------
    ! Local constants
    
    CHARACTER(*), PARAMETER :: comp = 'R509'
    
    !---------------------------------------------------------------------------
    ! Arguments

    INTEGER :: list_count     ! Number of items in the /coarray-spec/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Current token
    INTEGER :: i
    
    ! Error lists from procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)

    !***************************************************************************
    
    ALLOCATE(err_list(0))    
    
    ! We don't know what sort of /array-spec/ we are processing, therefore
    ! we don't know what to expect for a particular item.
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /coarray-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist)
    ALLOCATE(ss_list(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 1, SIZE(tlist), item_start, item_finish)
      CALL ParseCoshapeSpec(  &
          tlist(item_start:item_finish),  &
          ss_list(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    !---------------------------------------------------------------------------
    ! Check that we've not got something daft like [:,*] or [*,*]
    
    ! Test for mix of explicit and allocatable.
    IF (ss_list(SIZE(ss_list))%star) THEN
      ! Earlier dimensions must be explicit-spec (have upper bound).
      DO i = 1, SIZE(ss_list) - 1
        IF (.NOT. ALLOCATED(ss_list(i)%finish)) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='R511',  &
!              LOCATION=???,  &
              MSG='Syntax error - * specified for last dimension &
                &but ealier dimensions are not explicit.' )
          EXIT
        END IF
        IF (ss_list(i)%star) THEN 
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='R511',  &
!              LOCATION=???,  &
              MSG='Syntax error - * specified for dimensions other &
              &than the last.' )
          EXIT
        END IF
      END DO
    ELSE
      DO i = 1, SIZE(ss_list)
        IF ( ALLOCATED(ss_list(i)%start)  &
            .OR. ALLOCATED(ss_list(i)%finish) ) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='R511',  &
!              LOCATION=???,  &
              MSG='Syntax error - an /explicit-coshape-spec/ &
              &must have * specified for the last dimension, a &
              &/deferred-coshape-spec/ must consist only of ":".' )
          EXIT
        END IF
        IF (ss_list(i)%star) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='R511',  &
!              LOCATION=???,  &
              MSG='Syntax error - * specified for dimensions other &
              &than the last.' )
          EXIT
        END IF
      END DO
    END IF
    
  END SUBROUTINE ParseCoshapeSpec_list
  
  
  !*****************************************************************************
  !!
  !> Parse a /deferred-coshape-spec-list/.
  !!
  !! An /deferred-coshape-spec-list/ is one that has : for each dimension.  
  !!
  !! This procedure is remarkably similar to ParseDeferredShapeSpecList.
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[out]    ss_list           The list of subscripts.
  !!
  !! @param[out]    err_list          List of errors
  
  SUBROUTINE ParseDeferredCoshapeSpecList(tlist, ss_list, err_list)
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DimSpec), INTENT(OUT), ALLOCATABLE :: ss_list(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the /implicit-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist)
    ALLOCATE(ss_list(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 1, SIZE(tlist), item_start, item_finish)
      CALL parse_deferred_coshape_spec(  &
          tlist(item_start:item_finish),  &
          ss_list(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseDeferredCoshapeSpecList
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a coshape specification that 
  !! is comprised of optional lower bound and upper bound expressions, bar 
  !! the last upper bound specification, which must be "*".
  !!
  !! Both /allocate-coarray-spec/ and /explicit-coshape-spec/ could be 
  !! parsed by this, however the latter never appears in a context where it 
  !! can be distinguished prior to parsing.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! subscript list.  The first token should be the token that comes 
  !! immediately after the opening parentheses, and the last token should be 
  !! the token immediately before the closing parentheses.
  !!
  !! @param[in]     comp              The syntax rule number for the shape 
  !! spec, including the leading 'R'.  This is used as the error message 
  !! component.
  !!
  !! @param[in]     rule              The syntax rule name for the shape 
  !! spec.
  !!
  !! @param[in]     lower_comp        The syntax rule number for the 
  !! lower bound expression.
  !!
  !! @param[in]     lower_name        The syntax rlue name for the lower 
  !! bound expression.
  !!
  !! @param[in]     upper_comp        The syntax rule number for the 
  !! upper bound expression.
  !!
  !! @param[in]     upper_name        The syntax rlue name for the upper 
  !! bound expression.
  !!
  !! @param[out]    ss_list           The list of subscripts.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseAllocateCoarraySpec(tlist, start, finish, ss_list, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(DimSpec), INTENT(OUT), ALLOCATABLE :: ss_list(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /allocate-coarray-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R636'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the spec list.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    INTEGER :: colon          ! Index of the colon in the last item.
    TYPE(Error) :: err        ! Error object for Expect call.
    INTEGER :: it             ! Index of the current token.
    
    ! Error lists from procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected ' //comp // '.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Determine number of items and parse all but the last.
    
    list_count = ListItemCount(tlist(start:finish))
    ALLOCATE(ss_list(list_count))
    item_finish = 0
    DO item = 1, list_count - 1   ! Last item is special.
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      CALL parse_expr_based_shape_spec(  &
          tlist,  &
          item_start, item_finish,  &
          'R637', '/allocate-coshape-spec/',  &
          'R634', '/lower-bound-expr/',  &
          'R635', '/upper-cound-expr/',  &
          ss_list(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    !---------------------------------------------------------------------------
    ! Upper bound of the last spec must be "*".
    
    ASSOCIATE(sub => ss_list(list_count))
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      colon = FindBare(tlist(:item_finish), ittColon, scck_':', item_start)
      IF (colon /= 0) THEN
        sub%colon = .TRUE.
        IF (colon == 1) THEN
          ! No lower bound.  Might be assumed-shape-spec (R519) or 
          ! deferred-shape-spec (R520), but it isn't an explicit-shape-spec.
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT=comp,  &
!                LOCATION=tlist(part_end)%location,  &
              MSG='If ":" appears in the last spec in an &
                &/allocate-corarray-spec/ then /lower-bound-expr/ &
                &is required.' )
          RETURN
        END IF
        
        ! Get the lower bound
        CALL ParseExpr(tlist, item_start, colon - 1, sub%start, sub_err_list)
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        it = colon + 1
      ELSE
        sub%colon = .FALSE.
        it = item_start
      END IF
      
      CALL Expect(tlist, it, ittOperator, scck_'*', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ExpectEnd(tlist(:finish), it + 1, comp, err_list)
    END ASSOCIATE
    
  END SUBROUTINE ParseAllocateCoarraySpec
  
  
  !*****************************************************************************
  !!
  !> Parse a single item specified in an coshape-spec (R515).
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[out]    sub               The subscript (item in the array-spec 
  !! for a particular dimension).
  !!
  !! @param[out]    sub_err_list      List of errors.
  !!
  !! Note that we never allocate @a sub%stride - as that is not a valid 
  !! part of an array-spec.
  
  SUBROUTINE ParseCoshapeSpec_single(tlist, sub, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DimSpec), INTENT(OUT) :: sub
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! List of errors from procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
        
    INTEGER :: part_end     ! Position of the colon, if present.
    INTEGER :: it           ! Current token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! We don't know what sort of /array-spec/ we are processing, therefore
    ! we don't know what to expect for a particular item.
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          MSG='Expected a coarray shape specification.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! See if there's a colon.
    part_end = FindBare(tlist, ittColon, scck_':')
    IF (part_end /= 0) THEN
      ! We might have a lower bound, or we might be an deferred-coshape-spec.
      sub%colon = .TRUE.
      IF (part_end == 1) THEN
        ! No lower bound - must be deferred-coshape.  In that case, there
        ! must not be an upper bound.
        CALL ExpectEnd(tlist, 2, 'R510', err_list)
        sub%star = .FALSE.
        ! Everything else unallocated.
        RETURN
      END IF
      
      ! Get the lower bound
      CALL ParseExpr(tlist, 1, part_end - 1, sub%start, sub_err_list)
      ! lower-cobound is R512, part of R511
      CALL Add(err_list, sub_err_list, COMPONENT='R512' )
      IF (Failure(sub_err_list)) RETURN
      
      ! Is there an upper bound?
      IF (part_end == SIZE(tlist)) THEN
        ! No.  That's an error.
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='R511',  &
!            LOCATION=???,  &
            MSG='Syntax error - missing upper-cobound' )
        RETURN
      END IF
      it = part_end + 1
    ELSE
      sub%colon = .FALSE.
      it = 1
    END IF
    
    ! Do we just have a star?
    IF (IsToken(tlist(it), ittOperator, scck_'*')) THEN
      ! Assumed size spec.
      sub%star = .TRUE.
      ! everything else unallocated
      CALL ExpectEnd(tlist, it + 1, 'R511', err_list)
      RETURN
    END IF
    
    ! We must just have an upper bound of an explicit-coshape-spec.
    CALL ParseExpr(tlist, it, SIZE(tlist), sub%finish, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT='R513')
    sub%star = .FALSE.
    
  END SUBROUTINE ParseCoshapeSpec_single
  
  
  !*****************************************************************************
  !!
  !> Parse a /section-subscript-list/.
  !!
  !! @param[in]     tlist             Sequence of tokens.
  !!
  !! @param[out]    ss_list           The results of parsing the 
  !! section subscript list.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This cannot be used to parse the array-spec's that appear in dimension
  !! statements (despite there being many things in common) - use 
  !! ParseArraySpec instead.
  !!
  !! Note we are parsing a comma separated list of subscripts here.  Parsing 
  !! of a single subscript is done by parse_section_subscript, which is 
  !! not a public member of this module.
  !!
  !! Needs to be recursive, because an expression in a subscript list could 
  !! also involve subscript lists.
  
  RECURSIVE SUBROUTINE ParseSectionSubscriptList(tlist, ss_list, err_list)
    
    USE CompilerKinds
    USE Errors
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DimSpec), INTENT(OUT), ALLOCATABLE :: ss_list(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: item_count     ! Number of items in the list.
    INTEGER :: item_start     ! Start index in tlist for a particular item.
    INTEGER :: item_finish    ! Finish index in tlist for a particular item.
    INTEGER :: item           ! Item index
    
    ! Error lists from procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    item_count = ListItemCount(tlist)
    ALLOCATE(ss_list(item_count))
    item_finish = 0
    DO item = 1, item_count
      CALL GetNextListItem(tlist, 1, SIZE(tlist), item_start, item_finish)
      CALL parse_section_subscript(  &
          tlist(item_start:item_finish),  &
          ss_list(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseSectionSubscriptList
  
  
  !*****************************************************************************
  !!
  !> Parse a /section-subscript/ (R620).
  !!
  !! @param[in]     tlist             Sequence of tokens.
  !!
  !! @param[out]    sub               Results of parsing the single 
  !! /section-subscript/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Note that a subscript is part of this, a triplet can contain three
  !! subscripts.
  !!
  !! Needs to be recursive, because an expression that is a subscript could
  !! also involve subscripts.
  !!
  !! This procedure handles the scanner deficiency that we cannot (or do 
  !! not, perhaps) disambiguate between two adjacent tokens being a 
  !! double colon token (ittDeclare) or two separate colon tokens.
  
  RECURSIVE SUBROUTINE parse_section_subscript(tlist, sub, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DimSpec), INTENT(OUT) :: sub
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for /section-subscript/ for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R620'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! List of errors from procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Location of the end of the current part (subscript) being considered.
    INTEGER :: first_colon
    INTEGER :: second_colon
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      ! No component - we expect the parent syntax rule to apply here.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /section-subscript/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    sub%star = .FALSE.    ! Not applicable to section-subscripts
    
    ! Do we have a triplet?
    first_colon = get_first_colon()
    IF (first_colon == 0) THEN
      sub%colon = .FALSE.
      ! We just have a /subscript/ or a /vector-subscript/.
      CALL ParseExpr(tlist, 1, SIZE(tlist), sub%finish, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT='R619/R623')
      RETURN
    END IF
    
    ! We have a subscript-triplet.  It has at least one colon.
    sub%colon = .TRUE.
    
    ! Go looking for start
    IF (first_colon > 1) THEN
      CALL ParseExpr(tlist, 1, first_colon - 1, sub%start, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT='R621')
    END IF
    
    IF (tlist(first_colon) == ittDeclare) THEN
      ! No finish subscript, but a stride subscript is required. Locate token 
      ! before the stride.
      second_colon = first_colon
    ELSE
      ! Finish subscript (and stride) are optional.
      IF (first_colon == SIZE(tlist)) RETURN
      
      ! Go looking for the stride's colon
      second_colon = FindBare(tlist(first_colon+1:), ittColon, scck_':')
      IF (second_colon == 0) THEN
        CALL ParseExpr(  &
            tlist,  &
            first_colon + 1, SIZE(tlist),  &
            sub%finish,  &
            sub_err_list)
        CALL Add(err_list, sub_err_list, COMPONENT='R521')
        RETURN
      END IF
      
      ! Locate token before the stride.
      second_colon = second_colon + first_colon
    END IF
    
    ! Process the stride
    IF (second_colon == SIZE(tlist)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='R521',  &
!          LOCATION=???,  &
          MSG='Syntax error - missing stride' )
      RETURN
    END IF
    
    CALL ParseExpr(  &
        tlist,  &
        second_colon + 1, SIZE(tlist),  &
        sub%stride,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT='R521')
    
    ! All done!
  CONTAINS
    
    ! Logical for the first "colon" (either single or double) that 
    ! is present in the token list.
    FUNCTION get_first_colon()
      
      !-------------------------------------------------------------------------
      ! Characteristics
      
      ! Function result.
      INTEGER :: get_first_colon
      
      !-------------------------------------------------------------------------
      ! Local variables
      
      INTEGER :: single_colon ! Index of the first single colon, or zero.
      INTEGER :: double_colon ! Index of the first double colon, or zero.
      
      !*************************************************************************
      
      single_colon = FindBare(tlist, ittColon, scck_':')
      double_colon = FindBare(tlist, ittDeclare, scck_'::')
      
      IF (single_colon == 0) THEN
        IF (double_colon == 0) THEN
          ! No colons.
          get_first_colon = 0
        ELSE
          ! Double colon ahead.
          get_first_colon = double_colon
        END IF
      ELSE
        ! Three's a single colon somewhere ahead.
        IF (double_colon == 0) THEN
          ! No double colon.
          get_first_colon = single_colon
        ELSE
          ! There's a syntax error ahead, but for now...
          get_first_colon = MIN(single_colon, double_colon)
        END IF
      END IF
      
    END FUNCTION get_first_colon
    
  END SUBROUTINE parse_section_subscript
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an expression based shape 
  !! specification, such as an /explicit-shape-spec/ or an 
  !! /allocate-shape-spec/.
  !!
  !! These specifications have an optional lower bound followed by :,
  !! followed by a required upper bound, where both the lower and upper bounds 
  !! are expressions.
  !!
  !! In the cae of an /explicit-shape-expr/ the expressions are 
  !! /specification-expr/'s, but we don't care about that at this stage of 
  !! parsing.
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[in]     comp              The syntax rule number for the shape 
  !! spec, including the leading 'R'.  This is used as the error message 
  !! component.
  !!
  !! @param[in]     rule              The syntax rule name for the shape 
  !! spec.
  !!
  !! @param[in]     lower_comp        The syntax rule number for the 
  !! lower bound expression.
  !!
  !! @param[in]     lower_name        The syntax rlue name for the lower 
  !! bound expression.
  !!
  !! @param[in]     upper_comp        The syntax rule number for the 
  !! upper bound expression.
  !!
  !! @param[in]     upper_name        The syntax rlue name for the upper 
  !! bound expression.
  !!
  !! @param[out]    sub               The subscript specification.
  !!
  !! @param[out]    err_list          List of errors
  
  SUBROUTINE parse_expr_based_shape_spec( tlist, start, finish,  &
      comp, rule, lower_comp, lower_rule, upper_comp, upper_rule,  &
      sub, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CHARACTER(*), INTENT(IN) :: comp
    CHARACTER(*), INTENT(IN) :: rule
    CHARACTER(*), INTENT(IN) :: lower_comp
    CHARACTER(*), INTENT(IN) :: lower_rule
    CHARACTER(*), INTENT(IN) :: upper_comp
    CHARACTER(*), INTENT(IN) :: upper_rule
    TYPE(DimSpec), INTENT(OUT) :: sub
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    INTEGER :: colon          ! Index of the next colon.
    
    ! Error list for child parsing calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! Two ,, in succession (or ",)" at the end might trip this.
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected ' // rule // '.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Always, by syntax definition.
    sub%star = .FALSE.
    
    ! See if we have a lower bound.  If we have a lower bound then there 
    ! will be a : in the tokens.
    colon = FindBare(tlist(:finish), ittColon, scck_':', start)
    IF (colon /= 0) THEN
      ! We might have a lower bound
      sub%colon = .TRUE.
      IF (colon == 1) THEN
        ! No lower bound.  Might be assumed-shape-spec (R519) or 
        ! deferred-shape-spec (R520), but it isn't an explicit-shape-spec.
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
!            LOCATION=tlist(part_end)%location,  &
            MSG='If ":" appears in an ' // rule // ' then ' // lower_rule  &
              // ' is required.' )
        RETURN
      END IF
      
      ! Get the lower bound
      CALL ParseExpr(tlist, start, colon - 1, sub%start, sub_err_list)
      ! lower-bound is R517.
      CALL Add(err_list, sub_err_list, COMPONENT=lower_comp // ' via ' // comp)
      IF (Failure(sub_err_list)) RETURN
      
      it = colon + 1
    ELSE 
      sub%colon = .FALSE.
      it = start
    END IF
    ! it here the index of the first token of the upper bound.
    
    !---------------------------------------------------------------------------
    ! Upper bound
    
    IF (it > finish) THEN
      ! Missing upper bound.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
!          LOCATION=tlist(SIZE(tlist))%location,  &
          MSG='An ' // upper_rule // ' is required for an ' // rule // '.' )
      RETURN
    END IF
    
    ! Common cock-up.
    IF (IsToken(tlist(it), ittOperator, scck_'*')) THEN   
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
!          LOCATION=???,  &
          MSG='A "*" is not valid in an ' // rule // '.' )
      RETURN
    END IF
    
    CALL ParseExpr(tlist, it, finish, sub%finish, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=upper_comp // ' via ' // comp)
    
  END SUBROUTINE parse_expr_based_shape_spec
  
  
  !*****************************************************************************
  !!
  !> Parse a /deferred-shape-spec/.
  !!
  !! An /deferred-shape-spec/ is simply a ":", with no lower or upper bound.
  !!
  !! To be nice, we test to see if the array specification is one of the
  !! other shape-spec's.
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[out]    sub               The subscript specification.
  !!
  !! @param[out]    err_list          List of errors
  
  SUBROUTINE parse_deferred_shape_spec(tlist, sub, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DimSpec), INTENT(OUT) :: sub
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /deferred-shape-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R520'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /deferred-shape-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Alway, by syntax definition.  
    sub%star = .FALSE.
    sub%colon = .TRUE.
    
    ! Some simple diagnostics for potential cock-ups.
    it = FindBare(tlist, ittColon, scck_':', 1)
    IF (it == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(SIZE(tlist))),  &
          MSG='Expected ":".' )
    ELSE IF (it /= 1) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(1)),  &
          MSG='A /deferred-shape-spec/ must not have a lower bound.' )
    ELSE IF (it /= SIZE(tlist)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(it+1)),  &
          MSG='A /deferred-shape-spec/ must not have a upper bound.' )
    END IF
    
  END SUBROUTINE parse_deferred_shape_spec
  
  
  !*****************************************************************************
  !!
  !> Parse an /deferred-coshape-spec/.
  !!
  !! An /deferred-shape-spec/ is simply a ":", with no lower or upper bound.
  !!
  !! To be nice, we test to see if the array specification is one of the 
  !! other shape-spec's.
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[out]    sub               The subscript specification.
  !!
  !! @param[out]    err_list          List of errors
  
  SUBROUTINE parse_deferred_coshape_spec(tlist, sub, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DimSpec), INTENT(OUT) :: sub
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /deferred-coshape-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R510'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /deferred-coshape-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Alway, by syntax definition.  
    sub%star = .FALSE.
    sub%colon = .TRUE.

    ! Some simple diagnostics for potential cock-ups.
    it = FindBare(tlist, ittColon, scck_':', 1)  
    IF (it == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
!          LOCATION=tlist(part_end)%location,  &
          MSG='Expected ":".' )
    ELSE IF (it /= 1) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
!          LOCATION=tlist(part_end)%location,  &
          MSG='A /deferred-coshape-spec/ must not have a lower bound.' )
    ELSE IF (it /= SIZE(tlist)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
!          LOCATION=tlist(part_end)%location,  &
          MSG='A /deferred-coshape-spec/ must not have a upper bound.' )
    END IF
    
  END SUBROUTINE parse_deferred_coshape_spec
  
  
  !*****************************************************************************
  !!
  !> Write a comma separated list of DimSpec's to a logical unit.
  !!
  !! @param[in]     log_unit          Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     dim_specs         The DimSpec's to be written.  Must not
  !! be zero size.
  !!
  !! Output is nonadvancing.
  
  SUBROUTINE WriteDimSpec(log_unit, dim_specs)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(DimSpec), INTENT(IN) :: dim_specs(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Dimension specification index.
    
    !***************************************************************************
    
    i = 1
    DO 
      CALL write_dim_spec(log_unit, dim_specs(i))
      IF (i == SIZE(dim_specs)) EXIT
      WRITE (log_unit, "(',')", ADVANCE='NO')
      i = i + 1
    END DO
    
  END SUBROUTINE WriteDimSpec
  
  
  !*****************************************************************************
  !!
  !> Write a single DimSpec to a logical unit.
  !!
  !! @param[in]     log_unit          Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     dim_specs         The DimSpec to be written.
  !!
  !! Output is nonadvancing.
  !!
  !! @todo Handle expressions for start, finish and stride components.
  
  SUBROUTINE write_dim_spec(log_unit, dim_spec)
    
    USE Expressions
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(DimSpec), INTENT(IN) :: dim_spec
    
    !***************************************************************************
    
    IF (ALLOCATED(dim_spec%start))  &
        CALL dim_spec%start%Write(log_unit)
    IF (dim_spec%colon)  &
        WRITE (log_unit, "(':')", ADVANCE='NO')
    IF (ALLOCATED(dim_spec%finish))  &
        CALL dim_spec%finish%Write(log_unit)
    IF (dim_spec%star)  &
        WRITE (log_unit, "('*')", ADVANCE='NO')
    IF (ALLOCATED(dim_spec%stride)) THEN
      WRITE (log_unit, "(':')", ADVANCE='NO')
      CALL dim_spec%stride%Write(log_unit)
    END IF
    
  END SUBROUTINE write_dim_spec
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimSpec%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !1 @param[in]     pnode             The node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE ds_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimSpec), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL write_dim_spec(unit, pnode)
    
  END SUBROUTINE ds_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimSpec%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE ds_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimSpec), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%start)) THEN
      CALL pnode%start%Query(visitor)
    END IF
    
    IF (ALLOCATED(pnode%finish)) THEN
      CALL pnode%finish%Query(visitor)
    END IF
    
    IF (ALLOCATED(pnode%stride)) THEN
      CALL pnode%stride%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ds_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimSpec%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE ds_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimSpec), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%start)) THEN
      CALL pnode%start%Modify(visitor)
    END IF
    
    IF (ALLOCATED(pnode%finish)) THEN
      CALL pnode%finish%Modify(visitor)
    END IF
    
    IF (ALLOCATED(pnode%stride)) THEN
      CALL pnode%stride%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ds_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimSpec%Advertise - advertise 
  !! the identifiers referenced or defined in the dimension spec.
  !!
  !! @param[in]     pnode             The designator to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE ds_Advertise(pnode, ident_list)
    
    USE CharacterTypes
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimSpec), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
  END SUBROUTINE ds_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimensionStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'array-specish'.
  
  FUNCTION ds_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'array-specish'
    
  END FUNCTION ds_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of DimSpec%GetLocation - get the start of 
  !! the source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION ds_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(DimSpec), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%start)) THEN
      loc = .LocationOf. pnode%start
    ELSE IF (pnode%colon) THEN
      loc = pnode%colon_location
    ELSE IF (ALLOCATED(pnode%finish)) THEN
      loc = .LocationOf. pnode%finish
    ELSE IF (pnode%star) THEN
      loc = pnode%star_location
    ELSE
      STOP 'Internal error in SubscriptLists%%ds_GetLocation, &
          &there was no location!'
    END IF
    
  END FUNCTION ds_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Test whether a dimension specification list is substring range like.
  !!
  !! @param[in]     dim_specs         The dimension specifications to test.
  !!
  !! @returns True if the dimension specification could actually be 
  !! a /substring-range/.
  !!
  !! The criteria are:
  !! - only one subscript.
  !! - must have a colon.
  !! - must not have a stride.
  !!
  !! This criteria does not exclude the possibility of the dimension 
  !! specification being an array section of a rank one array.
  
  FUNCTION IsSubstringRangeLike(dim_specs) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DimSpec), INTENT(IN) :: dim_specs(:)
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = SIZE(dim_specs) == 1
    IF (.NOT. b) RETURN
    
    b = dim_specs(1)%colon
    IF (.NOT. b) RETURN
    
    b = .NOT. ALLOCATED(dim_specs(1)%stride)
    IF (.NOT. b) RETURN
    
  END FUNCTION IsSubstringRangeLike
  
  
  !*****************************************************************************
  !!
  !> Test whether a dimension specification list could not be a 
  !! function reference.
  !!
  !! @param[in]     dim_specs         The dimension specifications to test.
  !!
  !! @returns true if the dimension specification could not be the argument 
  !! list of a function reference.
  !!
  !! The criteria is that any one of the dimension specification contains 
  !! a colon.
  
  FUNCTION IsNotFunctionArgList(dim_specs) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DimSpec), INTENT(IN) :: dim_specs(:)
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = ANY(dim_specs%colon)
    
  END FUNCTION IsNotFunctionArgList
  
END MODULE SubscriptLists
