! $Id: SubstringRanges.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the SubstringRanges module.


!*******************************************************************************
!!
!> Types for storing substrings.

MODULE SubstringRanges
  
  USE Expressions
  USE ParseNodes
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseSubstringRange
  PUBLIC :: WriteSubstringRange
  
  !-----------------------------------------------------------------------------
  
  !> A substring.
  !!
  !! This is very similar to a subscript list, but a substring range always 
  !! has the colon in the middle.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: SubstringRange
    !> If not allocated then the substring starts from position one.
    CLASS(Expression), ALLOCATABLE :: start
    !> If not allocated then the substring ends at the end of the parent 
    !! string.
    CLASS(Expression), ALLOCATABLE :: finish
    !> The location of the colon in the middle of start and finish.
    TYPE(SourceLocation) :: colon_location
  CONTAINS
    PROCEDURE :: Write => sr_Write
    PROCEDURE :: Query => sr_Query
    PROCEDURE :: Modify => sr_Modify
    PROCEDURE, NOPASS :: What => sr_What
    PROCEDURE :: GetLocation => sr_GetLocation
  END TYPE SubstringRange
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a substring-range "[scalar-int-expr] : [scalar-int-expr]".
  !!
  !! At this stage of the parse we don't know whether the expressions
  !! are scalar-int-expr.
  !!
  !! Potential confusion with an array-element or array section, so only call 
  !! this if you know an array-element or array-section is not valid or you
  !! are prepared to resolve the ambiguity later.
  
  RECURSIVE SUBROUTINE ParseSubstringRange(ss_range, tlist, err_list)
    
    USE CompilerKinds
    USE Errors
    USE Expressions
    USE Tokens
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(SubstringRange), INTENT(OUT) :: ss_range
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Temporary for building the parsed representation of an expression.
    CLASS(Expression), ALLOCATABLE :: ex
    
    INTEGER :: finish         ! End of a particular part.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************    
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add(err_list, MSG='Expected substring-range')
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Get the extent of the start expression
    finish = FindBare(tlist, ittColon, scck_':')
    
    IF (finish == 0) THEN
      CALL Add(err_list, MSG='substring-range missing ":"', COMPONENT='R610')
      RETURN
    END IF
    
    IF (finish /= 1) THEN
      ! We appear to have the start expression.
      CALL ParseExpr(tlist, 1, finish - 1, ex, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT='R610')
      IF (Failure(sub_err_list)) RETURN
      
      CALL MOVE_ALLOC(ex, ss_range%start)
    END IF
    
    IF (finish < SIZE(tlist)) THEN
      CALL ParseExpr(tlist, finish+1, SIZE(tlist), ex, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT='R610')
      IF (Failure(sub_err_list)) RETURN
      
      CALL MOVE_ALLOC(ex, ss_range%finish)
    END IF
    
    ! Save the colon location.
    ss_range%colon_location = QueryLocation(tlist(finish))
    
  END SUBROUTINE ParseSubstringRange
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubstringRange%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !1 @param[in]     pnode             The node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE sr_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubstringRange), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL WriteSubstringRange(unit, pnode)
    
  END SUBROUTINE sr_Write
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of a substring range to a unit.
  !!
  !! @param[in]     log_unit          The unit connected for formatted output 
  !! to write the representation of the substring range to.
  !!
  !! @param[in]     subs              The substring range to write out.
  
  RECURSIVE SUBROUTINE WriteSubstringRange(log_unit, subs)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(SubstringRange), INTENT(IN) :: subs
    
    !***************************************************************************
    
    IF (ALLOCATED(subs%start)) THEN
      CALL WriteExpr(log_unit, subs%start)
    END IF
    WRITE (log_unit, "(':')", ADVANCE='NO')
    IF (ALLOCATED(subs%finish)) THEN
      CALL WriteExpr(log_unit, subs%finish)
    END IF
    
  END SUBROUTINE WriteSubstringRange
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubstringRange%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE sr_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubstringRange), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%start)) THEN
      CALL pnode%start%Query(visitor)
    END IF
    
    IF (ALLOCATED(pnode%finish)) THEN
      CALL pnode%finish%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sr_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubstringRange%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE sr_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubstringRange), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%start)) THEN
      CALL pnode%start%Modify(visitor)
    END IF
    
    IF (ALLOCATED(pnode%finish)) THEN
      CALL pnode%finish%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sr_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubstringRange%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns The syntax rule name represented by parse node.
  
  FUNCTION sr_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'substring-range'
    
  END FUNCTION sr_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubstringRange%GetLocation - get the start of the source 
  !! that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION sr_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(SubstringRange), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%start)) THEN
      loc = .LocationOf. pnode%start
    ELSE IF (ALLOCATED(pnode%finish)) THEN
      loc = .LocationOf. pnode%finish
    ELSE
      loc = pnode%colon_location
    END IF
    
  END FUNCTION sr_GetLocation
  
END MODULE SubstringRanges
