! $Id: VariableDefinitionContext.f90 2323 2016-08-15 03:15:16Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the VariableDefinitionContext program.


!*******************************************************************************
!!
!> Concept program that reports the names of base objects that appear in 
!! what is known to be a variable definition context, after parsing, in a 
!! scope.
!!
!! As this reporting happens prior to semantic analysis we do not 
!! report variable definition contexts that require knowledge across 
!! statements or program units - such as:
!!
!! - actual arguments associated with INTENT(OUT) or INTENT(INOUT) dummy 
!!   arguments,
!!
!! - definition of an object via association via an associate name that 
!!   is itself used in a variable definition context,
!!
!! - a variable in a namelist statement for a namelist that appears in a 
!!   namelist input statement,
!!
!! - an internal file variable in a write statement (we don't know if the 
!!   variable identifies an internal file or a logical unit).
!!
!! We are also oblivious to whether a component of an object is a pointer 
!! component - in which case the target of the pointer object is not 
!! considered a subobject of the base object - but we always report 
!! the base object anyway.
!!
!! The "name of a base object" that is reported for a particular variable 
!! definition context is the name that appears in the left most 
!! part ref of a designator.
!!
!! Provide the names of the source files to read as command line arguments.

PROGRAM VariableDefinitionContext
  
  USE VariableDefinitionContextImpl
  USE CmdLine
  USE Errors
  USE FileUtilities
  USE Strings
  
  USE, INTRINSIC :: ISO_FORTRAN_ENV, ONLY: OUTPUT_UNIT, ERROR_UNIT
  
  IMPLICIT NONE
  
  !-----------------------------------------------------------------------------
  ! Local variables
  
  ! Command line options.
  !
  ! In order, --args-only.
  TYPE(CmdLineOption) :: cmd_line_options(1)
  
  ! List of non-option arguments.
  TYPE(String), ALLOCATABLE :: arguments(:)
  
  ! Error message
  CHARACTER(:), ALLOCATABLE :: msg
  
  INTEGER :: stat             ! Status code for procedure calls.
  INTEGER :: stop_code        ! Stop code from procedure calls
  
  !*****************************************************************************
  
  !-----------------------------------------------------------------------------
  ! Parse our command line.
  
  cmd_line_options(1) = CmdLineOption(  &
      LONG_NAME='args-only',  &
      DESCRIPTION='Display information for dummy argument identifiers on&
                  &ly.',  &
      OPTION_TYPE=ioptFlag )

  CALL ParseCmdLine(cmd_line_options, arguments, stat, msg)
  IF (stat /= 0) THEN
    WRITE (ERROR_UNIT, "('Error parsing command line: ',A)") msg
    WRITE (ERROR_UNIT, "('Read the source and work out what you did wrong!')")
    STOP 3
  END IF
  
  !-----------------------------------------------------------------------------
  ! JUST DO IT!
  
  CALL Execute(  &
      arguments,  &
      cmd_line_options(1)%PresentCount /= 0,  &
      OUTPUT_UNIT,  &
      ERROR_UNIT,  &
      stop_code )
  
  IF (stop_code /= 0) CALL do_stop
  
CONTAINS
  
  ! F2003 doesn't allow an expression as a stop-code so this is a work around.
  SUBROUTINE do_stop
    SELECT CASE (stop_code)
    CASE (0)   ! Fall through to end statement.
    CASE (1) ; STOP 1
    CASE (2) ; STOP 2
    CASE DEFAULT;  STOP 3
    END SELECT
  END SUBROUTINE do_stop
  
END PROGRAM VariableDefinitionContext
