! $Id: VariableDefinitionContextImpl.f90 2323 2016-08-15 03:15:16Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the VariableDefinitionContext program.


!*******************************************************************************
!!
!> Concept program that reports the names that appear in a variable 
!! definition context (to the extent known after parsing) in a scope.

MODULE VariableDefinitionContextImpl
  
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: Execute
  
  !-----------------------------------------------------------------------------
  
  !> Information for the current subprogram scope.
  TYPE :: scope_data
    !> The dummy argument names for the current scope.
    TYPE(Identifier), ALLOCATABLE :: list(:)
  END TYPE scope_data
  
  !> Object to traverse the parse node tree and gather identifier information.
  TYPE, EXTENDS(ParseNodeQueryVisitor) :: vdc_visitor
    !> Unit for reporting output.
    INTEGER :: unit = 0
    !> The current stack of dummy argument names.
    !!
    !! May not be allocated, in which case we don't track dummy argument 
    !! names, and report any identifier that appears in a variable 
    !! definition context.
    TYPE(scope_data), ALLOCATABLE :: scope_stack(:)
  CONTAINS
    PROCEDURE :: Start => vdc_Start
    PROCEDURE :: Finish => vdc_Finish
  END TYPE vdc_visitor
  
  !> Format specification for an error message.
  CHARACTER(*), PARAMETER, PUBLIC :: fmt_error = "(A)"
  
  !> Format specification for subprogram name output.
  CHARACTER(*), PARAMETER :: fmt_subprogram = "(A)"
  
  !> Format specification for name output.
  CHARACTER(*), PARAMETER :: fmt_name = "(2X,A,' - ',A,' at ',A)"
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Execute the guts of the VariableDefinitionContext program.
  !!
  !! @param[in]     file_names        The list of file names to parse.
  !!
  !! @param[in]     out_unit          Logical unit for output.
  !!
  !! @param[in]     err_unit          Logical unit for reporting errors.
  !!
  !! @param[out]    stop_code         Stop code for reporting program status.
  
  SUBROUTINE Execute(file_names, args_only, out_unit, err_unit, stop_code)
    
    USE BaseStNodes
    USE BuildTreePass
    USE DefaultHosts
    USE Errors
    USE Programs
    USE Strings
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(String), INTENT(IN) :: file_names(:)
    LOGICAL, INTENT(IN) :: args_only
    INTEGER, INTENT(IN) :: out_unit
    INTEGER, INTENT(IN) :: err_unit
    INTEGER, INTENT(OUT) :: stop_code
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Object that specifies host behaviour.
    TYPE(DefaultHost) :: host
    
    ! The root of the statement tree for all files.
    TYPE(BaseStNode), POINTER :: base
    
    ! The statement tree for the current file.
    TYPE(BaseStNode), POINTER :: current
    
    ! The root of the parse tree for the file.
    TYPE(Program) :: program_node
    
    ! List of errors from various procedure calls.
    TYPE(Error), ALLOCATABLE :: err_list(:)
    
    INTEGER :: if               ! Filename index.
    INTEGER :: ie               ! Error index
    
    ! Our visitor.
    TYPE(vdc_visitor) :: visitor
    
    !***************************************************************************
    
    !---------------------------------------------------------------------------
    ! Build the statement trees.
    
    ALLOCATE(base)
    base%ensure_unique = .TRUE.
    ALLOCATE(base%prog_units(0))
    
    DO if = 1, SIZE(file_names)
      CALL DoBuildTreePass(  &
          host,  &
          file_names(if)%item,  &
          current,  &
          err_list,  &
          .TRUE. )
      DO ie = 1, SIZE(err_list)
        IF (Failure(err_list(ie))) THEN
          WRITE (err_unit, fmt_error) .Describe. err_list(ie)
        END IF
      END DO
      IF (.NOT. Failure(err_list)) THEN
        CALL MergeTrees(base, current, err_list)
        DO ie = 1, SIZE(err_list)
          WRITE (err_unit, fmt_error) .Describe. err_list(ie)
        END DO
      END IF
      ! current is allocated in DoBuildTreePass, so deallocate it here.
      DEALLOCATE(current)
      IF (Failure(err_list)) THEN
        stop_code = 2
        RETURN
      END IF
    END DO
    
    !---------------------------------------------------------------------------
    ! Parse the statement tree.
    
    CALL ParseProgram(base, program_node, err_list)
    DO ie = 1, SIZE(err_list)
      WRITE (err_unit, fmt_error) .Describe. err_list(ie)
    END DO
    IF (Failure(err_list)) THEN
      stop_code = 2
      RETURN
    END IF
    
    CALL Check(program_node, err_list)
    DO ie = 1, SIZE(err_list)
      WRITE (err_unit, fmt_error) .Describe. err_list(ie)
    END DO
    IF (Failure(err_list)) THEN
      stop_code = 2
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Go find me some variable definition contexts.
    
    IF (args_only) THEN
      visitor = vdc_visitor(out_unit)     ! , [scope_data::])
      ALLOCATE(visitor%scope_stack(0))
    ELSE
      visitor = vdc_visitor(out_unit)
    END IF
    
    CALL program_node%Query(visitor)
    
  END SUBROUTINE Execute
  
  
  !*****************************************************************************
  !!
  !> Start the query action (print out identifiers that are in 
  !! variable definition contexts) on a parse node.
  !!
  !! @param[in,out] visitor           The object executing the query 
  !! operation.
  !!
  !! @param[in]     pnode             The node being traversed.
  
  SUBROUTINE vdc_Start(visitor, pnode)
    
    USE AllocateStmts
    USE AssignmentStmts
    USE DoConstructs
    USE DoStmts
    USE FunctionStmts
    USE InquireStmts
    USE IoControlSpecs
    USE LockStmts
    USE NullifyStmts
    USE OpenStmts
    USE PointerAssignmentStmts
    USE PositionSpecs
    USE PrimaryExpressions
    USE ProgramUnits
    USE ReadStmts
    USE SubroutineStmts
    USE SyncStmts
    USE WaitStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(vdc_visitor), INTENT(INOUT) :: visitor
    CLASS(ParseNode), INTENT(IN) :: pnode
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Object index.
    
    !***************************************************************************
    
    SELECT TYPE (pnode)
    TYPE IS (FunctionSubprogram)
      IF (ALLOCATED(visitor%scope_stack)) THEN
        visitor%scope_stack  &
            = [visitor%scope_stack, scope_data(pnode%fn_stmt%args)]
      END IF
      
    TYPE IS (SubroutineSubprogram)
      IF (ALLOCATED(visitor%scope_stack)) THEN
        CALL push_scope(visitor%scope_stack, pnode%sub_stmt%args)
      END IF
      
    TYPE IS (AllocateStmt)
      IF (ALLOCATED(pnode%stat)) THEN
        CALL do_var(visitor, pnode%stat, 'STAT= variable')
      END IF
      IF (ALLOCATED(pnode%errmsg)) THEN
        CALL do_var(visitor, pnode%errmsg, 'ERRMSG= variable')
      END IF
      
    TYPE IS (AssignmentStmt)
      CALL do_var(visitor, pnode%var, 'assignment variable')
      
    TYPE IS (DoConstruct)
      IF (ALLOCATED(pnode%loop_control)) THEN
        SELECT TYPE (control => pnode%loop_control)
        CLASS IS (IndexLoopControl)
          CALL do_ident(visitor, control%name, 'do loop index')
        END SELECT
      END IF
      
    TYPE IS (FunctionStmt)
      WRITE (visitor%unit, fmt_subprogram) .NameOf. pnode%function_name
      
    TYPE IS (InquireStmt)
      IF (ALLOCATED(pnode%access)) THEN
        CALL do_var(visitor, pnode%access, 'ACCESS= variable')
      END IF
      IF (ALLOCATED(pnode%action)) THEN
        CALL do_var(visitor, pnode%action, 'ACTION= variable')
      END IF
      IF (ALLOCATED(pnode%asynchronous)) THEN
        CALL do_var(visitor, pnode%asynchronous, 'ASYNCHRONOUS= variable')
      END IF
      IF (ALLOCATED(pnode%blank)) THEN
        CALL do_var(visitor, pnode%blank, 'BLANK= variable')
      END IF
      IF (ALLOCATED(pnode%decimal)) THEN
        CALL do_var(visitor, pnode%decimal, 'DECIMAL= variable')
      END IF
      IF (ALLOCATED(pnode%delim)) THEN
        CALL do_var(visitor, pnode%delim, 'DELIM= variable')
      END IF
      IF (ALLOCATED(pnode%direct)) THEN
        CALL do_var(visitor, pnode%direct, 'DIRECT= variable')
      END IF
      IF (ALLOCATED(pnode%encoding)) THEN
        CALL do_var(visitor, pnode%encoding, 'ENCODING= variable')
      END IF
      IF (ALLOCATED(pnode%exist)) THEN
        CALL do_var(visitor, pnode%exist, 'EXIST= variable')
      END IF
      IF (ALLOCATED(pnode%form)) THEN
        CALL do_var(visitor, pnode%form, 'FORM= variable')
      END IF
      IF (ALLOCATED(pnode%formatted)) THEN
        CALL do_var(visitor, pnode%formatted, 'FORMATTED= variable')
      END IF
      IF (ALLOCATED(pnode%iomsg)) THEN
        CALL do_var(visitor, pnode%iomsg, 'IOMSG= variable')
      END IF
      IF (ALLOCATED(pnode%iostat)) THEN
        CALL do_var(visitor, pnode%iostat, 'IOSTAT= variable')
      END IF
      IF (ALLOCATED(pnode%name)) THEN
        CALL do_var(visitor, pnode%name, 'NAME= variable')
      END IF
      IF (ALLOCATED(pnode%named)) THEN
        CALL do_var(visitor, pnode%named, 'NAMED= variable')
      END IF
      IF (ALLOCATED(pnode%nextrec)) THEN
        CALL do_var(visitor, pnode%nextrec, 'NEXTREC= variable')
      END IF
      IF (ALLOCATED(pnode%number)) THEN
        CALL do_var(visitor, pnode%number, 'NUMBER= variable')
      END IF
      IF (ALLOCATED(pnode%opened)) THEN
        CALL do_var(visitor, pnode%opened, 'OPENED= variable')
      END IF
      IF (ALLOCATED(pnode%pad)) THEN
        CALL do_var(visitor, pnode%pad, 'PAD= variable')
      END IF
      IF (ALLOCATED(pnode%pending)) THEN
        CALL do_var(visitor, pnode%pending, 'PENDING= variable')
      END IF
      IF (ALLOCATED(pnode%pos)) THEN
        CALL do_var(visitor, pnode%pos, 'POS= variable')
      END IF
      IF (ALLOCATED(pnode%position)) THEN
        CALL do_var(visitor, pnode%position, 'POSITION= variable')
      END IF
      IF (ALLOCATED(pnode%read)) THEN
        CALL do_var(visitor, pnode%read, 'READ= variable')
      END IF
      IF (ALLOCATED(pnode%readwrite)) THEN
        CALL do_var(visitor, pnode%readwrite, 'READWRITE= variable')
      END IF
      IF (ALLOCATED(pnode%recl)) THEN
        CALL do_var(visitor, pnode%recl, 'RECL= variable')
      END IF
      IF (ALLOCATED(pnode%round)) THEN
        CALL do_var(visitor, pnode%round, 'ROUND= variable')
      END IF
      IF (ALLOCATED(pnode%sequential)) THEN
        CALL do_var(visitor, pnode%sequential, 'SEQUENTIAL= variable')
      END IF
      IF (ALLOCATED(pnode%sign)) THEN
        CALL do_var(visitor, pnode%sign, 'SIGN= variable')
      END IF
      IF (ALLOCATED(pnode%size)) THEN
        CALL do_var(visitor, pnode%size, 'SIZE= variable')
      END IF
      IF (ALLOCATED(pnode%stream)) THEN
        CALL do_var(visitor, pnode%stream, 'STREAM= variable')
      END IF
      IF (ALLOCATED(pnode%unformatted)) THEN
        CALL do_var(visitor, pnode%unformatted, 'UNFORMATTED= variable')
      END IF
      IF (ALLOCATED(pnode%write)) THEN
        CALL do_var(visitor, pnode%write, 'WRITE= variable')
      END IF
      IF (ALLOCATED(pnode%iolength)) THEN
        CALL do_var(visitor, pnode%iolength, 'IOLENGTH= variable')
      END IF
      
    TYPE IS (LockStmt)
      CALL do_var(visitor, pnode%lock_variable, 'lock variable')
      IF (ALLOCATED(pnode%acquired_lock)) THEN
        CALL do_var(visitor, pnode%acquired_lock, 'ACQUIRED_LOCK= variable')
      END IF
      IF (ALLOCATED(pnode%stat)) THEN
        CALL do_var(visitor, pnode%stat, 'STAT= variable')
      END IF
      IF (ALLOCATED(pnode%errmsg)) THEN
        CALL do_var(visitor, pnode%errmsg, 'ERRMSG= variable')
      END IF
      
    TYPE IS (NullifyStmt)
      DO i = 1, SIZE(pnode%object_list)
        CALL do_ident(  &
            visitor,  &
            pnode%object_list(i)%parts(1)%part_name,  &
            'nullify object' )
      END DO
      
    TYPE IS (OpenStmt)
      IF (ALLOCATED(pnode%newunit)) THEN
        CALL do_var(visitor, pnode%newunit, 'NEWUNIT= variable')
      END IF
      IF (ALLOCATED(pnode%iomsg)) THEN
        CALL do_var(visitor, pnode%iomsg, 'IOMSG= variable')
      END IF
      IF (ALLOCATED(pnode%iostat)) THEN
        CALL do_var(visitor, pnode%iostat, 'IOSTAT= variable')
      END IF
      
    TYPE IS (PointerAssignmentStmt)
      CALL do_ident(visitor, pnode%pointer(1)%part_name, 'pointer object')
      
    TYPE IS (ReadStmt)
      CALL handle_io_item_list(visitor, pnode%item_list)
      
    TYPE IS (SubroutineStmt)
      WRITE (visitor%unit, fmt_subprogram) .NameOf. pnode%subroutine_name
      
    TYPE IS (SyncAllStmt)
      IF (ALLOCATED(pnode%stat)) THEN
        CALL do_var(visitor, pnode%stat, 'STAT= variable')
      END IF
      IF (ALLOCATED(pnode%errmsg)) THEN
        CALL do_var(visitor, pnode%errmsg, 'ERRMSG= variable')
      END IF
      
    TYPE IS (SyncImagesStmt)
      IF (ALLOCATED(pnode%stat)) THEN
        CALL do_var(visitor, pnode%stat, 'STAT= variable')
      END IF
      IF (ALLOCATED(pnode%errmsg)) THEN
        CALL do_var(visitor, pnode%errmsg, 'ERRMSG= variable')
      END IF
      
    TYPE IS (UnlockStmt)
      CALL do_var(visitor, pnode%lock_variable, 'lock variable')
      IF (ALLOCATED(pnode%stat)) THEN
        CALL do_var(visitor, pnode%stat, 'STAT= variable')
      END IF
      IF (ALLOCATED(pnode%errmsg)) THEN
        CALL do_var(visitor, pnode%errmsg, 'ERRMSG= variable')
      END IF
    
    TYPE IS(WaitStmt)
      IF (ALLOCATED(pnode%iomsg)) THEN
        CALL do_var(visitor, pnode%iomsg, 'IOMSG= variable')
      END IF
      IF (ALLOCATED(pnode%iostat)) THEN
        CALL do_var(visitor, pnode%iostat, 'IOSTAT= variable')
      END IF
      
    TYPE IS (AllocateObjectStructureComponent)
      CALL do_ident(visitor, pnode%parts(1)%part_name, 'allocate object')
      
    TYPE IS (AllocateObjectVariable)
      CALL do_ident(visitor, pnode%name, 'allocate object')
      
    TYPE IS (IoControlSpecList)
      ! Why isn't this listed as a variable definition context?
      IF (ALLOCATED(pnode%id)) THEN
        CALL do_var(visitor, pnode%id, 'ID= variable')
      END IF
      IF (ALLOCATED(pnode%iomsg)) THEN
        CALL do_var(visitor, pnode%iomsg, 'IOMSG= variable')
      END IF
      IF (ALLOCATED(pnode%iostat)) THEN
        CALL do_var(visitor, pnode%iostat, 'IOSTAT= variable')
      END IF
      IF (ALLOCATED(pnode%size)) THEN
        CALL do_var(visitor, pnode%size, 'SIZE= variable')
      END IF
      
    TYPE IS (PositionSpecList)
      IF (ALLOCATED(pnode%iomsg)) THEN
        CALL do_var(visitor, pnode%iomsg, 'IOMSG= variable')
      END IF
      IF (ALLOCATED(pnode%iostat)) THEN
        CALL do_var(visitor, pnode%iostat, 'IOSTAT= variable')
      END IF
      
    END SELECT
    
  END SUBROUTINE vdc_Start
  
  
  !*****************************************************************************
  !!
  !> Finish the query action on a parse node.
  !!
  !! @param[in,out] visitor           The object executing the query 
  !! operation.
  !!
  !! @param[in]     pnode             The node being traversed.
  
  SUBROUTINE vdc_Finish(visitor, pnode)
    
    USE ProgramUnits
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(vdc_visitor), INTENT(INOUT) :: visitor
    CLASS(ParseNode), INTENT(IN) :: pnode
    
    !***************************************************************************
    
    SELECT TYPE (pnode)
    TYPE IS (FunctionSubprogram)
      IF (ALLOCATED(visitor%scope_stack)) THEN
        visitor%scope_stack = visitor%scope_stack(:SIZE(visitor%scope_stack)-1)
      END IF
      
    TYPE IS (SubroutineSubprogram)
      IF (ALLOCATED(visitor%scope_stack)) THEN
        visitor%scope_stack = visitor%scope_stack(:SIZE(visitor%scope_stack)-1)
      END IF
      
    END SELECT
    
  END SUBROUTINE vdc_Finish
  
  
  !*****************************************************************************
  !!
  !> Process recursively through an IOItemList.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for reporting.
  !!
  !! @param[in]     item_list         The item list to traverse.
  
  ! Recursive because recursive lists recursive.
  RECURSIVE SUBROUTINE handle_io_item_list(visitor, item_list)
    
    USE InputOutputItems
    USE PrimaryExpressions
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(vdc_visitor), INTENT(IN) :: visitor
    TYPE(IOItemList), INTENT(IN) :: item_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Item list index.
    
    !***************************************************************************
    
    DO I = 1, SIZE(item_list)
      SELECT TYPE (item => item_list(i)%item)
      TYPE IS (IOItemExpr)
        CALL do_var(visitor, item%expr, 'input item list variable')
        
      TYPE IS (IOItemImpliedDo)
        CALL do_ident(  &
            visitor,  &
            item%control%do_variable,  &
            'input item list implied do variable' )
        CALL handle_io_item_list(visitor, item%items)
        
      END SELECT
    END DO
    
  END SUBROUTINE handle_io_item_list
  
  
  !*****************************************************************************
  !!
  !> Checks whether an identifier is relevant, and if it is, outputs it.
  !!
  !! @param[in]     visitor           The visitor object execution the 
  !! query operation.
  !!
  !! @param[in]     ident             The identifier to check.
  !!
  !! @param[in]     ctx               The context to report for the variable.
  
  SUBROUTINE do_ident(visitor, ident, ctx)
    
    USE Identifiers
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(vdc_visitor), INTENT(IN) :: visitor
    TYPE(Identifier), INTENT(IN) :: ident
    CHARACTER(*), INTENT(IN) :: ctx
    
    !---------------------------------------------------------------------------
    ! Locals
    
    LOGICAL :: check          ! Result of the name check.
    
    ! Name of the identifier.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = .NameOf. ident
    
    IF (ALLOCATED(visitor%scope_stack)) THEN
      check = check_name(visitor%scope_stack, name)
    ELSE
      check = .TRUE.
    END IF
    
    IF (check) THEN
      WRITE (visitor%unit, fmt_name)  &
          name,  &
          ctx,  &
          .ToString. (.LocationOf. ident)
    END IF
    
  END SUBROUTINE do_ident
  
  
  !*****************************************************************************
  !!
  !> Checks whether a variable is relevant, and if it is, outputs it.
  !!
  !! @param[in]     visitor           The visitor object execution the 
  !! query operation.
  !!
  !! @param[in]     var               The parse node for the variable to 
  !! check.
  !!
  !! @param[in]     ctx               The context to report for the variable.
  
  SUBROUTINE do_var(visitor, var, ctx)
    
    USE Variables
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(vdc_visitor), INTENT(IN) :: visitor
    CLASS(Variable), INTENT(IN) :: var
    CHARACTER(*), INTENT(IN) :: ctx
    
    !---------------------------------------------------------------------------
    ! Locals
    
    LOGICAL :: check          ! Result of the name check.
    
    ! Name to use for the variable.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = get_name(var)
    
    IF (ALLOCATED(visitor%scope_stack)) THEN
      check = check_name(visitor%scope_stack, name)
    ELSE
      check = .TRUE.
    END IF
    
    IF (check) THEN
      WRITE (visitor%unit, fmt_name)  &
          name,  &
          ctx,  &
          .ToString. (.LocationOf. var)
    END IF
    
  END SUBROUTINE do_var
  
  
  !*****************************************************************************
  !!
  !> Check whether a name is considered relevant.
  !!
  !! @param[in]     scope_stack       The current stack of scope data.
  !!
  !! @param[in]     name              The name to check.
  !!
  !! @returns True if we consider name to be a dummy argument in the current 
  !! scope.
  
  FUNCTION check_name(scope_stack, name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(scope_data), INTENT(IN) :: scope_stack(:)
    CHARACTER(*), INTENT(IN) :: name
    
    ! Function result.
    LOGICAL :: check_name
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: is             ! Stack index.
    INTEGER :: il             ! List index.
    
    !***************************************************************************
    
    DO is = 1, SIZE(scope_stack)
      DO il = 1, SIZE(scope_stack(is)%list)
        check_name = name == .NameOf. scope_stack(is)%list(il)
        IF (check_name) RETURN
      END DO
    END DO
    
    check_name = .FALSE.
    
  END FUNCTION check_name
  
  
  !*****************************************************************************
  !!
  !> Get the name of the object nominated by an expression.
  !!
  !! @param[in]     var               The expression of interest.
  !!
  !! @returns The name of the base object of the expression, or lots 
  !! of question marks if that concept makes no sense.
  
  FUNCTION get_name(var) RESULT(name)
    
    USE Expressions
    USE PrimaryExpressions
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Expression), INTENT(IN) :: var
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    SELECT TYPE (var)
    TYPE IS (DesignatorExpr)
      IF (ALLOCATED(var%desig%part_refs)) THEN
        name = .NameOf. var%desig%part_refs(1)%part_name
      ELSE
        ! We are trying to use a substring of a string literal as a variable.
        name = '??'
      END IF
      
    CLASS DEFAULT
      ! Some other form of expression that we have not yet anticipated.
      name = '???'
      
    END SELECT
    
  END FUNCTION get_name
  
  
  !*****************************************************************************
  !!
  !> Little worker routine to push the argument names for a subroutine 
  !! onto the scope stack.
  !!
  !! @param[in]     scope_stack       The stack of scope data objects.
  !!
  !! @param[in]     args              The argument definitions to push.
  !!
  !! Argument definitions for subroutines can be alternate return 
  !! specifications, so we need special handling when assembling the 
  !! list of argument names.
  
  SUBROUTINE push_scope(scope_stack, args)
    
    USE SubroutineStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(scope_data), INTENT(INOUT), ALLOCATABLE :: scope_stack(:)
    TYPE(DummyArg), INTENT(IN) :: args(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Temporary for growing scope_stack.
    TYPE(scope_data), ALLOCATABLE :: tmp(:)
    
    INTEGER :: ia             ! args index
    INTEGER :: il             ! list index.
    
    !***************************************************************************
    
    ALLOCATE(tmp(SIZE(scope_stack)+1))
    tmp(:SIZE(scope_stack)) = scope_stack
    ALLOCATE(tmp(SIZE(tmp))%list(COUNT(.NOT. args%is_star)))
    
    il = 1
    DO ia = 1, SIZE(args)
      IF (.NOT. args(ia)%is_star) THEN
        tmp(SIZE(tmp))%list(il) = args(ia)%name
        il = il + 1
      END IF
    END DO
    
    CALL MOVE_ALLOC(tmp, scope_stack)
    
  END SUBROUTINE push_scope
  
END MODULE VariableDefinitionContextImpl
