! $Id: WaitStmts.f90 2835 2019-03-25 10:40:22Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the WaitStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /wait-stmt/'s (R904).

MODULE WaitStmts
  
  USE Expressions
  USE Labels
  USE ParseNodes
  USE Variables
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseWaitStmt
  
  !> The results of parsing a /wait-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: WaitStmt
    !> The UNIT specification.
    !!
    !! This is a /file-unit-number/.  Always allocated.
    CLASS(Expression), ALLOCATABLE :: unit
    !> The END specification.
    !!
    !! Not allocated if there was no END specifier.
    TYPE(Label), ALLOCATABLE :: end
    !> The EOR specification.
    !!
    !! Not allocated if there was no EOR specifier.
    TYPE(Label), ALLOCATABLE :: eor
    !> The ERR specification.
    !!
    !! Not allocated if there was no ERR specifier.
    TYPE(Label), ALLOCATABLE :: err
    !> The ID specification.
    !!
    !! For a /wait-stmt/ the id is a /scalar-int-expr/.  In other statements 
    !! it is a variable (the idea being that the value of that variable then 
    !! gets passed back to the /wait-stmt/).
    !!
    !! May not be allocated.
    CLASS(Expression), ALLOCATABLE :: id
    !> The IOMSG specification
    !!
    !! This is an /iomsg-variable/.
    !!
    !! May not be allocated.
    CLASS(Variable), ALLOCATABLE :: iomsg
    !> The IOSTAT specification.
    !!
    !! May not be allocated.
    CLASS(Variable), ALLOCATABLE :: iostat
  CONTAINS
    PROCEDURE :: Dump => wait_Dump
    PROCEDURE :: Query => wait_Query
    PROCEDURE :: Modify => wait_Modify
    PROCEDURE, NOPASS :: What => wait_What
  END TYPE WaitStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /wait-stmt/ (R922).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    wait_stmt         The results of parsing the /wait-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseWaitStmt(tlist, wait_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(WaitStmt), INTENT(OUT) :: wait_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /wait-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R9224'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: list_count     ! Number of items in the /wait-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.   
    LOGICAL :: no_spec        ! Flag to indicate UNIT= absent.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /wait-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! WAIT...
    
    CALL Expect(tlist, 1, ittName, scck_'WAIT', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN

    !---------------------------------------------------------------------------
    ! WAIT(...)
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! /wait-spec-list/
    
    list_count = ListItemCount(tlist(3:closing-1))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 3, closing - 1, item_start, item_finish)
      CALL parse_wait_spec(  &
          tlist,  &
          item_start, &
          item_finish,  &
          wait_stmt,  &
          no_spec,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      ! Continue on error.
      IF ((item /= 1) .AND. no_spec) THEN
        ! No syntax rule.
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C939',  &
            LOCATION=QueryLocation(tlist(item_start)),  &
            MSG='If the optional "UNIT=" is omitted, the /file-unit-number/ &
              &shall be the first item in the /wait-spec-list/.' )
        ! Continue on error.
      END IF
    END DO
    
    !---------------------------------------------------------------------------
    ! Check constraints.
    
    IF (.NOT. ALLOCATED(wait_stmt%unit)) THEN
      ! No syntax rule.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C939',  &
          LOCATION=QueryLocation(tlist(item_start)),  &
          MSG='A /file-unit-number/ shall be specified in a &
            &/wait-spec-list/.' )
    END IF
    
    !---------------------------------------------------------------------------
    ! That's it.
    
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseWaitStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /wait-spec/ (R923).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /wait-spec/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /wait-spec/ in @a tlist.
  !!
  !! @param[in,out] wait_stmt         Results of parsing the /wait-spec/.  
  !! Note that this is an INTENT(INOUT) argument - each spec sets a particular 
  !! component
  !!
  !! @param[out]    no_spec           Flag to indicate that the "XXX=" part of 
  !! the spec was absent.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_wait_spec( tlist, start, finish, wait_stmt, no_spec,  &
      err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE IoControlSpecs, ONLY: ParseFileUnitNumber
    USE MatchUtils
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(WaitStmt), INTENT(INOUT) :: wait_stmt
    LOGICAL, INTENT(OUT) :: no_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /wait-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R923'
    
    ! Keywords for each /wait-spec/.
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(7) = [  &
        scck_'UNIT  ', scck_'END   ', scck_'EOR   ', scck_'ERR   ',  &
        scck_'ID    ', scck_'IOMSG ', scck_'IOSTAT']
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: assign         ! Index of the =.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /wait-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Do we have an  argument keyword?
    
    assign = FindBare(tlist(:finish), ittAssign, scck_'=', start)
    IF (assign == 0) THEN
      ! No - this is a unit spec.
      no_spec = .TRUE.
      opt = 1
    ELSE
      no_spec = .FALSE.
      CALL ExpectKeyword(tlist, start, keywords, '/wait-spec/', opt, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    
      CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    SELECT CASE (opt)
    CASE (1)      ! UNIT
      IF (ALLOCATED(wait_stmt%unit)) THEN
        CALL already('UNIT')
      ELSE
        IF (assign == 0) THEN
          CALL ParseFileUnitNumber(  &
              tlist,  &
              start, finish,  &
              wait_stmt%unit,  &
              sub_err_list )
        ELSE
          CALL ParseFileUnitNumber(  &
              tlist,  &
              assign + 1, finish,  &
              wait_stmt%unit,  &
              sub_err_list )
        END IF
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (2)      ! END
      IF (ALLOCATED(wait_stmt%end)) THEN
        CALL already('END')
      ELSE
        ALLOCATE(wait_stmt%end)
        CALL ParseLabel(  &
            tlist,  &
            assign + 1,  &
            finish,  &
            wait_stmt%end,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE (3)     ! EOR
      IF (ALLOCATED(wait_stmt%eor)) THEN
        CALL already('EOR')
      ELSE
        ALLOCATE(wait_stmt%eor)
        CALL ParseLabel(  &
            tlist,  &
            assign + 1,  &
            finish,  &
            wait_stmt%eor,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE (4)     ! ERR
      IF (ALLOCATED(wait_stmt%err)) THEN
        CALL already('ERR')
      ELSE
        ALLOCATE(wait_stmt%err)
        CALL ParseLabel(  &
            tlist,  &
            finish,  &
            assign + 1,  &
            wait_stmt%err,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE (5)      ! ID 
      IF (ALLOCATED(wait_stmt%id)) THEN
        CALL already('ID')
      ELSE
        CALL ParseExpr(tlist, assign + 1, finish, wait_stmt%id, sub_err_list)
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (6)     ! IOMSG
      IF (ALLOCATED(wait_stmt%iomsg)) THEN
        CALL already('IOMSG')
      ELSE
        ! The syntax term is actually an /iomsg-variable/.
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            wait_stmt%iomsg,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (7)     ! IOSTAT
      IF (ALLOCATED(wait_stmt%iostat)) THEN
        CALL already('IOSTAT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            wait_stmt%iostat,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE DEFAULT
      STOP 'Unknown /wait-spec/.'
      
    END SELECT
    
  CONTAINS
    
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      ! No syntax rule.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C938',  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG=text // ' has already been specified for this /wait-stmt/.' )
    END SUBROUTINE already
    
  END SUBROUTINE parse_wait_spec
  
  
  !*****************************************************************************
  !!
  !> Implementation of WaitStmt%Dump - write a debugging representation 
  !! of the results of parsing a wait statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE wait_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WaitStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%unit)) THEN
      WRITE (unit, "(A,'UNIT = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%unit)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%end)) THEN
      WRITE (unit, "(A,'END = ',I0)") pfx, .ValueOf. pnode%end
    END IF
    IF (ALLOCATED(pnode%eor)) THEN
      WRITE (unit, "(A,'EOR = ',I0)") pfx, .ValueOf. pnode%eor
    END IF
    IF (ALLOCATED(pnode%err)) THEN
      WRITE (unit, "(A,'ERR = ',I0)") pfx, .ValueOf. pnode%err
    END IF
    IF (ALLOCATED(pnode%id)) THEN
      WRITE (unit, "(A,'ID = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%id)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%iomsg)) THEN
      WRITE (unit, "(A,'IOMSG = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%iomsg)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%iostat)) THEN
      WRITE (unit, "(A,'IOSTAT = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%iostat)
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE wait_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of WaitStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE wait_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WaitStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%unit%Query(visitor)
    
    IF (ALLOCATED(pnode%id)) CALL pnode%id%Query(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Query(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE wait_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of WaitStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE wait_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WaitStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%unit%Modify(visitor)
    
    IF (ALLOCATED(pnode%id)) CALL pnode%id%Modify(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Modify(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE wait_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of WaitStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'wait-stmt'.
  
  FUNCTION wait_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'wait-stmt'
    
  END FUNCTION wait_What
  
END MODULE WaitStmts
