! $Id: WhereConstructs.f90 2895 2019-04-07 01:12:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the WhereConstructs module.


!*******************************************************************************
!!
!> Procedures and types for parsing /where-construct/'s.

MODULE WhereConstructs
  
  USE Expressions
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseWhereConstruct
  PUBLIC :: ParseWhereStmtToConstruct
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing a particular "block" (as it is termed here) 
  !! of a where-construct.
  !!
  !! @todo Extend BlockStorage?
  TYPE :: sub_block
    !> The expression in the opening statement of the block.
    CLASS(Expression), ALLOCATABLE :: mask
    !> The block of executable statements following the initial statement.
    !!
    !! Always allocated, but maybe zero size.
    TYPE(SCParseNodeElement), ALLOCATABLE :: children(:)
  END TYPE sub_block
  
  !> Represents the results of parsing a /where-construct/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: WhereConstruct
    !> The construct name, if present.
    TYPE(Identifier), ALLOCATABLE :: construct_name
    !> The where or masked elsewhere blocks.
    !!
    !! Always allocated, may be zero sized.
    TYPE(sub_block), ALLOCATABLE :: blocks(:)
    !> The children of the elsewhere part.
    !!
    !! Not allocated if there was no elsewhere part.  Zero sized if there 
    !! was an elsewhere part, but that then had no child statements.
    TYPE(SCParseNodeElement), ALLOCATABLE :: else_children(:)
  CONTAINS
    PROCEDURE :: Dump => where_Dump
    PROCEDURE, NOPASS :: What => where_What
    PROCEDURE :: Query => where_Query
    PROCEDURE :: Modify => where_Modify
    PROCEDURE :: Advertise => where_Advertise
  END TYPE WhereConstruct
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a statement tree node for a /where-construct/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! construct.
  !!
  !! @param[out]    construct         The results of parsing the construct.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseWhereConstruct(st, construct, err_list)
    
    USE Errors
    USE ExecutionParts
    USE WhereConstructStNodes
    USE WhereStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(WhereConstructStNode), INTENT(INOUT) :: st
    TYPE(WhereConstruct), INTENT(OUT) :: construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: ib             ! Block index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ALLOCATE(construct%blocks(SIZE(st%blocks)))
    
    !---------------------------------------------------------------------------
    ! The opening where statement.
    
    CALL ParseWhereConstructStmt(  &
        st%blocks(1)%initial%tlist,  &
        construct%blocks(1)%mask,  &
        construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    ! Executable children of that initial part.
    CALL ParseBlock(  &
        st%blocks(1)%exec,  &
        construct%blocks(1)%children,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Subsequent masked elsewheres
    
    DO ib = 2, SIZE(st%blocks)
      CALL ParseMaskedElsewhereStmt(  &
          st%blocks(ib)%initial%tlist,  &
          construct%construct_name,  &
          construct%blocks(ib)%mask,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      
      ! Executable children of that subsequent part.
      CALL ParseBlock(  &
          st%blocks(ib)%exec,  &
          construct%blocks(ib)%children,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
    END DO
    
    !---------------------------------------------------------------------------
    ! The elsewhere block
    
    IF (ALLOCATED(st%else_block)) THEN
      CALL ParseElsewhereStmt(  &
          st%else_block%initial%tlist,  &
          construct%construct_name,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      
      CALL ParseBlock(  &
          st%else_block%exec,  &
          construct%else_children,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
    END IF
    
    !---------------------------------------------------------------------------
    ! The closing end where statement.
    
    CALL ParseEndWhereStmt(  &
        st%last%tlist,  &
        construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE ParseWhereConstruct
  
  
  !*****************************************************************************
  !!
  !> Implementation of WhereConstruct%Dump - write a debugging 
  !! representation of the results of parsing a where construct to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE where_Dump(pnode, unit, pfx)
    
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WhereConstruct), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: ib             ! Child block index.
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%construct_name))  &
        WRITE (unit, "(A,'construct_name = ',A)") pfx,  &
          .NameOf. pnode%construct_name
    
    ! The expression.
    WRITE (unit, "(A,'/mask-expr/ = ')", ADVANCE='NO') pfx
    CALL WriteExpr(unit, pnode%blocks(1)%mask)
    WRITE (unit, "()")
    
    WRITE (unit, "(A,'initial child statements:')") pfx
    CALL DumpBlock(pnode%blocks(1)%children, unit, pfx // '  ')
    
    DO ib = 2, SIZE(pnode%blocks)
      WRITE (unit, "(A,'masked-elsewhere expression:')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%blocks(ib)%mask)
      WRITE (unit, "()")
      
      WRITE (unit, "(A,'masked-elsewhere child statements:')") pfx
      CALL DumpBlock(pnode%blocks(ib)%children, unit, pfx // '  ')
    END DO
    
    IF (ALLOCATED(pnode%else_children)) THEN
      WRITE (unit, "(A,'elsewhere child statements:')") pfx
      CALL DumpBlock(pnode%else_children, unit, pfx // '  ')
    END IF
    
  END SUBROUTINE where_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of WhereConstruct%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'where-construct'.
  
  FUNCTION where_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'where-construct'
    
  END FUNCTION where_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of WhereConstruct%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE where_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(WhereConstruct), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child block index.
    INTEGER :: j              ! Child block child statement index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%blocks)
      DO j = 1, SIZE(pnode%blocks(i)%children)
        CALL pnode%blocks(i)%children(j)%Query(visitor)
      END DO
    END DO
    
    DO i = 1, SIZE(pnode%else_children)
      CALL pnode%else_children(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE where_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of WhereConstruct%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE where_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(WhereConstruct), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child block index.
    INTEGER :: j              ! Child block child statement index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%blocks)
      DO j = 1, SIZE(pnode%blocks(i)%children)
        CALL pnode%blocks(i)%children(j)%Modify(visitor)
      END DO
    END DO
    
    DO i = 1, SIZE(pnode%else_children)
      CALL pnode%else_children(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE where_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of WhereConstruct%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE where_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WhereConstruct), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=.NameOf. pnode%construct_name,  &
            NATURES=[NatureDefConstruct], &
            LOCATIONS=[.LocationOf. pnode%construct_name] ) )
    END IF
    
  END SUBROUTINE where_Advertise
  
  
  !*****************************************************************************
  !!
  !> Parse an IfStmt through to an IfConstruct structure.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    if_construct      The results of parsing the /if-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseWhereStmtToConstruct(tlist, construct, err_list)
    
    USE Classifier
    USE Errors
    USE ErrorCodes
    USE ExecStNodes
    USE ExecutionParts
    USE SingleStNodes
    USE Statements
    USE WhereStmts
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(INOUT) :: tlist(:)
    TYPE(WhereConstruct), INTENT(OUT) :: construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /forall-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R741'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Tokens that make up the child /forall-assignment-stmt/ of the 
    ! /forall-stmt/.
    TYPE(Token), ALLOCATABLE :: child_tlist(:)
    
    ! Execution part statement tree node.
    TYPE(ExecStNode) :: exec_st
    
    INTEGER :: ist            ! Child statement number.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ALLOCATE(construct%blocks(1))
    
    !---------------------------------------------------------------------------
    ! Parse the leading where-stmt.
    
    CALL ParseWhereStmt(  &
        tlist,  &
        construct%blocks(1)%mask,  &
        child_tlist,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    ! No point continuing if we can't parse the front bit.
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Work out what the action statement is.
    
    CALL Classify(child_tlist, [INTEGER ::], ist)
    SELECT CASE (ist)
    CASE (istUnclassified)
      CALL Add( err_list,  &
          CODE=errUnclassifiedStatement,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(child_tlist(1)),  &
          MSG='Unclassifiable /forall-assignment-stmt/.' )
      
    CASE (istAssignment)
      ! Parse it.
      CALL MakeExecStNode(child_tlist, ist, exec_st)
      CALL ParseBlock(exec_st, construct%blocks(1)%children, sub_err_list)
      CALL Add(err_list, sub_err_list)
      
    CASE DEFAULT
      CALL Add( err_list,  &
          CODE=errIfBadChildStmt,  &
          COMPONENT='R748 via ' // comp,  &  
          LOCATION=QueryLocation(child_tlist(1)),  &
          MSG='A ' // GetStName(ist) // ' is not a valid &
            &/where-assignment-stmt/.' )
      RETURN
      
    END SELECT
    
  END SUBROUTINE ParseWhereStmtToConstruct
  
END MODULE WhereConstructs
