! $Id: WriteStmts.f90 2803 2019-03-22 19:36:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the WriteStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /write-stmt/'s (R911).

MODULE WriteStmts
  
  USE InputOutputItems
  USE IoControlSpecs
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseWriteStmt
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing a /write-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: WriteStmt
    !> The /io-control-spec-list/.
    TYPE(IoControlSpecList) :: io_control_spec_list
    !> The list of input items.
    !!
    !! Always allocated, but might be zero size.
    TYPE(IOItemList), ALLOCATABLE :: item_list(:)
  CONTAINS
    PROCEDURE :: Dump => write_Dump
    PROCEDURE :: Query => write_Query
    PROCEDURE :: Modify => write_Modify
    PROCEDURE, NOPASS :: What => write_What
    PROCEDURE :: Check => write_Check
  END TYPE WriteStmt
  
  !-----------------------------------------------------------------------------
  
  ! Syntax rule number for an /write-stmt/.
  CHARACTER(*), PARAMETER :: comp = 'R911'
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /write-stmt/ (R911).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    write_stmt        The results of parsing the /write-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseWriteStmt(tlist, write_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(WriteStmt), INTENT(OUT) :: write_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /write-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! WRITE...
    
    CALL Expect(tlist, 1, ittName, scck_'WRITE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN

    !---------------------------------------------------------------------------
    ! WRITE(...)
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The /io-control-spec-list/.
    
    CALL ParseIOControlSpecList(  &
        tlist,  &
        3,  &
        closing - 1,  &
        .FALSE.,  &
        write_stmt%io_control_spec_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ! Continue on error.
    
    ! That can be it.
    IF (closing + 1 > SIZE(tlist)) THEN
      ALLOCATE(write_stmt%item_list(0))
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! The /output-item-list/.
    
    CALL ParseIOItemList(  &
        tlist,  &
        closing + 1, SIZE(tlist),  &
        .TRUE.,  &            ! Specify an /output-item-list/.
        write_stmt%item_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseWriteStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of WriteStmt%Dump - write a debugging representation 
  !! of the results of parsing a write statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE write_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WriteStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/io-control-spec-list/')") pfx
    CALL DumpIoControlSpecList(unit, pnode%io_control_spec_list, LEN(pfx) + 2)
      
    WRITE (unit, "(A,'/output-item-list/')") pfx
    CALL DumpIOItemList(pnode%item_list, unit, pfx // '  ')
    
  END SUBROUTINE write_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of WriteStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE write_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WriteStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%io_control_spec_list%Query(visitor)
    
    CALL QueryIOItemList(pnode%item_list, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE write_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of WriteStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE write_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WriteStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%io_control_spec_list%Modify(visitor)
    
    CALL ModifyIOItemList(pnode%item_list, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE write_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of WriteStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'write-stmt'.
  
  FUNCTION write_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'write-stmt'
    
  END FUNCTION write_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of WriteStmt%Check - check the parse node for 
  !! conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE write_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WriteStmt), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Component identifier for the /io-control-spec/.
    CHARACTER(*), PARAMETER :: ics_comp = 'R913 via ' // comp
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF ( ALLOCATED(pnode%io_control_spec_list%nml)  &
        .AND. (SIZE(pnode%item_list) /= 0) ) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C916 on R913 via ' // comp,  &
          LOCATION=.LocationOf. pnode%io_control_spec_list%nml,  &
          MSG='A /namelist-group-name/ shall not appear if the /write-stmt/ &
            &contains an /output-item-list/.' )
    END IF
    
    ! Similar comments as for ReadStmt%Check - these were split out from 
    ! basic parsing of an IoControlSpec.
    
    IF (ALLOCATED(pnode%io_control_spec_list%blank)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C914 on ' // ics_comp,  &
          LOCATION=.LocationOf. pnode%io_control_spec_list%blank,  &
          MSG='A "BLANK=" specifier shall not appear in a /write-stmt/.' )
    END IF
    IF (ALLOCATED(pnode%io_control_spec_list%pad)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C914 on ' // ics_comp,  &
          LOCATION=.LocationOf. pnode%io_control_spec_list%pad,  &
          MSG='A "PAD=" specifier shall not appear in a /write-stmt/.' )
    END IF
    IF (ALLOCATED(pnode%io_control_spec_list%end)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C914 on ' // ics_comp,  &
          LOCATION=.LocationOf. pnode%io_control_spec_list%end,  &
          MSG='An "END=" specifier shall not appear in a /write-stmt/.' )
    END IF
    IF (ALLOCATED(pnode%io_control_spec_list%eor)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C914 on ' // ics_comp,  &
          LOCATION=.LocationOf. pnode%io_control_spec_list%eor,  &
          MSG='An "EOR=" specifier shall not appear in a /write-stmt/.' )
    END IF
    IF (ALLOCATED(pnode%io_control_spec_list%size)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C914 on ' // comp,  &
          LOCATION=.LocationOf. pnode%io_control_spec_list%size,  &
          MSG='A "SIZE=" specifier shall not appear in a /write-stmt/.' )
    END IF
    
  END SUBROUTINE write_Check
  
END MODULE WriteStmts
